########################################
# adapted from: https://github.com/HazyResearch/fm_data_tasks/blob/updates/fm_data_tasks/utils/data_utils.py 
########################################
"""Constants."""

import os
from pathlib import Path

# os.environ["DATASET_PATH"] = "/Users/Effy/workspace/full_llm_wrangler/data/datasets/"
DATASET_PATH = os.environ.get("DATASET_PATH", Path("data/datasets").resolve())
# print(f"Current DATASET_PATH: {DATASET_PATH}")

DATA2TASK = {
    f"{DATASET_PATH}/entity_matching/structured/Amazon-Google": "entity_matching",
    f"{DATASET_PATH}/entity_matching/structured/Beer": "entity_matching",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-ACM": "entity_matching",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-GoogleScholar": "entity_matching",
    f"{DATASET_PATH}/entity_matching/structured/Fodors-Zagats": "entity_matching",
    f"{DATASET_PATH}/entity_matching/structured/iTunes-Amazon": "entity_matching",
    f"{DATASET_PATH}/entity_matching/structured/Walmart-Amazon": "entity_matching",
    f"{DATASET_PATH}/data_imputation/Buy": "data_imputation",
    f"{DATASET_PATH}/data_imputation/Restaurant": "data_imputation",
    f"{DATASET_PATH}/error_detection/Hospital": "error_detection_spelling",
    f"{DATASET_PATH}/error_detection/Adult": "error_detection_spelling",
    f"{DATASET_PATH}/schema_matching/Synthea": "schema_matching",
    f"{DATASET_PATH}/data_transformation/benchmark-bing-query-logs": "data_transformation",
    f"{DATASET_PATH}/data_transformation/benchmark-stackoverflow": "data_transformation",
    f"{DATASET_PATH}/data_transformation/benchmark-FF-Trifacta-GoogleRefine": "data_transformation",
    f"{DATASET_PATH}/data_transformation/benchmark-headcase": "data_transformation",
    f"{DATASET_PATH}/data_transformation/benchmark-bing-query-logs-semantics": "data_transformation",
    f"{DATASET_PATH}/data_transformation/benchmark-bing-query-logs-unit": "data_transformation",
    f"{DATASET_PATH}/data_transformation/md-query-logs": "data_transformation"
}

IMPUTE_COLS = {
    f"{DATASET_PATH}/data_imputation/Buy": "manufacturer",
    f"{DATASET_PATH}/data_imputation/Restaurant": "city",
}

MATCH_PROD_NAME = {
    f"{DATASET_PATH}/entity_matching/structured/Amazon-Google": "Product",
    f"{DATASET_PATH}/entity_matching/structured/Beer": "Product",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-ACM": "Product",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-GoogleScholar": "Product",
    f"{DATASET_PATH}/entity_matching/structured/Fodors-Zagats": "Product",
    f"{DATASET_PATH}/entity_matching/structured/iTunes-Amazon": "Song",
    f"{DATASET_PATH}/entity_matching/structured/Walmart-Amazon": "Product",
    f"{DATASET_PATH}/schema_matching/Synthea": "",
}

# Dropping happens before renaming
DATA2DROPCOLS = {
    f"{DATASET_PATH}/entity_matching/structured/Amazon-Google": [],
    f"{DATASET_PATH}/entity_matching/structured/Beer": ["Style", "ABV"],
    f"{DATASET_PATH}/entity_matching/structured/DBLP-ACM": [],
    f"{DATASET_PATH}/entity_matching/structured/DBLP-GoogleScholar": [],
    f"{DATASET_PATH}/entity_matching/structured/Fodors-Zagats": [],
    f"{DATASET_PATH}/entity_matching/structured/iTunes-Amazon": ["CopyRight"],
    f"{DATASET_PATH}/entity_matching/structured/Walmart-Amazon": [
        "category",
        "price",
        "brand",
    ],
    f"{DATASET_PATH}/data_imputation/Buy": ["manufacturer"],
    f"{DATASET_PATH}/data_imputation/Restaurant": ["city"],
    f"{DATASET_PATH}/error_detection/Hospital": [],
    f"{DATASET_PATH}/error_detection/Adult": [],
    f"{DATASET_PATH}/schema_matching/Synthea": ["des1", "des2", "d1", "d2", "d3", "d4"],
    f"{DATASET_PATH}/data_transformation/benchmark-bing-query-logs": [],
    f"{DATASET_PATH}/data_transformation/benchmark-stackoverflow": [],
    f"{DATASET_PATH}/data_transformation/md-query-logs": [],
    f"{DATASET_PATH}/data_transformation/benchmark-FF-Trifacta-GoogleRefine": []
}

DATA2COLREMAP = {
    f"{DATASET_PATH}/entity_matching/structured/Amazon-Google": {},
    f"{DATASET_PATH}/entity_matching/structured/Beer": {
        "id": "id",
        "Beer_Name": "name",
        "Brew_Factory_Name": "factory",
        "Style": "style",
        "ABV": "ABV",
    },
    f"{DATASET_PATH}/entity_matching/structured/DBLP-ACM": {},
    f"{DATASET_PATH}/entity_matching/structured/DBLP-GoogleScholar": {},
    f"{DATASET_PATH}/entity_matching/structured/Fodors-Zagats": {},
    f"{DATASET_PATH}/entity_matching/structured/iTunes-Amazon": {
        "id": "id",
        "Song_Name": "name",
        "Artist_Name": "artist name",
        "Album_Name": "album name",
        "Genre": "genre",
        "Price": "price",
        "CopyRight": "CopyRight",
        "Time": "time",
        "Released": "released",
    },
    f"{DATASET_PATH}/entity_matching/structured/Walmart-Amazon": {},
    f"{DATASET_PATH}/data_imputation/Buy": {},
    f"{DATASET_PATH}/data_imputation/Restaurant": {},
    f"{DATASET_PATH}/error_detection/Hospital": {},
    f"{DATASET_PATH}/error_detection/Adult": {},
    f"{DATASET_PATH}/schema_matching/Synthea": {
        "omop": "left",
        "table": "right",
        "label": "label",
    },
    f"{DATASET_PATH}/data_transformation/benchmark-bing-query-logs": {},
    f"{DATASET_PATH}/data_transformation/benchmark-stackoverflow": {},
    f"{DATASET_PATH}/data_transformation/md-query-logs": {},
    f"{DATASET_PATH}/data_transformation/benchmark-FF-Trifacta-GoogleRefine": {}
}


DATA2INSTRUCT = {
    f"{DATASET_PATH}/entity_matching/structured/Amazon-Google": "Are Product A and Product B the same? Yes or No? ",
    f"{DATASET_PATH}/entity_matching/structured/Beer": "Are Product A and Product B the same? Yes or No? ",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-ACM": "Are Product A and Product B the same? Yes or No? ",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-GoogleScholar": "Are Product A and Product B the same? Yes or No? ",
    f"{DATASET_PATH}/entity_matching/structured/Fodors-Zagats": "Are Product A and Product B the same? Yes or No? ",
    f"{DATASET_PATH}/entity_matching/structured/iTunes-Amazon": "Are Song A and Song B the same? Yes or No? ",
    f"{DATASET_PATH}/entity_matching/structured/Walmart-Amazon": "Are Product A and Product B the same? Yes or No? ",
    f"{DATASET_PATH}/data_imputation/Buy": "Who is the manufacturer? apple, sony, lg electronics? ",
    f"{DATASET_PATH}/data_imputation/Restaurant": "What is the city? san fransisco, new york, denver? ",
    f"{DATASET_PATH}/error_detection/Hospital": "Is there a x spelling error? Yes or No? ",
    f"{DATASET_PATH}/error_detection/Hospital": "Is there a x spelling error? Yes or No? ",
    f"{DATASET_PATH}/schema_matching/Synthea": "Are A and B semantically equal?\n\n",
    f"{DATASET_PATH}/error_detection/Adult": "Classify errors. nan values are not errors.",
    f"{DATASET_PATH}/data_transformation/benchmark-bing-query-logs": "Transform from input to output",
    f"{DATASET_PATH}/data_transformation/benchmark-stackoverflow": "Transform from input to output"
}

DATA2SUFFIX = {
    f"{DATASET_PATH}/entity_matching/structured/Amazon-Google": " Are Product A and Product B the same?",
    f"{DATASET_PATH}/entity_matching/structured/Beer": " Are Product A and Product B the same?",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-ACM": " Are Product A and Product B the same?",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-GoogleScholar": " Are Product A and Product B the same?",
    f"{DATASET_PATH}/entity_matching/structured/Fodors-Zagats": " Are Product A and Product B the same?",
    f"{DATASET_PATH}/entity_matching/structured/iTunes-Amazon": " Same Song?",
    f"{DATASET_PATH}/entity_matching/structured/Walmart-Amazon": " Are A and B the Same?",
    f"{DATASET_PATH}/data_imputation/Buy": " Who is the manufacturer?",
    f"{DATASET_PATH}/data_imputation/Restaurant": " What is the city?",
    f"{DATASET_PATH}/error_detection/Hospital": "?",
    f"{DATASET_PATH}/schema_matching/Synthea": "Are A and B the same? Yes or No?",
    f"{DATASET_PATH}/error_detection/Adult": "? ",
    f"{DATASET_PATH}/data_transformation/benchmark-bing-query-logs": "?",
    f"{DATASET_PATH}/data_transformation/benchmark-stackoverflow": "?"
}


DATA2EXAMPLE_SUBKEY_ATTR = {
    f"{DATASET_PATH}/entity_matching/structured/Amazon-Google": "manufacturer_A",
    f"{DATASET_PATH}/entity_matching/structured/Beer": None,
    f"{DATASET_PATH}/entity_matching/structured/DBLP-ACM": "venue_A",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-GoogleScholar": "venue_A",
    f"{DATASET_PATH}/entity_matching/structured/Fodors-Zagats": None,
    f"{DATASET_PATH}/entity_matching/structured/iTunes-Amazon": None,
    f"{DATASET_PATH}/entity_matching/structured/Walmart-Amazon": None,
    f"{DATASET_PATH}/data_imputation/Buy": None,
    f"{DATASET_PATH}/data_imputation/Restaurant": None,
    f"{DATASET_PATH}/error_detection/Hospital": "col_name",
    f"{DATASET_PATH}/schema_matching/Synthea": "left",
    f"{DATASET_PATH}/error_detection/Adult": "col_name",
    f"{DATASET_PATH}/data_transformation/benchmark-bing-query-logs": None,
    f"{DATASET_PATH}/data_transformation/benchmark-stackoverflow": None
}


PREFIXES = {
    f"{DATASET_PATH}/entity_matching/structured/iTunes-Amazon": "Song A is name: Illusion (  feat . Emeli Sand ' © & Professor Green ). Song B is name: Transmission [ feat . Emeli Sand ' © & Professor Green ]. Same Song? No\n\nSong A is name: Praise the sky ( feat . Meghan Trainor ). Song B is name: Praise the sky. Same Song? Yes\n\nSong A is name: Drive it  ( Featuring Jackson T. Brown ). Song B is name: Drive it ( Featuring Jackson T. Brown ) ( Main Version ). Same Song? Yes\n\nSong A is name: Jump Up ( feat . Tyler ). Song B is name: Jump Up ( feat . Tyler ) [ Explicit ]. Same Song? Yes\n\nSong A is name: Fight Song Anthem. genre: Country , Music , Urban Cowboy , Contemporary Country. Song B is name: Fight Song Anthem [ Explicit ]. genre: Country. Same Song? Yes\n\nSong A is name: Over it. Song B is name: Over it [ Clean ]. Same Song? Yes\n\nSong A is name: Original Don ( feat . Nicki Minaj ) ( Extended Version ). Song B is name: Original Don ( feat . Nicki Minaj ) [Remix]. Same Song? No\n\nSong A is name: Amazing 2.0 ( feat . Justin Bieber ). time: 3:14. Song B is name: Amazing ( feat . Justin Bieber ).  time: 6:14. Same Song? Yes\n\nSong A is name: Run this town ( feat . JT ). Song B is name: Fight the law ( feat . JT ). Same Song? No\n",
    f"{DATASET_PATH}/entity_matching/structured/Beer": "Product A is name: Bourbon Barrel Red Oak Ale. factory: Big Brewing Company. style: American Amber / Red Ale. ABV: 7.60%. Product B is name: Bourbon Barrel Red Oak Ale. factory: Halo Street Brewery. style: American Amber / Red Ale. ABV: 7.60%. Are Product A and Product B the same? Yes\n\nProduct A is name: Red Imperial Red Ale. factory: Brew Brewing Company. style: American Amber / Red Ale. ABV: 5.40%. Product B is name: Red Imperial Red Ale - Bourbon Barrel Aged. factory: Brew Brewing Company. style: American Amber / Red Ale. ABV: 5.40%. Are Product A and Product B the same? No\n\nProduct A is name: Red Rocket Amber Ale. factory: Third Base Sports Bar & Brewery. style: American Amber / Red Ale. ABV: 6.40%. Product B is name: Blue Cat Red Toad Amber Ale. factory: Blue Cat Brew Pub. style: Premium Bitter/ESB. ABV: 5.70%. Are Product A and Product B the same? No\n\nProduct A is name: The Gentle Mongoose Holy City. factory: Brewing City Brewing. style: American Amber / Red Ale. ABV: 6.90%. Product B is name: Holy City The Gentle Mongoose. factory: Holy City Brewing. style: Amber Ale. ABV: 6.90%. Are Product A and Product B the same? Yes\n\nProduct A is name: Freedom Soho Red. factory: Freedom Brewery Ltd. style: American Amber / Red Ale. ABV: 4.70%. Product B is name: Freedom Soho Red. factory: Freedom. style: Amber Lager/Vienna. ABV: 4.70%. Are Product A and Product B the same? Yes\n\nProduct A is name: Blarney Rock Irish Ale. factory: Rockyard Brewing. style: American Amber / Red Ale. ABV: 6.00%. Product B is name: Rockyard Blarney Rock Irish Ale. factory: Rockyard Brewing Company. style: Irish Ale. ABV: -. Are Product A and Product B the same? Yes\n\nProduct A is name: Red Rocket Amber Ale. factory: Third Base Sports Bar & Brewery. style: American Amber / Red Ale. ABV: 6.40%. Product B is name: Red Duck Amber Ale. factory: Purrumbete Brewing. style: Amber Ale. ABV: 4.80%. Are Product A and Product B the same? No\n",
    f"{DATASET_PATH}/entity_matching/structured/Fodors-Zagats": "Product A is name: ` bamboo garden '. addr: ' 4850 flamingo rd. '. city: ` las vegas '. phone: 702/871 -3262. type: asian. class: 425. Product B is name: ` buzio \\ 's in the rio '. addr: ' 3700 w. flamingo rd. '. city: ` las vegas '. phone: 702-252-7697. type: seafood. class: 658. Are Product A and Product B the same? No, Product A class: 425 and Product B class: 658 are not the same\n\nProduct A is name: ` le chardonnay '. addr: ' 8284 melrose ave. '. city: ` los angeles '. phone: 213/655 -8880. type: french. class: type: 12. Product B is name: ` le chardonnay ( los angeles ) '. addr: ' 8284 melrose ave. '. city: ` los angeles '. phone: 213-655-8880. type: ` french bistro '. class: type: 12. Are Product A and Product B the same? Yes\n\nProduct A is name: lespinasse. addr: ' 2 e. 55th st. '. city: ` new york '. phone: 212/339 -6719. type: american. class: type: 43. Product B is name: ` lespinasse ( new york city ) '. addr: ' 2 e. 55th st. '. city: ` new york city '. phone: 212-339-6719. type: asian. class: type: 43. Are Product A and Product B the same? Yes\n\nProduct A is name: ` cafe roma '. addr: ' 3570 las vegas blvd. s '. city: ` las vegas '. phone: 702/731 -7547. type: ` coffee shops/diners '. class: 433. Product B is name: 'em eril \\ 's new orleans fish house '. addr: ' 3799 las vegas blvd. s. '. city: ` las vegas '. phone: 702-891-7374. type: seafood. class: 659. Are Product A and Product B the same? No, Product A class: 443 and Product B class: 659 are not the same\n\nProduct A is name: ` smith & wollensky '. addr: ' 201 e. 49th st. '. city: ` new york '. phone: 212/753 -1530. type: american. class: 62. Product B is name: ` smith & wollensky '. addr: ' 797 third ave. '. city: ` new york city '. phone: 212-753-1530. type: steakhouses. class: type: 62. Are Product A and B the same? Yes\n",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-GoogleScholar": {
        "nan": "Product A is title: keynote address. authors: . venue: . year: 1995. Product B is title: keynote address. authors: t bailey. venue: women in aviation international conference , reno , nv , march ,. year: . Same Product? No\n\nProduct A is title: foreword : management of semistructured data. authors: . venue: . year: 1997. Product B is title: lore : a database management system for semistructued data. authors: j mchugh , s abiteboul , r goldmanâ ? ¦. venue: acm sigmod ,. year: . Same Product? No\n\nProduct A is title: guest editorial. authors: . venue: . year: 2003. Product B is title: guest editorial. authors: v atluri , a joshi , y yesha. venue: the vldb journal the international journal on very large & hellip ; ,. year: 2003.0. Same Product? Yes\n\nProduct A is title: storage and retrieval of xml data using relational databases. authors: . venue: . year: 2001. Product B is title: storage and retrieval of xml data using relational databases. authors: s chaudhuri , k shim. venue: data engineering , 2003 . proceedings . 19th international & hellip ; ,. year: 2003.0. Same Product? Yes\n\nProduct A is title: editorial. authors: . venue: . year: 2001. Product B is title: eic editorial. authors: ps yu. venue: ieee transactions on knowledge and data engineering ,. year: 2004.0. Same Product? No\n\nProduct A is title: state of the art in workflow management research and products. authors: . venue: . year: 1996. Product B is title: workflow management systems : state of the art on research and products , tutorial notes. authors: c mohan. venue: int . conf . on extending database technology , avignon ,. year: . Same Product? Yes\n\nProduct A is title: information management for genome level bioinformatics. authors: . venue: . year: 2001. Product B is title: implementing information management strategies .. authors: cet rohm. venue: new directions for higher education ,. year: 1986.0. Same Product? No\n\nProduct A is title: open object database management systems. authors: . venue: . year: 1994. Product B is title: case for output management systems .. authors: k schwaber. venue: computerworld. ,. year: 1981.0. Same Product? No\n\nProduct A is title: querying and mining data streams : you only get one look a tutorial. authors: . venue: . year: 2002. Product B is title: query and mining data streams : you only get one look. authors: m garofalakis , j gehrke , r rastogi. venue: . year: . Same Product? Yes\n\nProduct A is title: reminiscences on influential papers. authors: . venue: . year: 2001. Product B is title: reminiscences on influential papers. authors: ka ross. venue: sigmod. year: 2002.0. Same Product? Yes\n\n",
        "acm trans . database syst .": "Product A is title: semantics for update rule programs and implementations in a relational database management system. authors: l raschid , j lobo. venue: acm trans . database syst .. year: 1996. Product B is title: query processing in a relational database management system. authors: k youssefi , e wong. venue: vldb ,. year: 1979.0. Same Product? No\n\nProduct A is title: an efficient method for checking object-oriented database schema correctness. authors: a formica , h grіger , m missikoff. venue: acm trans . database syst .. year: 1998. Product B is title: an efficient method for checking object-oriented database schema correctness. authors: a formica , hd groger , m missikoff. venue: acm transactions on database systems ,. year: 1998.0. Same Product? Yes\n\nProduct A is title: semantics for update rule programs and implementations in a relational database management system. authors: l raschid , j lobo. venue: acm trans . database syst .. year: 1996. Product B is title: performance analysis of a relational data base management system. authors: p hawthorn , m stonebraker. venue: proceedings of the 1979 acm sigmod international conference. year: . Same Product? No\n\nProduct A is title: theory of dependence values. authors: r meo. venue: acm trans . database syst .. year: 2000. Product B is title: theory of dependence values. authors: r meo. venue: acm transactions on database systems ,. year: 2000.0. Same Product? Yes\n\nProduct A is title: gloss : text-source discovery over the internet. authors: l gravano , h garcia-molina , a tomasic. venue: acm trans . database syst .. year: 1999. Product B is title: gloss : text-source discovery over the internet. authors: l gravano , h garcia-molina , a tomasic. venue: acm transactions on database systems ,. year: 1999.0. Same Product? Yes\n\nProduct A is title: spatial queries in dynamic environments. authors: y tao , d papadias. venue: acm trans . database syst .. year: 2003. Product B is title: location-based spatial queries with data sharing in mobile environment. authors: ws ku , r zimmermann , cn wan. venue: . year: . Same Product? No\n\nProduct A is title: path sharing and predicate evaluation for high-performance xml filtering. authors: y diao , m altinel , m franklin , h zhang , p fischer. venue: acm trans . database syst .. year: 2003. Product B is title: yfilter : path sharing and predicate evaluation for high-performance xmpsll filtering. authors: y diao , m altinel , mj franklin , h zhang , p fischer. venue: acm transactions on database systems ( tods ). year: . Same Product? Yes\n\nProduct A is title: a cost model for query processing in high dimensional data spaces. authors: c bіhm. venue: acm trans . database syst .. year: 2000. Product B is title: generalized multi-dimensional data mapping and query processing. authors: r zhang , p kalnis , bc ooi , kl tan. venue: acm transactions on database systems ,. year: 2005.0. Same Product? No\n\nProduct A is title: applying an update method to a set of receivers. authors: m andries , l cabibbo , j paredaens , j bussche. venue: acm trans . database syst .. year: 2001. Product B is title: applying an update method to a set of receivers. authors: m andries , l cabibbo , j paredaens , j van den. venue: acm transactions on database systems ,. year: 2001.0. Same Product? Yes\n\nProduct A is title: formal query languages for secure relational databases. authors: m winslett , k smith , x qian. venue: acm trans . database syst .. year: 1994. Product B is title: implementation of logiclal query languages for databases ( abstract only ). authors: jd ullman. venue: proceedings of the 1985 acm sigmod international conference & hellip ; ,. year: 1985.0. Same Product? No\n\n",
        "ecord": "Product A is title: david dewitt speaks out. authors: m winslett. venue: ecord. year: 2002. Product B is title: jim gray speaks out. authors: m winslett. venue: sigmod record ,. year: 2003.0. Same Product? No\n\nProduct A is title: david dewitt speaks out. authors: m winslett. venue: ecord. year: 2002. Product B is title: david dewitt speaks out. authors: s be proud , hsfis very , p spent. venue: sigmod record ,. year: 2002.0. Same Product? Yes\n\nProduct A is title: david dewitt speaks out. authors: m winslett. venue: ecord. year: 2002. Product B is title: pat selinger speaks out. authors: p selinger , m winslett. venue: sigmod record ,. year: . Same Product? No\n\n",
        "sigmod conference": "Product A is title: dynamat : a dynamic view management system for data warehouses. authors: y kotidis , n roussopoulos. venue: sigmod conference. year: 1999. Product B is title: dynamat : a dynamic viewmanagement system fordata warehouses. authors: ykn roussopoulos. venue: proc . of the acm sigmod. year: . Same Product? Yes\n\nProduct A is title: data management issues in electronic commerce ( panel ). authors: m жzsu. venue: sigmod conference. year: 1999. Product B is title: special section on data management issues in e-commerce. authors: a dogac. venue: acm sigmod record ,. year: . Same Product? No\n\nProduct A is title: database systems management and oracle8. authors: c doherty. venue: sigmod conference. year: 1998. Product B is title: the cedar database management system. authors: m brownâ ? ¦. venue: proc . 1981 acmsigmod conference on management of data , ann & hellip ; ,. year: . Same Product? No\n\nProduct A is title: online association rule mining. authors: c hidber. venue: sigmod conference. year: 1999. Product B is title: mining fuzzy association rules in database. authors: mk chan , f ada , hw man. venue: sigmod record ,. year: . Same Product? No\n\nProduct A is title: a performance study of transitive closure algorithms. authors: s dar , r ramakrishnan. venue: sigmod conference. year: 1994. Product B is title: a performance study of transitive closure algorithms. authors: s dar , r ramakrishnan. venue: sigmod conference ,. year: 1994.0. Same Product? Yes\n\nProduct A is title: an open storage system for abstract objects. authors: s blott , l relly , h schek. venue: sigmod conference. year: 1996. Product B is title: an abstract-object storage manager. authors: s blott , l relly , hj schek. venue: & hellip ; 1996 acm sigmod int . conf . on management of data , acm sigmod. year: . Same Product? Yes\n\nProduct A is title: communication efficient distributed mining of association rules. authors: a schuster , r wolff. venue: sigmod conference. year: 2001. Product B is title: mining generalized association rules. authors: r agrawal , r srikant. venue: proceedings of the 1995 international conference of very & hellip ; ,. year: . Same Product? No\n\nProduct A is title: ariadne : a system for constructing mediators for internet sources. authors: j ambite , n ashish , g barish , c knoblock , s minton , p modi , i muslea , a philpot , s tejada. venue: sigmod conference. year: 1998. Product B is title: . and sheila tejada . ariadne : a system for constructing mediators for internet sources ( system. authors: jl ambite , n ashish , g barish. venue: proc . of a cm sigmod conf . on management of. year: . Same Product? Yes\n\nProduct A is title: dna-miner : a system prototype for mining dna sequences. authors: j han , h jamil , y lu , l chen , y liao , j pei. venue: sigmod conference. year: 2001. Product B is title: n. stefanovic 1997 â ?? geominer : a system prototype for spatial data miningâ ??. authors: j han , k koperski. venue: proc . acm-sigmod int . conf . on management of data ( sigmod &#39; 97. year: . Same Product? No\n\nProduct A is title: a critique of ansi sql isolation levels. authors: h berenson , p bernstein , j gray , j melton , e o'neil , p o'neil. venue: sigmod conference. year: 1995. Product B is title: a critique of ansi sql isolation levels. authors: h berenson , p bernstein , j gray , j melton. venue: proceedings of the acm sigmod international conference on & hellip ; ,. year: 1995.0. Same Product? Yes\n\n",
        "sigmod record": "Product A is title: exploiting main memory dbms features to improve real-time concurrency control protocols. authors: ж ulusoy , a buchmann. venue: sigmod record. year: 1996. Product B is title: index concurrency control in firm real-time dbms. authors: b goyal , j haritsa , s seshadri , v srinivasan. venue: proceedings of the international conference on very large & hellip ; ,. year: 1995.0. Same Product? No\n\nProduct A is title: database techniques for the world-wide web : a survey. authors: d florescu , a levy , a mendelzon. venue: sigmod record. year: 1998. Product B is title: mendelzon . database techniques for the world-w ide-wide : a survey. authors: d florescu , a levy. venue: sigmod record ,. year: . Same Product? Yes\n\nProduct A is title: an extensible compressor for xml data. authors: h liefke , d suciu. venue: sigmod record. year: 2000. Product B is title: xmill : an efficient compressor for xml data. authors: l hartmut , d suciu. venue: . year: . Same Product? No\n\nProduct A is title: a consensus glossary of temporal database concepts. authors: c jensen , j clifford , r elmasri , s gadia , p hayes , s jajodia. venue: sigmod record. year: 1994. Product B is title: , sk gadi p. hayes and s. jajodia ( e & ) : â ?? a consensus glossary of temporal database conceptsâ ??. authors: cs jensen , j cliftord. venue: a chf sig3fod. year: . Same Product? Yes\n\nProduct A is title: on the issue of valid time ( s ) in temporal databases. authors: s kokkotos , e ioannidis , t panayiotopoulos , c spyropoulos. venue: sigmod record. year: 1995. Product B is title: temporal databases. authors: rt snodgrass. venue: . year: . Same Product? No\n\nProduct A is title: asserting beliefs in mls relational models. authors: n jukic , s vrbsky. venue: sigmod record. year: 1997. Product B is title: learning probabilistic relational models. authors: l getoorâ ? ¦. venue: lecture notes in computer science ,. year: 2000.0. Same Product? No\n\nProduct A is title: mining fuzzy association rules in databases. authors: c kuok , a fu , m wong. venue: sigmod record. year: 1998. Product B is title: mining association rules between sets of items in large databases . 1993. authors: r agrawal , t imielinski , a swami. venue: proc . acm sigmod international conference on management of. year: . Same Product? No\n\nProduct A is title: supply chain infrastructures : system integration and information sharing. authors: m ball , m ma , l raschid , z zhao. venue: sigmod record. year: 2002. Product B is title: supply chain infrastructures : system integration and information sharing. authors: mo ball , m ma , l raschid , z zhao. venue: sigmod record ,. year: 2002.0. Same Product? Yes\n\nProduct A is title: database research : achievements and opportunities into the 21st century. authors: a silberschatz , m stonebraker , j ullman. venue: sigmod record. year: 1996. Product B is title: database research : achievements and opportunities into the 1st century. authors: a silberschatz , m stonebraker , j ullman. venue: acm sigmod record ,. year: 1996.0. Same Product? Yes\n\nProduct A is title: constraint databases : a tutorial introduction. authors: j bussche. venue: sigmod record. year: 2000. Product B is title: constraint databases : a tutorial introduction .. authors: acms anthology. venue: sigmod record ,. year: 2000.0. Same Product? Yes\n\n",
        "vldb": "Product A is title: fast algorithms for mining association rules in large databases. authors: r agrawal , r srikant. venue: vldb. year: 1994. Product B is title: data mining using two-dimensional optimized association rules for numeric data : scheme , algorithms ,. authors: t fukuda , y morimoto , s morishita , t tokuyama. venue: proceedings of the acm sigmod international conference on. year: . Same Product? No\n\nProduct A is title: client-server paradise. authors: d dewitt , n kabra , j luo , j patel , j yu. venue: vldb. year: 1994. Product B is title: and j. yu . client-server paradise. authors: dj dewitt , n kabra , j luo , jm pate. venue: proceedings ofthe twentieth international conference on vey & hellip ; ,. year: . Same Product? Yes\n\nProduct A is title: mixed mode xml query processing. authors: a halverson , j burger , l galanis , a kini , r krishnamurthy , a rao , f tian , s viglas , y wang , j naughton , d dewitt. venue: vldb. year: 2003. Product B is title: a transducer-based xml query processor. authors: , p mukhopadhyay , y papakonstantinou. venue: vldb ,. year: 2002.0. Same Product? No\n\nProduct A is title: database tuning : principles , experiments , and troubleshooting techniques. authors: d shasha , p bonnet. venue: vldb. year: 2002. Product B is title: database tuning : principles , experiments , and troubleshooting techniques ( part i ). authors: d shasha , p bonnet. venue: proceedings of the 2002 acm sigmod international conference & hellip ; ,. year: 2002.0. Same Product? Yes\n\nProduct A is title: mining generalized association rules. authors: r srikant , r agrawal. venue: vldb. year: 1995. Product B is title: mining association rules from semi-structured data. authors: k maruyama , k uehara. venue: icdcs workshop of knowledge discovery and data mining in the & hellip ; ,. year: 2000.0. Same Product? No\n\nProduct A is title: the x-tree : an index structure for high-dimensional data. authors: s berchtold , d keim , h kriegel. venue: vldb. year: 1996. Product B is title: the tv-tree { an index structure for high-dimensional data . the vldb journal , 3 ( 4 ) : 517 { 549. authors: ki lin , hv jagadish , c faloutsos. venue: . year: 1994.0. Same Product? No\n\nProduct A is title: dual-buffering strategies in object bases. authors: a kemper , d kossmann. venue: vldb. year: 1994. Product B is title: dual-buffering strategies in object bases. authors: a kemper , d kossmann. venue: vldb ,. year: 1994.0. Same Product? Yes\n\nProduct A is title: banks : browsing and keyword searching in relational databases. authors: b aditya , g bhalotia , s chakrabarti , a hulgeri , c nakhe , p parag , s sudarshan. venue: vldb. year: 2002. Product B is title: distance browsing in spatial databases. authors: gr hjaltason , h samet. venue: acm transactions on database systems ,. year: 1999.0. Same Product? No\n\nProduct A is title: on-demand data elevation in hierarchical multimedia storage servers. authors: p triantafillou , t papadakis. venue: vldb. year: 1997. Product B is title: on-demand data elevation in a hierarchical multimedia storage server. authors: p triantallou , t papadakis. venue: proc . of 23rd intl. conf . on very large data bases , vldb ,. year: 1997.0. Same Product? Yes\n\nProduct A is title: a raster approximation for processing of spatial joins. authors: g zimbrao , j souza. venue: vldb. year: 1998. Product B is title: a raster approximation for processing of spatial joins. authors: g zimbrao. venue: vldb ,. year: 1998.0. Same Product? Yes\n\n",
        "vldb j.": "Product A is title: the demarcation protocol : a technique for maintaining constraints in distributed database systems. authors: d barbarс , h garcia-molina. venue: vldb j.. year: 1994. Product B is title: local verification of global integrity constraints in distributed databases. authors: a gupta , j widom. venue: . year: . Same Product? No\n\nProduct A is title: an introduction to spatial database systems. authors: r gќting. venue: vldb j.. year: 1994. Product B is title: an introduction to spatial database systems. authors: portal.acm.org. venue: . year: 1994.0. Same Product? Yes\n\nProduct A is title: re-tree : an efficient index structure for regular expressions. authors: c chan , m garofalakis , r rastogi. venue: vldb j.. year: 2003. Product B is title: the tv-tree : an index structure forhigh-dimensional data. authors: ki lin , h jagadish , c faloutsos. venue: the vldb journal ,. year: . Same Product? No\n\nProduct A is title: one-dimensional and multi-dimensional substring selectivity estimation. authors: h jagadish , o kapitskaia , r ng , d srivastava. venue: vldb j.. year: 2000. Product B is title: one-dimensional and multi-dimensional substring selectivity estimation. authors: hv jagadish , o kapitskaia , rt ng , d srivastava. venue: the vldb journal the international journal on very large & hellip ; ,. year: 2000.0. Same Product? Yes\n\nProduct A is title: active rules for xml : a new paradigm for e-services. authors: a bonifati , s ceri , s paraboschi. venue: vldb j.. year: 2001. Product B is title: active rules for xml : a new paradigm for e-services. authors: a bonifati , s ceri , s paraboschi. venue: the vldb journal the international journal on very large & hellip ; ,. year: 2001.0. Same Product? Yes\n\nProduct A is title: optimal clip ordering for multi-clip queries. authors: r ng , p shum. venue: vldb j.. year: 1998. Product B is title: optimal clip ordering for multi-clip queries. authors: rt ng , p shum. venue: the vldb journal the international journal on very large & hellip ; ,. year: 1998.0. Same Product? Yes\n\nProduct A is title: priority assignment in real-time active databases. authors: r sivasankaran , j stankovic , d towsley , b purimetla , k ramamritham. venue: vldb j.. year: 1996. Product B is title: priority assignment in real-time active databases. authors: rm sivasankaran , ja stankovic , d towsley , b. venue: the vldb journal the international journal on very large & hellip ; ,. year: 1996.0. Same Product? Yes\n\nProduct A is title: global transaction support for workflow management systems : from formal specification to practical implementation. authors: p grefen , j vonk , p apers. venue: vldb j.. year: 2001. Product B is title: distributed transaction support for workflow management applications. authors: j vonk , p grefen , e boertjes , p apers. venue: procs . 10th int . conf . on database and expert system & hellip ; ,. year: . Same Product? No\n\nProduct A is title: answering queries using views : a survey. authors: a halevy. venue: vldb j.. year: 2001. Product B is title: a scalable algorithm for answering queries using views. authors: ay levy. venue: . year: . Same Product? No\n\nProduct A is title: index nesting - an efficient approach to indexing in object-oriented databases. authors: b ooi , j han , h lu , k tan. venue: vldb j.. year: 1996. Product B is title: spatial query processing in an object-oriented database system. authors: ja orenstein. venue: proceedings of the 1986 acm sigmod international conference. year: . Same Product? No\n\n",
    },
    f"{DATASET_PATH}/entity_matching/structured/Walmart-Amazon": "Product A is modelno: c789u1mg. Product B is modelno: c644u1mg. Are A and B the Same? No, because modelno: c789u1mg and modelno: c644u1mg are not the same\n\nProduct A is modelno: hp ce278a. Product B is modelno: hp27 / ce278a ( xl ). Are A and B the Same? No, because modelno: hp ce278a and modelno: hp27 / c8727a ( xl ) are not the same\n\nProduct A is modelno: c7826an. Product B is modelno: c7826an # 56. Are A and B the Same? No, because modelno: c7826 and modelno: c7826 # 56 are not the same\n\nProduct A is modelno: oczssd22vtxe60g. Product B is modelno: oczssd2-3vtx60g. Are A and B the Same? No, because modelno: va1932 and modelno: va705b are not the same\n\nProduct A is modelno: ktd-dm6400a / 1g. Product B is modelno: ktd-dm6400bc / 1g. Are A and B the Same? No, because modelno: ktd-dm8400a / 1g and modelno: ktd-dm8400bc / 1g are not the same\n\nProduct A is title: sharp electronics xldh259n 160w micro system with ipod dock black. modelno: xldh259n. Product B is title: sharp electronics xldh259n 160w micro system with docking slot for ipod black. modelno: xldh259n. Are A and B the Same? Yes, because modelno: xldh259n and modelno: xldh259 are the same\n\nProduct A is title: philips vibe 4gb mp3 video player. modelno: sa2vbe04kc/17. Product B is title: philips gogear vibe 8gb mp3 player sa3vbe08k 37. modelno: sa3vbe08k/37. Are A and B the Same? No, because modelno: sa2vbe04kc/17 and  modelno: sa3vbe08k/37 are not the same\n\nProduct A is title: innovera heavyweight photo paper matte 8-1 2 x 11 50 sheets pack. modelno: 99650. Product B title: innovera 99650 - heavyweight photo paper matte 8-1 2 x 11 50 sheets pack. modelno: . Are A and B the Same? Yes, because modelno: 99650 and Product B title: innovera 99650 are the same",
    f"{DATASET_PATH}/entity_matching/structured/DBLP-ACM": {
        "vldb": "Product A is title: mapinfo spatialware : a spatial information server for rdbms. authors: chebel mina. venue: vldb. year: 1998. Product B is title: mapinfo spatialware : a spatial information server for rdbms. authors: chebel mina. venue: very large data bases. year: 1998. Are Product A and Product B the same? Yes\n\nProduct A is title: sampling large databases for association rules. authors: hannu toivonen. venue: vldb. year: 1996. Product B is title: sampling large databases for association rules. authors: hannu toivonen. venue: very large data bases. year: 1996. Are Product A and Product B the same? Yes\n\nProduct A is title: an ultra highly available dbms. authors: svein-olaf hvasshovd , svein erik bratsberg , øystein torbjørnsen. venue: vldb. year: 2000. Product B is title: an ultra highly available dbms. authors: svein-olaf hvasshovd , svein erik bratsberg , &#216; ystein torbj &#248; rnsen. venue: very large data bases. year: 2000. Are Product A and Product B the same? Yes\n\nProduct A is title: e.piphany epicenter technology overview. authors: sridhar ramaswamy. venue: vldb. year: 2000. Product B is title: e.piphany epicenter technology overview. authors: sridhar ramaswamy. venue: very large data bases. year: 2000. Are Product A and Product B the same? Yes\n\nProduct A is title: on the costs of multilingualism in database systems. authors: a. kumaran , jayant r. haritsa. venue: vldb. year: 2003. Product B is title: analysis of locking behavior in three real database systems. authors: vigyan singhal , alan jay smith. venue: the vldb journal -- the international journal on very large data bases. year: 1997. Are Product A and Product B the same? No\n\nProduct A is title: dynamic multi-resource load balancing in parallel database systems. authors: robert marek , erhard rahm. venue: vldb. year: 1995. Product B is title: parallel database systems 101. authors: jim gray. venue: international conference on management of data. year: 1995. Are Product A and Product B the same? No\n\nProduct A is title: processing sliding window multi-joins in continuous queries over data streams. authors: lukasz golab , m. tamer özsu. venue: vldb. year: 2003. Product B is title: processing set expressions over continuous update streams. authors: sumit ganguly , minos garofalakis , rajeev rastogi. venue: international conference on management of data. year: 2003. Are Product A and Product B the same? No\n\nProduct A is title: odefs : a file system interface to an object-oriented database. authors: narain h. gehani , william d. roome , h. v. jagadish. venue: vldb. year: 1994. Product B is title: unisql/x unified relational and object-oriented database system. authors: won kim. venue: the vldb journal -- the international journal on very large data bases. year: 1994. Are Product A and Product B the same? No\n\nProduct A is title: dwms : data warehouse management system. authors: narendra mohan. venue: vldb. year: 1996. Product B is title: strudel : a web site management system. authors: mary fernandez , daniela florescu , jaewoo kang , alon levy , dan suciu. venue: the vldb journal -- the international journal on very large data bases. year: 1997. Are Product A and Product B the same? No\n\nProduct A is title: sampling large databases for association rules. authors: hannu toivonen. venue: vldb. year: 1996. Product A is title: sampling large databases for association rules. authors: hannu toivonen. venue: the vldb journal -- the international journal on very large data bases. year: 1996. Are Product A and Product B the same? No\n",
        "sigmod record": "Product A is title: automata theory for xml researchers. authors: frank neven. venue: sigmod record. year: 2002. Product B is title: automata theory for xml researchers. authors: frank neven. venue: acm sigmod record. year: 2002. Are Product A and Product B the same? Yes\n\nProduct A is title: reasoning on regular path queries. authors: giuseppe de giacomo , moshe y. vardi , maurizio lenzerini , diego calvanese. venue: sigmod record. year: 2003. Product B is title: reasoning on regular path queries. authors: d. calvanese , g. de giacomo , m. lenzerini , m. y. vardi. venue: acm sigmod record. year: 2003. Are Product A and Product B the same? Yes\n\nProduct A is title: 3d geographic network displays. authors: taosong he , stephen g. eick , kenneth c. cox. venue: sigmod record. year: 1996. Product B is title: 3d geographic network displays. authors: kenneth c. cox , stephen g. eick , taosong he. venue: acm sigmod record. year: 1996. Are Product A and Product B the same? Yes\n\nProduct A is title: optimizing jan jannink 's implementation of b + - tree deletion. authors: h. olivie , r. maelbrancke. venue: sigmod record. year: 1995. Product B is title: optimizing jan jannink 's implementation of b + - tree deletion. authors: r. maelbrancke , h. olivi &#233;. venue: acm sigmod record. year: 1995. Are Product A and Product B the same? Yes\n\nProduct A is title: automata theory for xml researchers. authors: frank neven. venue: sigmod record. year: 2002. Product B is title: automata theory for xml researchers. authors: frank neven. venue: very large databases year: 2002. Are Product A and Product B the same? No\n\nProduct A is title: book review column. authors: karl aberer. venue: sigmod record. year: 2002. Product B is title: book review column. authors: karl aberer. venue: acm sigmod record. year: 2002. Are Product A and Product B the same? Yes\n\nProduct A is title: book review column. authors: karl aberer. venue: sigmod record. year: 2002. Product B is title: book reviews. authors: karl aberer. venue: acm sigmod record. year: 2002. Are Product A and Product B the same? No\n\nProduct A is title: book review column. authors: karl aberer. venue: sigmod record. year: 2003. Product B is title: book review column. authors: karl aberer. venue: acm sigmod record. year: 2003. Are Product A and Product B the same? Yes\n\nProduct A is title: book review column. authors: karl aberer. venue: sigmod record. year: 2003. Product B is title: book reviews. authors: karl aberer. venue: acm sigmod record. year: 2003. Are Product A and Product B the same? No\n",
        "vldb j.": "Product A is title: an architecture to support scalable online personalization on the web. authors: kaushik dutta , debra e. vandermeer , krithi ramamritham , anindya datta , shamkant b. navathe. venue: vldb j.. year: 2001. Product B is title: an architecture to support scalable online personalization on the web. authors: anindya datta , kaushik dutta , debra vandermeer , krithi ramamritham , shamkant b. navathe. venue: the vldb journal -- the international journal on very large data bases. year: 2001. Are Product A and Product B the same? Yes\n\nProduct A is title: answering queries using views : a survey. authors: alon y. halevy. venue: vldb j.. year: 2001. Product B is title: answering queries using views : a survey. authors: alon y. halevy. venue: the vldb journal -- the international journal on very large data bases. year: 2001. Are Product A and Product B the same? Yes\n\nProduct A is title: efficient schemes for managing multiversionxml documents. authors: shu-yao chien , carlo zaniolo , vassilis j. tsotras. venue: vldb j.. year: 2002. Product B is title: efficient schemes for managing multiversionxml documents. authors: s.-y . chien , v. j. tsotras , c. zaniolo. venue: the vldb journal -- the international journal on very large data bases. year: 2002. Are Product A and Product B the same? Yes\n\nProduct A is title: index configuration in object-oriented databases. authors: elisa bertino. venue: vldb j.. year: 1994. Product B is title: a cost model for clustered object-oriented databases. authors: georges gardarin , jean-robert gruser , zhao-hui tang. venue: very large data bases. year: 1995. Are Product A and Product B the same? No\n\nProduct A is title: incremental computation and maintenance of temporal aggregates. authors: jun yang , jennifer widom. venue: vldb j.. year: 2003. Product B is title: incremental maintenance of recursive views using relational calculus/sql. authors: guozhu dong , jianwen su. venue: acm sigmod record. year: 2000. Are Product A and Product B the same? No\n\nProduct A is title: index nesting - an efficient approach to indexing in object-oriented databases. authors: jiawei han , beng chin ooi , hongjun lu , kian-lee tan. venue: vldb j.. year: 1996. Product B is title: converting relational to object-oriented databases. authors: joseph fong. venue: acm sigmod record. year: 1997. Are Product A and Product B the same? No\n\nProduct A is title: concurrency control in hierarchical multidatabase systems. authors: henry f. korth , abraham silberschatz , sharad mehrotra. venue: vldb j.. year: 1997. Product B is title: dynamic load balancing in hierarchical parallel database systems. authors: luc bouganim , daniela florescu , patrick valduriez. venue: very large data bases. year: 1996. Are Product A and Product B the same? No\n\nProduct A is title: instance-based attribute identification in database integration. authors: roger h. l. chiang , ee-peng lim , chua eng huang cecil. venue: vldb j.. year: 2003. Product B is title: a case-based approach to information integration. authors: maurizio panti , luca spalazzi , alberto giretti. venue: very large data bases. year: 2000. Are Product A and Product B the same? No\n\nProduct A is title: answering queries using views : a survey. authors: alon y. halevy. venue: vldb j.. year: 2001. Product A is title: answering queries using views : a survey. authors: alon y. halevy. venue: international conference on management of data year: 2001.. Are Product A and Product B the same? No\n",
        "acm trans . database syst .": "Product A is title: tail recursion elimination in deductive databases. authors: kenneth a. ross. venue: acm trans . database syst .. year: 1996. Product B is title: tail recursion elimination in deductive databases. authors: kenneth a. ross. venue: acm transactions on database systems ( tods ). year: 1996. Are Product A and Product B the same? Yes\n\nProduct A is title: disjunctive datalog. authors: heikki mannila , thomas eiter , georg gottlob. venue: acm trans . database syst .. year: 1997. Product B is title: disjunctive datalog. authors: thomas eiter , georg gottlob , heikki mannila. venue: acm transactions on database systems ( tods ). year: 1997. Are Product A and Product B the same? Yes\n\nProduct A is title: space optimization in deductive databases. authors: s. sudarshan , divesh srivastava , raghu ramakrishnan , jeffrey f. naughton. venue: acm trans . database syst .. year: 1995. Product B is title: introduction to constraint databases. authors: bart kuijpers. venue: acm sigmod record. year: 2002. Are Product A and Product B the same? No\n\nProduct A is title: solving satisfiability and implication problems in database systems. authors: sha guo , mark allen weiss , wei sun. venue: acm trans . database syst .. year: 1996. Product B is title: temporal database system implementations. authors: michael h. b &#246; hlen. venue: acm sigmod record. year: 1995. Are Product A and Product B the same? No\n\nProduct A is title: disjunctive datalog. authors: heikki mannila , thomas eiter , georg gottlob. venue: acm trans . database syst .. year: 1997. Product B is title: disjunctive datalog. authors: heikki mannila , thomas eiter , georg gottlob. venue: acm sigmod record. year: 1997. Are Product A and Product B the same? No\n",
        "sigmod conference": "Product A is title: javax.xxl : a prototype for a library of query processing algorithms. authors: bernhard seeger , jens-peter dittrich , jochen van den bercken. venue: sigmod conference. year: 2000. Product B is title: javax.xxl : a prototype for a library of query processing algorithms. authors: jochen van den bercken , jens-peter dittrich , bernhard seeger. venue: international conference on management of data. year: 2000. Are Product A and Product B the same? Yes\n\nProduct A is title: orthogonal optimization of subqueries and aggregation. authors: milind joshi , césar a. galindo-legaria. venue: sigmod conference. year: 2001. Product B is title: orthogonal optimization of subqueries and aggregation. authors: c &#233; sar galindo-legaria , milind joshi. venue: international conference on management of data. year: 2001. Are Product A and Product B the same? Yes\n\nProduct A is title: from structured documents to novel query facilities. authors: sophie cluet , michel scholl , serge abiteboul , vassilis christophides. venue: sigmod conference. year: 1994. Product B is title: from structured documents to novel query facilities. authors: v. christophides , s. abiteboul , s. cluet , m. scholl. venue: international conference on management of data. year: 1994. Are Product A and Product B the same? Yes\n\nProduct A is title: the naos system. authors: christine collet , thierry coupaye. venue: sigmod conference. year: 1995. Product B is title: the naos system. authors: c. collet , t. coupaye. venue: international conference on management of data. year: 1995. Are Product A and Product B the same? Yes\n\nProduct A is title: highly concurrent cache consistency for indices in client-server database systems. authors: michael j. carey , markos zaharioudakis. venue: sigmod conference. year: 1997. Product B is title: index concurrency control in firm real-time database systems. authors: brajesh goyal , jayant r. haritsa , s. seshadri , v. srinivasan. venue: very large data bases. year: 1995. Are Product A and Product B the same? No\n\nProduct A is title: selectivity estimation in spatial databases. authors: viswanath poosala , sridhar ramaswamy , swarup acharya. venue: sigmod conference. year: 1999. Product B is title: selectivity estimation for spatio-temporal queries to moving objects. authors: yong-jin choi , chin-wan chung. venue: international conference on management of data. year: 2002. Are Product A and Product B the same? No\n\nProduct A is title: the need for distributed asynchronous transactions. authors: lyman do , prabhu ram , pamela drew. venue: sigmod conference. year: 1999. Product B is title: atomicity versus anonymity : distributed transactions for electronic commerce. authors: j. d. tygar. venue: very large data bases. year: 1998. Are Product A and Product B the same? No, because Product A venue: sigmod conference and Product B venue: very large data bases are not the same\n\nProduct A is title: query optimization in compressed database systems. authors: zhiyuan chen , flip korn , johannes gehrke. venue: sigmod conference. year: 2001. Product B is title: query caching and optimization in distributed mediator systems. authors: s. adali , k. s. candan , y. papakonstantinou , v. s. subrahmanian. venue: the vldb journal -- the international journal on very large data bases. year: 1996. Are Product A and Product B the same? No\n\nProduct A is title: orthogonal optimization of subqueries and aggregation. authors: milind joshi , césar a. galindo-legaria. venue: sigmod conference. year: 2001. Product B is title: orthogonal optimization of subqueries and aggregation. authors: milind joshi , césar a. galindo-legaria. venue: very large databases. year: 2001. Are Product A and Product B the same? No\n",
    },
    f"{DATASET_PATH}/schema_matching/Synthea": {
        "person-person_id": "A is person-person_id. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n\nA is person-person_id. B is patients-suffix. Are A and B the same? Yes or No? No\n\nA is person-person_id. B is patients-zip. Are A and B the same? Yes or No? No\n",
        "person-gender_concept_id": "A is person-gender_concept_id. B is patients-gender. Are A and B the same? Yes or No? Yes\n\nA is person-gender_concept_id. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is person-gender_concept_id. B is organizations-utilization. Are A and B the same? Yes or No? No\n",
        "person-year_of_birth": "A is person-year_of_birth. B is patients-birthdate. Are A and B the same? Yes or No? Yes\n\nA is person-year_of_birth. B is immunizations-cost. Are A and B the same? Yes or No? No\n\nA is person-year_of_birth. B is procedures-date. Are A and B the same? Yes or No? No\n",
        "person-month_of_birth": "A is person-month_of_birth. B is patients-birthdate. Are A and B the same? Yes or No? Yes\n\nA is person-month_of_birth. B is careplans-encounter. Are A and B the same? Yes or No? No\n\nA is person-month_of_birth. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n",
        "person-day_of_birth": "A is person-day_of_birth. B is patients-birthdate. Are A and B the same? Yes or No? Yes\n\nA is person-day_of_birth. B is careplans-patient. Are A and B the same? Yes or No? No\n\nA is person-day_of_birth. B is observations-encounter. Are A and B the same? Yes or No? No\n",
        "person-birth_datetime": "A is person-birth_datetime. B is patients-birthdate. Are A and B the same? Yes or No? Yes\n\nA is person-birth_datetime. B is providers-address. Are A and B the same? Yes or No? No\n\nA is person-birth_datetime. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n",
        "person-death_datetime": "A is person-death_datetime. B is patients-deathdate. Are A and B the same? Yes or No? Yes\n\nA is person-death_datetime. B is patients-id. Are A and B the same? Yes or No? No\n\nA is person-death_datetime. B is encounters-reasoncode. Are A and B the same? Yes or No? No\n",
        "person-race_concept_id": "A is person-race_concept_id. B is patients-race. Are A and B the same? Yes or No? Yes\n\nA is person-race_concept_id. B is imaging_studies-encounter. Are A and B the same? Yes or No? No\n\nA is person-race_concept_id. B is imaging_studies-id. Are A and B the same? Yes or No? No\n",
        "person-ethnicity_concept_id": "A is person-ethnicity_concept_id. B is patients-ethnicity. Are A and B the same? Yes or No? Yes\n\nA is person-ethnicity_concept_id. B is encounters-cost. Are A and B the same? Yes or No? No\n\nA is person-ethnicity_concept_id. B is patients-state. Are A and B the same? Yes or No? No\n",
        "person-location_id": "A is person-location_id. B is imaging_studies-id. Are A and B the same? Yes or No? No\n\nA is person-location_id. B is encounters-stop. Are A and B the same? Yes or No? No\n\nA is person-location_id. B is imaging_studies-date. Are A and B the same? Yes or No? No\n",
        "person-provider_id": "A is person-provider_id. B is providers-speciality. Are A and B the same? Yes or No? No\n\nA is person-provider_id. B is procedures-description. Are A and B the same? Yes or No? No\n\nA is person-provider_id. B is patients-race. Are A and B the same? Yes or No? No\n",
        "person-care_site_id": "A is person-care_site_id. B is patients-ssn. Are A and B the same? Yes or No? No\n\nA is person-care_site_id. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n\nA is person-care_site_id. B is allergies-description. Are A and B the same? Yes or No? No\n",
        "person-person_source_value": "A is person-person_source_value. B is patients-id. Are A and B the same? Yes or No? Yes\n\nA is person-person_source_value. B is organizations-phone. Are A and B the same? Yes or No? No\n\nA is person-person_source_value. B is encounters-reasondescription. Are A and B the same? Yes or No? No\n",
        "person-gender_source_value": "A is person-gender_source_value. B is patients-gender. Are A and B the same? Yes or No? Yes\n\nA is person-gender_source_value. B is immunizations-description. Are A and B the same? Yes or No? No\n\nA is person-gender_source_value. B is providers-gender. Are A and B the same? Yes or No? No\n",
        "person-gender_source_concept_id": "A is person-gender_source_concept_id. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is person-gender_source_concept_id. B is procedures-date. Are A and B the same? Yes or No? No\n\nA is person-gender_source_concept_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n",
        "person-race_source_value": "A is person-race_source_value. B is patients-race. Are A and B the same? Yes or No? Yes\n\nA is person-race_source_value. B is observations-type. Are A and B the same? Yes or No? No\n\nA is person-race_source_value. B is patients-id. Are A and B the same? Yes or No? No\n",
        "person-race_source_concept_id": "A is person-race_source_concept_id. B is imaging_studies-sop description. Are A and B the same? Yes or No? No\n\nA is person-race_source_concept_id. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is person-race_source_concept_id. B is observations-code. Are A and B the same? Yes or No? No\n",
        "person-ethnicity_source_value": "A is person-ethnicity_source_value. B is patients-ethnicity. Are A and B the same? Yes or No? Yes\n\nA is person-ethnicity_source_value. B is observations-code. Are A and B the same? Yes or No? No\n\nA is person-ethnicity_source_value. B is allergies-stop. Are A and B the same? Yes or No? No\n",
        "person-ethnicity_source_concept_id": "A is person-ethnicity_source_concept_id. B is patients-state. Are A and B the same? Yes or No? No\n\nA is person-ethnicity_source_concept_id. B is careplans-id. Are A and B the same? Yes or No? No\n\nA is person-ethnicity_source_concept_id. B is careplans-patient. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_occurrence_id": "A is visit_occurrence-visit_occurrence_id. B is careplans-encounter. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_occurrence_id. B is encounters-provider. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_occurrence_id. B is encounters-code. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-person_id": "A is visit_occurrence-person_id. B is encounters-patient. Are A and B the same? Yes or No? Yes\n\nA is visit_occurrence-person_id. B is imaging_studies-modality code. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-person_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_concept_id": "A is visit_occurrence-visit_concept_id. B is encounters-encounterclass. Are A and B the same? Yes or No? Yes\n\nA is visit_occurrence-visit_concept_id. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_concept_id. B is allergies-description. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_start_date": "A is visit_occurrence-visit_start_date. B is encounters-start. Are A and B the same? Yes or No? Yes\n\nA is visit_occurrence-visit_start_date. B is observations-code. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_start_date. B is patients-id. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_start_datetime": "A is visit_occurrence-visit_start_datetime. B is encounters-start. Are A and B the same? Yes or No? Yes\n\nA is visit_occurrence-visit_start_datetime. B is encounters-code. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_start_datetime. B is procedures-encounter. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_end_date": "A is visit_occurrence-visit_end_date. B is encounters-stop. Are A and B the same? Yes or No? Yes\n\nA is visit_occurrence-visit_end_date. B is organizations-id. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_end_date. B is patients-drivers. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_end_datetime": "A is visit_occurrence-visit_end_datetime. B is encounters-stop. Are A and B the same? Yes or No? Yes\n\nA is visit_occurrence-visit_end_datetime. B is encounters-provider. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_end_datetime. B is organizations-name. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_type_concept_id": "A is visit_occurrence-visit_type_concept_id. B is providers-address. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_type_concept_id. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_type_concept_id. B is careplans-reasoncode. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-provider_id": "A is visit_occurrence-provider_id. B is medications-reasoncode. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-provider_id. B is patients-zip. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-provider_id. B is medications-start. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-care_site_id": "A is visit_occurrence-care_site_id. B is imaging_studies-modality description. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-care_site_id. B is careplans-stop. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-care_site_id. B is providers-name. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_source_value": "A is visit_occurrence-visit_source_value. B is encounters-encounterclass. Are A and B the same? Yes or No? Yes\n\nA is visit_occurrence-visit_source_value. B is organizations-phone. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_source_value. B is conditions-patient. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-visit_source_concept_id": "A is visit_occurrence-visit_source_concept_id. B is observations-value. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_source_concept_id. B is careplans-encounter. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-visit_source_concept_id. B is providers-state. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-admitted_from_concept_id": "A is visit_occurrence-admitted_from_concept_id. B is careplans-start. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-admitted_from_concept_id. B is providers-state. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-admitted_from_concept_id. B is patients-address. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-admitted_from_source_value": "A is visit_occurrence-admitted_from_source_value. B is providers-city. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-admitted_from_source_value. B is immunizations-description. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-admitted_from_source_value. B is conditions-patient. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-discharge_to_concept_id": "A is visit_occurrence-discharge_to_concept_id. B is allergies-start. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-discharge_to_concept_id. B is allergies-code. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-discharge_to_concept_id. B is patients-state. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-discharge_to_source_value": "A is visit_occurrence-discharge_to_source_value. B is medications-start. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-discharge_to_source_value. B is providers-address. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-discharge_to_source_value. B is allergies-stop. Are A and B the same? Yes or No? No\n",
        "visit_occurrence-preceding_visit_occurrence_id": "A is visit_occurrence-preceding_visit_occurrence_id. B is patients-marital. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-preceding_visit_occurrence_id. B is careplans-patient. Are A and B the same? Yes or No? No\n\nA is visit_occurrence-preceding_visit_occurrence_id. B is providers-address. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_id": "A is visit_detail-visit_detail_id. B is providers-city. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_id. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_id. B is medications-description. Are A and B the same? Yes or No? No\n",
        "visit_detail-person_id": "A is visit_detail-person_id. B is organizations-city. Are A and B the same? Yes or No? No\n\nA is visit_detail-person_id. B is procedures-date. Are A and B the same? Yes or No? No\n\nA is visit_detail-person_id. B is organizations-state. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_concept_id": "A is visit_detail-visit_detail_concept_id. B is immunizations-code. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_concept_id. B is encounters-provider. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_concept_id. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_start_date": "A is visit_detail-visit_detail_start_date. B is conditions-description. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_start_date. B is imaging_studies-modality code. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_start_date. B is patients-id. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_start_datetime": "A is visit_detail-visit_detail_start_datetime. B is medications-encounter. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_start_datetime. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_start_datetime. B is encounters-provider. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_end_date": "A is visit_detail-visit_detail_end_date. B is procedures-encounter. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_end_date. B is medications-reasondescription. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_end_date. B is patients-gender. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_end_datetime": "A is visit_detail-visit_detail_end_datetime. B is encounters-start. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_end_datetime. B is patients-state. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_end_datetime. B is patients-passport. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_type_concept_id": "A is visit_detail-visit_detail_type_concept_id. B is immunizations-description. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_type_concept_id. B is organizations-name. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_type_concept_id. B is medications-reasondescription. Are A and B the same? Yes or No? No\n",
        "visit_detail-provider_id": "A is visit_detail-provider_id. B is patients-first. Are A and B the same? Yes or No? No\n\nA is visit_detail-provider_id. B is organizations-address. Are A and B the same? Yes or No? No\n\nA is visit_detail-provider_id. B is imaging_studies-sop description. Are A and B the same? Yes or No? No\n",
        "visit_detail-care_site_id": "A is visit_detail-care_site_id. B is providers-name. Are A and B the same? Yes or No? No\n\nA is visit_detail-care_site_id. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is visit_detail-care_site_id. B is providers-organization. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_source_value": "A is visit_detail-visit_detail_source_value. B is immunizations-patient. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_source_value. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_source_value. B is providers-address. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_source_concept_id": "A is visit_detail-visit_detail_source_concept_id. B is encounters-stop. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_source_concept_id. B is patients-race. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_source_concept_id. B is medications-description. Are A and B the same? Yes or No? No\n",
        "visit_detail-admitted_from_source_value": "A is visit_detail-admitted_from_source_value. B is organizations-city. Are A and B the same? Yes or No? No\n\nA is visit_detail-admitted_from_source_value. B is observations-date. Are A and B the same? Yes or No? No\n\nA is visit_detail-admitted_from_source_value. B is procedures-patient. Are A and B the same? Yes or No? No\n",
        "visit_detail-admitted_from_concept_id": "A is visit_detail-admitted_from_concept_id. B is allergies-stop. Are A and B the same? Yes or No? No\n\nA is visit_detail-admitted_from_concept_id. B is medications-code. Are A and B the same? Yes or No? No\n\nA is visit_detail-admitted_from_concept_id. B is procedures-date. Are A and B the same? Yes or No? No\n",
        "visit_detail-discharge_to_source_value": "A is visit_detail-discharge_to_source_value. B is encounters-id. Are A and B the same? Yes or No? No\n\nA is visit_detail-discharge_to_source_value. B is providers-organization. Are A and B the same? Yes or No? No\n\nA is visit_detail-discharge_to_source_value. B is encounters-provider. Are A and B the same? Yes or No? No\n",
        "visit_detail-discharge_to_concept_id": "A is visit_detail-discharge_to_concept_id. B is patients-id. Are A and B the same? Yes or No? No\n\nA is visit_detail-discharge_to_concept_id. B is immunizations-code. Are A and B the same? Yes or No? No\n\nA is visit_detail-discharge_to_concept_id. B is immunizations-cost. Are A and B the same? Yes or No? No\n",
        "visit_detail-preceding_visit_detail_id": "A is visit_detail-preceding_visit_detail_id. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is visit_detail-preceding_visit_detail_id. B is allergies-patient. Are A and B the same? Yes or No? No\n\nA is visit_detail-preceding_visit_detail_id. B is observations-units. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_detail_parent_id": "A is visit_detail-visit_detail_parent_id. B is allergies-stop. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_parent_id. B is allergies-code. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_detail_parent_id. B is encounters-start. Are A and B the same? Yes or No? No\n",
        "visit_detail-visit_occurrence_id": "A is visit_detail-visit_occurrence_id. B is providers-utilization. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_occurrence_id. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n\nA is visit_detail-visit_occurrence_id. B is patients-passport. Are A and B the same? Yes or No? No\n",
        "specimen-specimen_id": "A is specimen-specimen_id. B is imaging_studies-body site code. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_id. B is providers-name. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_id. B is imaging_studies-modality description. Are A and B the same? Yes or No? No\n",
        "specimen-person_id": "A is specimen-person_id. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is specimen-person_id. B is observations-type. Are A and B the same? Yes or No? No\n\nA is specimen-person_id. B is conditions-description. Are A and B the same? Yes or No? No\n",
        "specimen-specimen_concept_id": "A is specimen-specimen_concept_id. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_concept_id. B is encounters-reasondescription. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_concept_id. B is providers-name. Are A and B the same? Yes or No? No\n",
        "specimen-specimen_type_concept_id": "A is specimen-specimen_type_concept_id. B is immunizations-description. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_type_concept_id. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_type_concept_id. B is observations-code. Are A and B the same? Yes or No? No\n",
        "specimen-specimen_date": "A is specimen-specimen_date. B is careplans-description. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_date. B is patients-state. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_date. B is organizations-state. Are A and B the same? Yes or No? No\n",
        "specimen-specimen_datetime": "A is specimen-specimen_datetime. B is organizations-zip. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_datetime. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_datetime. B is providers-id. Are A and B the same? Yes or No? No\n",
        "specimen-quantity": "A is specimen-quantity. B is patients-first. Are A and B the same? Yes or No? No\n\nA is specimen-quantity. B is careplans-id. Are A and B the same? Yes or No? No\n\nA is specimen-quantity. B is encounters-reasondescription. Are A and B the same? Yes or No? No\n",
        "specimen-unit_concept_id": "A is specimen-unit_concept_id. B is imaging_studies-body site description. Are A and B the same? Yes or No? No\n\nA is specimen-unit_concept_id. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is specimen-unit_concept_id. B is patients-state. Are A and B the same? Yes or No? No\n",
        "specimen-anatomic_site_concept_id": "A is specimen-anatomic_site_concept_id. B is observations-patient. Are A and B the same? Yes or No? No\n\nA is specimen-anatomic_site_concept_id. B is medications-description. Are A and B the same? Yes or No? No\n\nA is specimen-anatomic_site_concept_id. B is organizations-address. Are A and B the same? Yes or No? No\n",
        "specimen-disease_status_concept_id": "A is specimen-disease_status_concept_id. B is organizations-utilization. Are A and B the same? Yes or No? No\n\nA is specimen-disease_status_concept_id. B is observations-code. Are A and B the same? Yes or No? No\n\nA is specimen-disease_status_concept_id. B is encounters-description. Are A and B the same? Yes or No? No\n",
        "specimen-specimen_source_id": "A is specimen-specimen_source_id. B is immunizations-encounter. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_source_id. B is observations-date. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_source_id. B is imaging_studies-modality description. Are A and B the same? Yes or No? No\n",
        "specimen-specimen_source_value": "A is specimen-specimen_source_value. B is patients-suffix. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_source_value. B is imaging_studies-date. Are A and B the same? Yes or No? No\n\nA is specimen-specimen_source_value. B is organizations-id. Are A and B the same? Yes or No? No\n",
        "specimen-unit_source_value": "A is specimen-unit_source_value. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n\nA is specimen-unit_source_value. B is organizations-name. Are A and B the same? Yes or No? No\n\nA is specimen-unit_source_value. B is imaging_studies-body site code. Are A and B the same? Yes or No? No\n",
        "specimen-anatomic_site_source_value": "A is specimen-anatomic_site_source_value. B is encounters-patient. Are A and B the same? Yes or No? No\n\nA is specimen-anatomic_site_source_value. B is providers-state. Are A and B the same? Yes or No? No\n\nA is specimen-anatomic_site_source_value. B is providers-city. Are A and B the same? Yes or No? No\n",
        "specimen-disease_status_source_value": "A is specimen-disease_status_source_value. B is procedures-date. Are A and B the same? Yes or No? No\n\nA is specimen-disease_status_source_value. B is providers-gender. Are A and B the same? Yes or No? No\n\nA is specimen-disease_status_source_value. B is careplans-description. Are A and B the same? Yes or No? No\n",
        "measurement-measurement_id": "A is measurement-measurement_id. B is careplans-id. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_id. B is observations-encounter. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_id. B is careplans-encounter. Are A and B the same? Yes or No? No\n",
        "measurement-person_id": "A is measurement-person_id. B is observations-patient. Are A and B the same? Yes or No? Yes\n\nA is measurement-person_id. B is careplans-reasoncode. Are A and B the same? Yes or No? No\n\nA is measurement-person_id. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n",
        "measurement-measurement_concept_id": "A is measurement-measurement_concept_id. B is procedures-code. Are A and B the same? Yes or No? Yes\n\nA is measurement-measurement_concept_id. B is patients-address. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_concept_id. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n",
        "measurement-measurement_date": "A is measurement-measurement_date. B is observations-date. Are A and B the same? Yes or No? Yes\n\nA is measurement-measurement_date. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_date. B is allergies-description. Are A and B the same? Yes or No? No\n",
        "measurement-measurement_datetime": "A is measurement-measurement_datetime. B is procedures-date. Are A and B the same? Yes or No? Yes\n\nA is measurement-measurement_datetime. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_datetime. B is encounters-start. Are A and B the same? Yes or No? No\n",
        "measurement-measurement_time": "A is measurement-measurement_time. B is procedures-date. Are A and B the same? Yes or No? Yes\n\nA is measurement-measurement_time. B is medications-patient. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_time. B is observations-type. Are A and B the same? Yes or No? No\n",
        "measurement-measurement_type_concept_id": "A is measurement-measurement_type_concept_id. B is medications-encounter. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_type_concept_id. B is imaging_studies-date. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_type_concept_id. B is conditions-code. Are A and B the same? Yes or No? No\n",
        "measurement-operator_concept_id": "A is measurement-operator_concept_id. B is providers-state. Are A and B the same? Yes or No? No\n\nA is measurement-operator_concept_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is measurement-operator_concept_id. B is medications-reasondescription. Are A and B the same? Yes or No? No\n",
        "measurement-value_as_number": "A is measurement-value_as_number. B is observations-value. Are A and B the same? Yes or No? Yes\n\nA is measurement-value_as_number. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n\nA is measurement-value_as_number. B is procedures-description. Are A and B the same? Yes or No? No\n",
        "measurement-value_as_concept_id": "A is measurement-value_as_concept_id. B is immunizations-date. Are A and B the same? Yes or No? No\n\nA is measurement-value_as_concept_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is measurement-value_as_concept_id. B is organizations-utilization. Are A and B the same? Yes or No? No\n",
        "measurement-unit_concept_id": "A is measurement-unit_concept_id. B is observations-units. Are A and B the same? Yes or No? No\n\nA is measurement-unit_concept_id. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is measurement-unit_concept_id. B is imaging_studies-encounter. Are A and B the same? Yes or No? No\n",
        "measurement-range_low": "A is measurement-range_low. B is patients-zip. Are A and B the same? Yes or No? No\n\nA is measurement-range_low. B is organizations-utilization. Are A and B the same? Yes or No? No\n\nA is measurement-range_low. B is encounters-description. Are A and B the same? Yes or No? No\n",
        "measurement-range_high": "A is measurement-range_high. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n\nA is measurement-range_high. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is measurement-range_high. B is allergies-patient. Are A and B the same? Yes or No? No\n",
        "measurement-provider_id": "A is measurement-provider_id. B is observations-code. Are A and B the same? Yes or No? No\n\nA is measurement-provider_id. B is observations-description. Are A and B the same? Yes or No? No\n\nA is measurement-provider_id. B is providers-organization. Are A and B the same? Yes or No? No\n",
        "measurement-visit_occurrence_id": "A is measurement-visit_occurrence_id. B is observations-encounter. Are A and B the same? Yes or No? Yes\n\nA is measurement-visit_occurrence_id. B is organizations-zip. Are A and B the same? Yes or No? No\n\nA is measurement-visit_occurrence_id. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n",
        "measurement-visit_detail_id": "A is measurement-visit_detail_id. B is allergies-start. Are A and B the same? Yes or No? No\n\nA is measurement-visit_detail_id. B is imaging_studies-id. Are A and B the same? Yes or No? No\n\nA is measurement-visit_detail_id. B is organizations-utilization. Are A and B the same? Yes or No? No\n",
        "measurement-measurement_source_value": "A is measurement-measurement_source_value. B is observations-code. Are A and B the same? Yes or No? Yes\n\nA is measurement-measurement_source_value. B is patients-last. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_source_value. B is organizations-phone. Are A and B the same? Yes or No? No\n",
        "measurement-measurement_source_concept_id": "A is measurement-measurement_source_concept_id. B is procedures-code. Are A and B the same? Yes or No? Yes\n\nA is measurement-measurement_source_concept_id. B is immunizations-code. Are A and B the same? Yes or No? No\n\nA is measurement-measurement_source_concept_id. B is procedures-date. Are A and B the same? Yes or No? No\n",
        "measurement-unit_source_value": "A is measurement-unit_source_value. B is observations-units. Are A and B the same? Yes or No? Yes\n\nA is measurement-unit_source_value. B is patients-maiden. Are A and B the same? Yes or No? No\n\nA is measurement-unit_source_value. B is medications-stop. Are A and B the same? Yes or No? No\n",
        "measurement-value_source_value": "A is measurement-value_source_value. B is observations-value. Are A and B the same? Yes or No? Yes\n\nA is measurement-value_source_value. B is encounters-id. Are A and B the same? Yes or No? No\n\nA is measurement-value_source_value. B is observations-encounter. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-procedure_occurrence_id": "A is procedure_occurrence-procedure_occurrence_id. B is patients-first. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_occurrence_id. B is observations-code. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_occurrence_id. B is medications-stop. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-person_id": "A is procedure_occurrence-person_id. B is procedures-patient. Are A and B the same? Yes or No? Yes\n\nA is procedure_occurrence-person_id. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-person_id. B is conditions-start. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-procedure_concept_id": "A is procedure_occurrence-procedure_concept_id. B is procedures-code. Are A and B the same? Yes or No? Yes\n\nA is procedure_occurrence-procedure_concept_id. B is organizations-utilization. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_concept_id. B is organizations-name. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-procedure_date": "A is procedure_occurrence-procedure_date. B is procedures-date. Are A and B the same? Yes or No? Yes\n\nA is procedure_occurrence-procedure_date. B is observations-code. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_date. B is careplans-encounter. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-procedure_datetime": "A is procedure_occurrence-procedure_datetime. B is procedures-date. Are A and B the same? Yes or No? Yes\n\nA is procedure_occurrence-procedure_datetime. B is observations-type. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_datetime. B is allergies-encounter. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-procedure_type_concept_id": "A is procedure_occurrence-procedure_type_concept_id. B is providers-speciality. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_type_concept_id. B is medications-description. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_type_concept_id. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-modifier_concept_id": "A is procedure_occurrence-modifier_concept_id. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-modifier_concept_id. B is encounters-reasoncode. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-modifier_concept_id. B is patients-prefix. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-quantity": "A is procedure_occurrence-quantity. B is encounters-reasondescription. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-quantity. B is providers-state. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-quantity. B is immunizations-cost. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-provider_id": "A is procedure_occurrence-provider_id. B is allergies-description. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-provider_id. B is immunizations-patient. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-provider_id. B is patients-suffix. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-visit_occurrence_id": "A is procedure_occurrence-visit_occurrence_id. B is procedures-encounter. Are A and B the same? Yes or No? Yes\n\nA is procedure_occurrence-visit_occurrence_id. B is observations-patient. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-visit_occurrence_id. B is observations-type. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-visit_detail_id": "A is procedure_occurrence-visit_detail_id. B is medications-code. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-visit_detail_id. B is conditions-start. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-visit_detail_id. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-procedure_source_value": "A is procedure_occurrence-procedure_source_value. B is procedures-code. Are A and B the same? Yes or No? Yes\n\nA is procedure_occurrence-procedure_source_value. B is conditions-description. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_source_value. B is patients-address. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-procedure_source_concept_id": "A is procedure_occurrence-procedure_source_concept_id. B is procedures-code. Are A and B the same? Yes or No? Yes\n\nA is procedure_occurrence-procedure_source_concept_id. B is organizations-zip. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-procedure_source_concept_id. B is careplans-id. Are A and B the same? Yes or No? No\n",
        "procedure_occurrence-modifier_source_value": "A is procedure_occurrence-modifier_source_value. B is imaging_studies-body site description. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-modifier_source_value. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is procedure_occurrence-modifier_source_value. B is allergies-start. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_occurrence_id": "A is condition_occurrence-condition_occurrence_id. B is immunizations-encounter. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_occurrence_id. B is patients-maiden. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_occurrence_id. B is observations-code. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-person_id": "A is condition_occurrence-person_id. B is conditions-patient. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-person_id. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-person_id. B is imaging_studies-patient. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_concept_id": "A is condition_occurrence-condition_concept_id. B is conditions-code. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-condition_concept_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_concept_id. B is careplans-encounter. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_start_date": "A is condition_occurrence-condition_start_date. B is conditions-start. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-condition_start_date. B is providers-address. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_start_date. B is observations-date. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_start_datetime": "A is condition_occurrence-condition_start_datetime. B is conditions-start. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-condition_start_datetime. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_start_datetime. B is providers-city. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_end_date": "A is condition_occurrence-condition_end_date. B is conditions-stop. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-condition_end_date. B is imaging_studies-date. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_end_date. B is immunizations-date. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_end_datetime": "A is condition_occurrence-condition_end_datetime. B is conditions-stop. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-condition_end_datetime. B is providers-utilization. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_end_datetime. B is imaging_studies-modality description. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_type_concept_id": "A is condition_occurrence-condition_type_concept_id. B is allergies-patient. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_type_concept_id. B is medications-reasondescription. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_type_concept_id. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_status_concept_id": "A is condition_occurrence-condition_status_concept_id. B is patients-birthplace. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_status_concept_id. B is organizations-name. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_status_concept_id. B is encounters-patient. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-stop_reason": "A is condition_occurrence-stop_reason. B is organizations-zip. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-stop_reason. B is providers-address. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-stop_reason. B is conditions-start. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-provider_id": "A is condition_occurrence-provider_id. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-provider_id. B is patients-prefix. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-provider_id. B is patients-state. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-visit_occurrence_id": "A is condition_occurrence-visit_occurrence_id. B is conditions-encounter. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-visit_occurrence_id. B is imaging_studies-body site code. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-visit_occurrence_id. B is providers-gender. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-visit_detail_id": "A is condition_occurrence-visit_detail_id. B is encounters-patient. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-visit_detail_id. B is imaging_studies-id. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-visit_detail_id. B is observations-date. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_source_value": "A is condition_occurrence-condition_source_value. B is conditions-code. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-condition_source_value. B is providers-utilization. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_source_value. B is immunizations-description. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_source_concept_id": "A is condition_occurrence-condition_source_concept_id. B is conditions-code. Are A and B the same? Yes or No? Yes\n\nA is condition_occurrence-condition_source_concept_id. B is encounters-start. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_source_concept_id. B is medications-stop. Are A and B the same? Yes or No? No\n",
        "condition_occurrence-condition_status_source_value": "A is condition_occurrence-condition_status_source_value. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_status_source_value. B is patients-first. Are A and B the same? Yes or No? No\n\nA is condition_occurrence-condition_status_source_value. B is providers-state. Are A and B the same? Yes or No? No\n",
        "death-person_id": "A is death-person_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is death-person_id. B is organizations-address. Are A and B the same? Yes or No? No\n\nA is death-person_id. B is allergies-patient. Are A and B the same? Yes or No? No\n",
        "death-death_date": "A is death-death_date. B is organizations-utilization. Are A and B the same? Yes or No? No\n\nA is death-death_date. B is procedures-patient. Are A and B the same? Yes or No? No\n\nA is death-death_date. B is conditions-code. Are A and B the same? Yes or No? No\n",
        "death-death_datetime": "A is death-death_datetime. B is organizations-state. Are A and B the same? Yes or No? No\n\nA is death-death_datetime. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n\nA is death-death_datetime. B is patients-drivers. Are A and B the same? Yes or No? No\n",
        "death-death_type_concept_id": "A is death-death_type_concept_id. B is patients-id. Are A and B the same? Yes or No? No\n\nA is death-death_type_concept_id. B is procedures-encounter. Are A and B the same? Yes or No? No\n\nA is death-death_type_concept_id. B is imaging_studies-date. Are A and B the same? Yes or No? No\n",
        "death-cause_concept_id": "A is death-cause_concept_id. B is conditions-encounter. Are A and B the same? Yes or No? No\n\nA is death-cause_concept_id. B is imaging_studies-body site description. Are A and B the same? Yes or No? No\n\nA is death-cause_concept_id. B is immunizations-cost. Are A and B the same? Yes or No? No\n",
        "death-cause_source_value": "A is death-cause_source_value. B is patients-race. Are A and B the same? Yes or No? No\n\nA is death-cause_source_value. B is procedures-reasondescription. Are A and B the same? Yes or No? No\n\nA is death-cause_source_value. B is allergies-code. Are A and B the same? Yes or No? No\n",
        "death-cause_source_concept_id": "A is death-cause_source_concept_id. B is encounters-reasoncode. Are A and B the same? Yes or No? No\n\nA is death-cause_source_concept_id. B is careplans-encounter. Are A and B the same? Yes or No? No\n\nA is death-cause_source_concept_id. B is providers-gender. Are A and B the same? Yes or No? No\n",
        "location-location_id": "A is location-location_id. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is location-location_id. B is encounters-start. Are A and B the same? Yes or No? No\n\nA is location-location_id. B is observations-type. Are A and B the same? Yes or No? No\n",
        "location-address_1": "A is location-address_1. B is allergies-code. Are A and B the same? Yes or No? No\n\nA is location-address_1. B is organizations-name. Are A and B the same? Yes or No? No\n\nA is location-address_1. B is imaging_studies-modality code. Are A and B the same? Yes or No? No\n",
        "location-address_2": "A is location-address_2. B is imaging_studies-body site code. Are A and B the same? Yes or No? No\n\nA is location-address_2. B is conditions-code. Are A and B the same? Yes or No? No\n\nA is location-address_2. B is providers-gender. Are A and B the same? Yes or No? No\n",
        "location-city": "A is location-city. B is patients-ssn. Are A and B the same? Yes or No? No\n\nA is location-city. B is patients-gender. Are A and B the same? Yes or No? No\n\nA is location-city. B is imaging_studies-sop description. Are A and B the same? Yes or No? No\n",
        "location-state": "A is location-state. B is encounters-provider. Are A and B the same? Yes or No? No\n\nA is location-state. B is providers-zip. Are A and B the same? Yes or No? No\n\nA is location-state. B is patients-zip. Are A and B the same? Yes or No? No\n",
        "location-zip": "A is location-zip. B is careplans-reasoncode. Are A and B the same? Yes or No? No\n\nA is location-zip. B is patients-marital. Are A and B the same? Yes or No? No\n\nA is location-zip. B is patients-ssn. Are A and B the same? Yes or No? No\n",
        "location-county": "A is location-county. B is conditions-patient. Are A and B the same? Yes or No? No\n\nA is location-county. B is encounters-cost. Are A and B the same? Yes or No? No\n\nA is location-county. B is patients-gender. Are A and B the same? Yes or No? No\n",
        "location-location_source_value": "A is location-location_source_value. B is patients-zip. Are A and B the same? Yes or No? No\n\nA is location-location_source_value. B is patients-city. Are A and B the same? Yes or No? No\n\nA is location-location_source_value. B is medications-encounter. Are A and B the same? Yes or No? No\n",
        "observation-observation_id": "A is observation-observation_id. B is procedures-encounter. Are A and B the same? Yes or No? No\n\nA is observation-observation_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is observation-observation_id. B is medications-encounter. Are A and B the same? Yes or No? No\n",
        "observation-person_id": "A is observation-person_id. B is allergies-patient. Are A and B the same? Yes or No? Yes\n\nA is observation-person_id. B is encounters-start. Are A and B the same? Yes or No? No\n\nA is observation-person_id. B is allergies-description. Are A and B the same? Yes or No? No\n",
        "observation-observation_concept_id": "A is observation-observation_concept_id. B is conditions-code. Are A and B the same? Yes or No? Yes\n\nA is observation-observation_concept_id. B is providers-name. Are A and B the same? Yes or No? No\n\nA is observation-observation_concept_id. B is providers-organization. Are A and B the same? Yes or No? No\n",
        "observation-observation_date": "A is observation-observation_date. B is allergies-start. Are A and B the same? Yes or No? Yes\n\nA is observation-observation_date. B is careplans-start. Are A and B the same? Yes or No? No\n\nA is observation-observation_date. B is medications-cost. Are A and B the same? Yes or No? No\n",
        "observation-observation_datetime": "A is observation-observation_datetime. B is allergies-start. Are A and B the same? Yes or No? Yes\n\nA is observation-observation_datetime. B is conditions-encounter. Are A and B the same? Yes or No? No\n\nA is observation-observation_datetime. B is medications-description. Are A and B the same? Yes or No? No\n",
        "observation-observation_type_concept_id": "A is observation-observation_type_concept_id. B is encounters-code. Are A and B the same? Yes or No? No\n\nA is observation-observation_type_concept_id. B is observations-code. Are A and B the same? Yes or No? No\n\nA is observation-observation_type_concept_id. B is providers-name. Are A and B the same? Yes or No? No\n",
        "observation-value_as_number": "A is observation-value_as_number. B is careplans-encounter. Are A and B the same? Yes or No? No\n\nA is observation-value_as_number. B is imaging_studies-patient. Are A and B the same? Yes or No? No\n\nA is observation-value_as_number. B is encounters-description. Are A and B the same? Yes or No? No\n",
        "observation-value_as_string": "A is observation-value_as_string. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is observation-value_as_string. B is imaging_studies-date. Are A and B the same? Yes or No? No\n\nA is observation-value_as_string. B is patients-race. Are A and B the same? Yes or No? No\n",
        "observation-value_as_concept_id": "A is observation-value_as_concept_id. B is providers-id. Are A and B the same? Yes or No? No\n\nA is observation-value_as_concept_id. B is careplans-patient. Are A and B the same? Yes or No? No\n\nA is observation-value_as_concept_id. B is providers-address. Are A and B the same? Yes or No? No\n",
        "observation-qualifier_concept_id": "A is observation-qualifier_concept_id. B is immunizations-patient. Are A and B the same? Yes or No? No\n\nA is observation-qualifier_concept_id. B is immunizations-code. Are A and B the same? Yes or No? No\n\nA is observation-qualifier_concept_id. B is medications-start. Are A and B the same? Yes or No? No\n",
        "observation-unit_concept_id": "A is observation-unit_concept_id. B is procedures-code. Are A and B the same? Yes or No? No\n\nA is observation-unit_concept_id. B is allergies-patient. Are A and B the same? Yes or No? No\n\nA is observation-unit_concept_id. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n",
        "observation-provider_id": "A is observation-provider_id. B is careplans-stop. Are A and B the same? Yes or No? No\n\nA is observation-provider_id. B is organizations-phone. Are A and B the same? Yes or No? No\n\nA is observation-provider_id. B is conditions-stop. Are A and B the same? Yes or No? No\n",
        "observation-visit_occurrence_id": "A is observation-visit_occurrence_id. B is allergies-encounter. Are A and B the same? Yes or No? Yes\n\nA is observation-visit_occurrence_id. B is providers-address. Are A and B the same? Yes or No? No\n\nA is observation-visit_occurrence_id. B is organizations-phone. Are A and B the same? Yes or No? No\n",
        "observation-visit_detail_id": "A is observation-visit_detail_id. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n\nA is observation-visit_detail_id. B is imaging_studies-encounter. Are A and B the same? Yes or No? No\n\nA is observation-visit_detail_id. B is providers-utilization. Are A and B the same? Yes or No? No\n",
        "observation-observation_source_value": "A is observation-observation_source_value. B is allergies-code. Are A and B the same? Yes or No? Yes\n\nA is observation-observation_source_value. B is immunizations-encounter. Are A and B the same? Yes or No? No\n\nA is observation-observation_source_value. B is observations-type. Are A and B the same? Yes or No? No\n",
        "observation-observation_source_concept_id": "A is observation-observation_source_concept_id. B is allergies-code. Are A and B the same? Yes or No? Yes\n\nA is observation-observation_source_concept_id. B is patients-last. Are A and B the same? Yes or No? No\n\nA is observation-observation_source_concept_id. B is patients-prefix. Are A and B the same? Yes or No? No\n",
        "observation-unit_source_value": "A is observation-unit_source_value. B is immunizations-description. Are A and B the same? Yes or No? No\n\nA is observation-unit_source_value. B is providers-organization. Are A and B the same? Yes or No? No\n\nA is observation-unit_source_value. B is careplans-patient. Are A and B the same? Yes or No? No\n",
        "observation-qualifier_source_value": "A is observation-qualifier_source_value. B is careplans-description. Are A and B the same? Yes or No? No\n\nA is observation-qualifier_source_value. B is conditions-patient. Are A and B the same? Yes or No? No\n\nA is observation-qualifier_source_value. B is conditions-description. Are A and B the same? Yes or No? No\n",
        "observation-observation_event_id": "A is observation-observation_event_id. B is observations-type. Are A and B the same? Yes or No? No\n\nA is observation-observation_event_id. B is patients-address. Are A and B the same? Yes or No? No\n\nA is observation-observation_event_id. B is providers-utilization. Are A and B the same? Yes or No? No\n",
        "observation-obs_event_field_concept_id": "A is observation-obs_event_field_concept_id. B is procedures-description. Are A and B the same? Yes or No? No\n\nA is observation-obs_event_field_concept_id. B is medications-cost. Are A and B the same? Yes or No? No\n\nA is observation-obs_event_field_concept_id. B is procedures-code. Are A and B the same? Yes or No? No\n",
        "observation-value_as_datetime": "A is observation-value_as_datetime. B is patients-marital. Are A and B the same? Yes or No? No\n\nA is observation-value_as_datetime. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is observation-value_as_datetime. B is procedures-cost. Are A and B the same? Yes or No? No\n",
        "observation_period-observation_period_id": "A is observation_period-observation_period_id. B is allergies-description. Are A and B the same? Yes or No? No\n\nA is observation_period-observation_period_id. B is observations-description. Are A and B the same? Yes or No? No\n\nA is observation_period-observation_period_id. B is providers-utilization. Are A and B the same? Yes or No? No\n",
        "observation_period-person_id": "A is observation_period-person_id. B is encounters-patient. Are A and B the same? Yes or No? Yes\n\nA is observation_period-person_id. B is patients-id. Are A and B the same? Yes or No? No\n\nA is observation_period-person_id. B is conditions-description. Are A and B the same? Yes or No? No\n",
        "observation_period-observation_period_start_date": "A is observation_period-observation_period_start_date. B is encounters-start. Are A and B the same? Yes or No? Yes\n\nA is observation_period-observation_period_start_date. B is procedures-patient. Are A and B the same? Yes or No? No\n\nA is observation_period-observation_period_start_date. B is patients-birthplace. Are A and B the same? Yes or No? No\n",
        "observation_period-observation_period_end_date": "A is observation_period-observation_period_end_date. B is encounters-stop. Are A and B the same? Yes or No? Yes\n\nA is observation_period-observation_period_end_date. B is imaging_studies-modality code. Are A and B the same? Yes or No? No\n\nA is observation_period-observation_period_end_date. B is procedures-code. Are A and B the same? Yes or No? No\n",
        "observation_period-period_type_concept_id": "A is observation_period-period_type_concept_id. B is patients-prefix. Are A and B the same? Yes or No? No\n\nA is observation_period-period_type_concept_id. B is conditions-start. Are A and B the same? Yes or No? No\n\nA is observation_period-period_type_concept_id. B is procedures-reasondescription. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_exposure_id": "A is drug_exposure-drug_exposure_id. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_exposure_id. B is procedures-date. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_exposure_id. B is medications-start. Are A and B the same? Yes or No? No\n",
        "drug_exposure-person_id": "A is drug_exposure-person_id. B is immunizations-patient. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-person_id. B is observations-encounter. Are A and B the same? Yes or No? No\n\nA is drug_exposure-person_id. B is patients-passport. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_concept_id": "A is drug_exposure-drug_concept_id. B is medications-code. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-drug_concept_id. B is conditions-patient. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_concept_id. B is allergies-stop. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_exposure_start_date": "A is drug_exposure-drug_exposure_start_date. B is medications-start. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-drug_exposure_start_date. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_exposure_start_date. B is patients-deathdate. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_exposure_start_datetime": "A is drug_exposure-drug_exposure_start_datetime. B is medications-start. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-drug_exposure_start_datetime. B is patients-ssn. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_exposure_start_datetime. B is immunizations-patient. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_exposure_end_date": "A is drug_exposure-drug_exposure_end_date. B is medications-stop. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-drug_exposure_end_date. B is conditions-description. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_exposure_end_date. B is organizations-city. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_exposure_end_datetime": "A is drug_exposure-drug_exposure_end_datetime. B is medications-start. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-drug_exposure_end_datetime. B is conditions-start. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_exposure_end_datetime. B is observations-patient. Are A and B the same? Yes or No? No\n",
        "drug_exposure-verbatim_end_date": "A is drug_exposure-verbatim_end_date. B is immunizations-date. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-verbatim_end_date. B is organizations-phone. Are A and B the same? Yes or No? No\n\nA is drug_exposure-verbatim_end_date. B is medications-cost. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_type_concept_id": "A is drug_exposure-drug_type_concept_id. B is providers-id. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_type_concept_id. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_type_concept_id. B is immunizations-encounter. Are A and B the same? Yes or No? No\n",
        "drug_exposure-stop_reason": "A is drug_exposure-stop_reason. B is patients-gender. Are A and B the same? Yes or No? No\n\nA is drug_exposure-stop_reason. B is careplans-start. Are A and B the same? Yes or No? No\n\nA is drug_exposure-stop_reason. B is procedures-date. Are A and B the same? Yes or No? No\n",
        "drug_exposure-refills": "A is drug_exposure-refills. B is observations-type. Are A and B the same? Yes or No? No\n\nA is drug_exposure-refills. B is immunizations-encounter. Are A and B the same? Yes or No? No\n\nA is drug_exposure-refills. B is organizations-zip. Are A and B the same? Yes or No? No\n",
        "drug_exposure-quantity": "A is drug_exposure-quantity. B is patients-address. Are A and B the same? Yes or No? No\n\nA is drug_exposure-quantity. B is encounters-reasoncode. Are A and B the same? Yes or No? No\n\nA is drug_exposure-quantity. B is providers-id. Are A and B the same? Yes or No? No\n",
        "drug_exposure-days_supply": "A is drug_exposure-days_supply. B is medications-stop. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-days_supply. B is patients-birthplace. Are A and B the same? Yes or No? No\n\nA is drug_exposure-days_supply. B is medications-description. Are A and B the same? Yes or No? No\n",
        "drug_exposure-sig": "A is drug_exposure-sig. B is imaging_studies-id. Are A and B the same? Yes or No? No\n\nA is drug_exposure-sig. B is medications-cost. Are A and B the same? Yes or No? No\n\nA is drug_exposure-sig. B is organizations-name. Are A and B the same? Yes or No? No\n",
        "drug_exposure-route_concept_id": "A is drug_exposure-route_concept_id. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n\nA is drug_exposure-route_concept_id. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is drug_exposure-route_concept_id. B is conditions-encounter. Are A and B the same? Yes or No? No\n",
        "drug_exposure-lot_number": "A is drug_exposure-lot_number. B is providers-gender. Are A and B the same? Yes or No? No\n\nA is drug_exposure-lot_number. B is allergies-code. Are A and B the same? Yes or No? No\n\nA is drug_exposure-lot_number. B is encounters-patient. Are A and B the same? Yes or No? No\n",
        "drug_exposure-provider_id": "A is drug_exposure-provider_id. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n\nA is drug_exposure-provider_id. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is drug_exposure-provider_id. B is careplans-reasoncode. Are A and B the same? Yes or No? No\n",
        "drug_exposure-visit_occurrence_id": "A is drug_exposure-visit_occurrence_id. B is medications-encounter. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-visit_occurrence_id. B is observations-units. Are A and B the same? Yes or No? No\n\nA is drug_exposure-visit_occurrence_id. B is observations-encounter. Are A and B the same? Yes or No? No\n",
        "drug_exposure-visit_detail_id": "A is drug_exposure-visit_detail_id. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is drug_exposure-visit_detail_id. B is medications-reasoncode. Are A and B the same? Yes or No? No\n\nA is drug_exposure-visit_detail_id. B is imaging_studies-id. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_source_value": "A is drug_exposure-drug_source_value. B is medications-code. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-drug_source_value. B is procedures-description. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_source_value. B is allergies-start. Are A and B the same? Yes or No? No\n",
        "drug_exposure-drug_source_concept_id": "A is drug_exposure-drug_source_concept_id. B is conditions-code. Are A and B the same? Yes or No? Yes\n\nA is drug_exposure-drug_source_concept_id. B is careplans-encounter. Are A and B the same? Yes or No? No\n\nA is drug_exposure-drug_source_concept_id. B is immunizations-date. Are A and B the same? Yes or No? No\n",
        "drug_exposure-route_source_value": "A is drug_exposure-route_source_value. B is patients-suffix. Are A and B the same? Yes or No? No\n\nA is drug_exposure-route_source_value. B is imaging_studies-modality description. Are A and B the same? Yes or No? No\n\nA is drug_exposure-route_source_value. B is conditions-stop. Are A and B the same? Yes or No? No\n",
        "drug_exposure-dose_unit_source_value": "A is drug_exposure-dose_unit_source_value. B is careplans-encounter. Are A and B the same? Yes or No? No\n\nA is drug_exposure-dose_unit_source_value. B is patients-state. Are A and B the same? Yes or No? No\n\nA is drug_exposure-dose_unit_source_value. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_exposure_id": "A is device_exposure-device_exposure_id. B is organizations-utilization. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_id. B is organizations-address. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_id. B is allergies-start. Are A and B the same? Yes or No? No\n",
        "device_exposure-person_id": "A is device_exposure-person_id. B is medications-description. Are A and B the same? Yes or No? No\n\nA is device_exposure-person_id. B is encounters-code. Are A and B the same? Yes or No? No\n\nA is device_exposure-person_id. B is patients-passport. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_concept_id": "A is device_exposure-device_concept_id. B is observations-units. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_concept_id. B is patients-address. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_concept_id. B is encounters-stop. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_exposure_start_date": "A is device_exposure-device_exposure_start_date. B is patients-address. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_start_date. B is allergies-start. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_start_date. B is encounters-patient. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_exposure_start_datetime": "A is device_exposure-device_exposure_start_datetime. B is imaging_studies-modality description. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_start_datetime. B is observations-encounter. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_start_datetime. B is conditions-code. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_exposure_end_date": "A is device_exposure-device_exposure_end_date. B is organizations-utilization. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_end_date. B is medications-patient. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_end_date. B is encounters-start. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_exposure_end_datetime": "A is device_exposure-device_exposure_end_datetime. B is observations-value. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_end_datetime. B is medications-reasondescription. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_exposure_end_datetime. B is encounters-reasoncode. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_type_concept_id": "A is device_exposure-device_type_concept_id. B is procedures-encounter. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_type_concept_id. B is careplans-description. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_type_concept_id. B is observations-patient. Are A and B the same? Yes or No? No\n",
        "device_exposure-unique_device_id": "A is device_exposure-unique_device_id. B is encounters-provider. Are A and B the same? Yes or No? No\n\nA is device_exposure-unique_device_id. B is patients-suffix. Are A and B the same? Yes or No? No\n\nA is device_exposure-unique_device_id. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n",
        "device_exposure-quantity": "A is device_exposure-quantity. B is conditions-start. Are A and B the same? Yes or No? No\n\nA is device_exposure-quantity. B is immunizations-cost. Are A and B the same? Yes or No? No\n\nA is device_exposure-quantity. B is conditions-code. Are A and B the same? Yes or No? No\n",
        "device_exposure-provider_id": "A is device_exposure-provider_id. B is providers-gender. Are A and B the same? Yes or No? No\n\nA is device_exposure-provider_id. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n\nA is device_exposure-provider_id. B is immunizations-description. Are A and B the same? Yes or No? No\n",
        "device_exposure-visit_occurrence_id": "A is device_exposure-visit_occurrence_id. B is encounters-reasondescription. Are A and B the same? Yes or No? No\n\nA is device_exposure-visit_occurrence_id. B is conditions-encounter. Are A and B the same? Yes or No? No\n\nA is device_exposure-visit_occurrence_id. B is providers-gender. Are A and B the same? Yes or No? No\n",
        "device_exposure-visit_detail_id": "A is device_exposure-visit_detail_id. B is medications-reasondescription. Are A and B the same? Yes or No? No\n\nA is device_exposure-visit_detail_id. B is patients-city. Are A and B the same? Yes or No? No\n\nA is device_exposure-visit_detail_id. B is patients-prefix. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_source_value": "A is device_exposure-device_source_value. B is allergies-patient. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_source_value. B is immunizations-date. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_source_value. B is conditions-description. Are A and B the same? Yes or No? No\n",
        "device_exposure-device_source_concept_id": "A is device_exposure-device_source_concept_id. B is patients-suffix. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_source_concept_id. B is careplans-description. Are A and B the same? Yes or No? No\n\nA is device_exposure-device_source_concept_id. B is encounters-id. Are A and B the same? Yes or No? No\n",
        "note-note_id": "A is note-note_id. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is note-note_id. B is medications-code. Are A and B the same? Yes or No? No\n\nA is note-note_id. B is allergies-start. Are A and B the same? Yes or No? No\n",
        "note-person_id": "A is note-person_id. B is conditions-description. Are A and B the same? Yes or No? No\n\nA is note-person_id. B is patients-race. Are A and B the same? Yes or No? No\n\nA is note-person_id. B is careplans-id. Are A and B the same? Yes or No? No\n",
        "note-note_event_id": "A is note-note_event_id. B is patients-first. Are A and B the same? Yes or No? No\n\nA is note-note_event_id. B is observations-code. Are A and B the same? Yes or No? No\n\nA is note-note_event_id. B is careplans-stop. Are A and B the same? Yes or No? No\n",
        "note-note_event_field_concept_id": "A is note-note_event_field_concept_id. B is imaging_studies-modality code. Are A and B the same? Yes or No? No\n\nA is note-note_event_field_concept_id. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n\nA is note-note_event_field_concept_id. B is careplans-code. Are A and B the same? Yes or No? No\n",
        "note-note_date": "A is note-note_date. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is note-note_date. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is note-note_date. B is medications-reasoncode. Are A and B the same? Yes or No? No\n",
        "note-note_datetime": "A is note-note_datetime. B is encounters-id. Are A and B the same? Yes or No? No\n\nA is note-note_datetime. B is patients-id. Are A and B the same? Yes or No? No\n\nA is note-note_datetime. B is procedures-encounter. Are A and B the same? Yes or No? No\n",
        "note-note_type_concept_id": "A is note-note_type_concept_id. B is encounters-description. Are A and B the same? Yes or No? No\n\nA is note-note_type_concept_id. B is patients-city. Are A and B the same? Yes or No? No\n\nA is note-note_type_concept_id. B is medications-reasondescription. Are A and B the same? Yes or No? No\n",
        "note-note_class_concept_id": "A is note-note_class_concept_id. B is providers-organization. Are A and B the same? Yes or No? No\n\nA is note-note_class_concept_id. B is observations-date. Are A and B the same? Yes or No? No\n\nA is note-note_class_concept_id. B is medications-code. Are A and B the same? Yes or No? No\n",
        "note-note_title": "A is note-note_title. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n\nA is note-note_title. B is patients-last. Are A and B the same? Yes or No? No\n\nA is note-note_title. B is providers-organization. Are A and B the same? Yes or No? No\n",
        "note-note_text": "A is note-note_text. B is immunizations-patient. Are A and B the same? Yes or No? No\n\nA is note-note_text. B is imaging_studies-id. Are A and B the same? Yes or No? No\n\nA is note-note_text. B is imaging_studies-body site description. Are A and B the same? Yes or No? No\n",
        "note-encoding_concept_id": "A is note-encoding_concept_id. B is patients-ssn. Are A and B the same? Yes or No? No\n\nA is note-encoding_concept_id. B is organizations-state. Are A and B the same? Yes or No? No\n\nA is note-encoding_concept_id. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n",
        "note-language_concept_id": "A is note-language_concept_id. B is imaging_studies-body site code. Are A and B the same? Yes or No? No\n\nA is note-language_concept_id. B is careplans-reasoncode. Are A and B the same? Yes or No? No\n\nA is note-language_concept_id. B is imaging_studies-modality description. Are A and B the same? Yes or No? No\n",
        "note-provider_id": "A is note-provider_id. B is providers-gender. Are A and B the same? Yes or No? No\n\nA is note-provider_id. B is immunizations-code. Are A and B the same? Yes or No? No\n\nA is note-provider_id. B is imaging_studies-id. Are A and B the same? Yes or No? No\n",
        "note-visit_occurrence_id": "A is note-visit_occurrence_id. B is patients-gender. Are A and B the same? Yes or No? No\n\nA is note-visit_occurrence_id. B is careplans-encounter. Are A and B the same? Yes or No? No\n\nA is note-visit_occurrence_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n",
        "note-visit_detail_id": "A is note-visit_detail_id. B is allergies-encounter. Are A and B the same? Yes or No? No\n\nA is note-visit_detail_id. B is immunizations-date. Are A and B the same? Yes or No? No\n\nA is note-visit_detail_id. B is providers-gender. Are A and B the same? Yes or No? No\n",
        "note-note_source_value": "A is note-note_source_value. B is patients-ssn. Are A and B the same? Yes or No? No\n\nA is note-note_source_value. B is medications-patient. Are A and B the same? Yes or No? No\n\nA is note-note_source_value. B is immunizations-encounter. Are A and B the same? Yes or No? No\n",
        "fact_relationship-domain_concept_id_1": "A is fact_relationship-domain_concept_id_1. B is imaging_studies-patient. Are A and B the same? Yes or No? No\n\nA is fact_relationship-domain_concept_id_1. B is careplans-id. Are A and B the same? Yes or No? No\n\nA is fact_relationship-domain_concept_id_1. B is medications-encounter. Are A and B the same? Yes or No? No\n",
        "fact_relationship-fact_id_1": "A is fact_relationship-fact_id_1. B is careplans-reasoncode. Are A and B the same? Yes or No? No\n\nA is fact_relationship-fact_id_1. B is medications-encounter. Are A and B the same? Yes or No? No\n\nA is fact_relationship-fact_id_1. B is immunizations-description. Are A and B the same? Yes or No? No\n",
        "fact_relationship-domain_concept_id_2": "A is fact_relationship-domain_concept_id_2. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is fact_relationship-domain_concept_id_2. B is conditions-patient. Are A and B the same? Yes or No? No\n\nA is fact_relationship-domain_concept_id_2. B is patients-last. Are A and B the same? Yes or No? No\n",
        "fact_relationship-fact_id_2": "A is fact_relationship-fact_id_2. B is medications-start. Are A and B the same? Yes or No? No\n\nA is fact_relationship-fact_id_2. B is procedures-patient. Are A and B the same? Yes or No? No\n\nA is fact_relationship-fact_id_2. B is providers-speciality. Are A and B the same? Yes or No? No\n",
        "fact_relationship-relationship_concept_id": "A is fact_relationship-relationship_concept_id. B is immunizations-encounter. Are A and B the same? Yes or No? No\n\nA is fact_relationship-relationship_concept_id. B is allergies-encounter. Are A and B the same? Yes or No? No\n\nA is fact_relationship-relationship_concept_id. B is patients-gender. Are A and B the same? Yes or No? No\n",
        "payer_plan_period-payer_plan_period_id": "A is payer_plan_period-payer_plan_period_id. B is providers-zip. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-payer_plan_period_id. B is medications-encounter. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-payer_plan_period_id. B is allergies-stop. Are A and B the same? Yes or No? No\n",
        "payer_plan_period-person_id": "A is payer_plan_period-person_id. B is patients-birthdate. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-person_id. B is medications-cost. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-person_id. B is patients-maiden. Are A and B the same? Yes or No? No\n",
        "payer_plan_period-payer_plan_period_start_date": "A is payer_plan_period-payer_plan_period_start_date. B is organizations-zip. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-payer_plan_period_start_date. B is patients-race. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-payer_plan_period_start_date. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n",
        "payer_plan_period-payer_plan_period_end_date": "A is payer_plan_period-payer_plan_period_end_date. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-payer_plan_period_end_date. B is allergies-patient. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-payer_plan_period_end_date. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n",
        "payer_plan_period-payer_source_value": "A is payer_plan_period-payer_source_value. B is conditions-stop. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-payer_source_value. B is patients-suffix. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-payer_source_value. B is patients-passport. Are A and B the same? Yes or No? No\n",
        "payer_plan_period-plan_source_value": "A is payer_plan_period-plan_source_value. B is organizations-id. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-plan_source_value. B is allergies-start. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-plan_source_value. B is medications-stop. Are A and B the same? Yes or No? No\n",
        "payer_plan_period-family_source_value": "A is payer_plan_period-family_source_value. B is patients-zip. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-family_source_value. B is observations-encounter. Are A and B the same? Yes or No? No\n\nA is payer_plan_period-family_source_value. B is patients-gender. Are A and B the same? Yes or No? No\n",
        "cdm_source-cdm_source_name": "A is cdm_source-cdm_source_name. B is immunizations-encounter. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_source_name. B is conditions-description. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_source_name. B is immunizations-date. Are A and B the same? Yes or No? No\n",
        "cdm_source-cdm_source_abbreviation": "A is cdm_source-cdm_source_abbreviation. B is imaging_studies-id. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_source_abbreviation. B is patients-race. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_source_abbreviation. B is patients-city. Are A and B the same? Yes or No? No\n",
        "cdm_source-cdm_holder": "A is cdm_source-cdm_holder. B is medications-patient. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_holder. B is allergies-description. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_holder. B is observations-encounter. Are A and B the same? Yes or No? No\n",
        "cdm_source-source_description": "A is cdm_source-source_description. B is careplans-reasoncode. Are A and B the same? Yes or No? No\n\nA is cdm_source-source_description. B is procedures-cost. Are A and B the same? Yes or No? No\n\nA is cdm_source-source_description. B is providers-speciality. Are A and B the same? Yes or No? No\n",
        "cdm_source-source_documentation_reference": "A is cdm_source-source_documentation_reference. B is procedures-cost. Are A and B the same? Yes or No? No\n\nA is cdm_source-source_documentation_reference. B is patients-drivers. Are A and B the same? Yes or No? No\n\nA is cdm_source-source_documentation_reference. B is careplans-encounter. Are A and B the same? Yes or No? No\n",
        "cdm_source-cdm_etl_reference": "A is cdm_source-cdm_etl_reference. B is procedures-description. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_etl_reference. B is procedures-reasondescription. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_etl_reference. B is procedures-date. Are A and B the same? Yes or No? No\n",
        "cdm_source-source_release_date": "A is cdm_source-source_release_date. B is providers-speciality. Are A and B the same? Yes or No? No\n\nA is cdm_source-source_release_date. B is patients-birthplace. Are A and B the same? Yes or No? No\n\nA is cdm_source-source_release_date. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n",
        "cdm_source-cdm_release_date": "A is cdm_source-cdm_release_date. B is conditions-stop. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_release_date. B is patients-passport. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_release_date. B is conditions-encounter. Are A and B the same? Yes or No? No\n",
        "cdm_source-cdm_version": "A is cdm_source-cdm_version. B is patients-state. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_version. B is patients-suffix. Are A and B the same? Yes or No? No\n\nA is cdm_source-cdm_version. B is patients-id. Are A and B the same? Yes or No? No\n",
        "cdm_source-vocabulary_version": "A is cdm_source-vocabulary_version. B is encounters-encounterclass. Are A and B the same? Yes or No? No\n\nA is cdm_source-vocabulary_version. B is conditions-patient. Are A and B the same? Yes or No? No\n\nA is cdm_source-vocabulary_version. B is allergies-code. Are A and B the same? Yes or No? No\n",
        "care_site-care_site_id": "A is care_site-care_site_id. B is patients-race. Are A and B the same? Yes or No? No\n\nA is care_site-care_site_id. B is patients-address. Are A and B the same? Yes or No? No\n\nA is care_site-care_site_id. B is providers-id. Are A and B the same? Yes or No? No\n",
        "care_site-care_site_name": "A is care_site-care_site_name. B is careplans-patient. Are A and B the same? Yes or No? No\n\nA is care_site-care_site_name. B is careplans-description. Are A and B the same? Yes or No? No\n\nA is care_site-care_site_name. B is organizations-phone. Are A and B the same? Yes or No? No\n",
        "care_site-place_of_service_concept_id": "A is care_site-place_of_service_concept_id. B is encounters-reasondescription. Are A and B the same? Yes or No? No\n\nA is care_site-place_of_service_concept_id. B is patients-id. Are A and B the same? Yes or No? No\n\nA is care_site-place_of_service_concept_id. B is medications-stop. Are A and B the same? Yes or No? No\n",
        "care_site-location_id": "A is care_site-location_id. B is patients-city. Are A and B the same? Yes or No? No\n\nA is care_site-location_id. B is immunizations-cost. Are A and B the same? Yes or No? No\n\nA is care_site-location_id. B is organizations-phone. Are A and B the same? Yes or No? No\n",
        "care_site-care_site_source_value": "A is care_site-care_site_source_value. B is immunizations-cost. Are A and B the same? Yes or No? No\n\nA is care_site-care_site_source_value. B is careplans-id. Are A and B the same? Yes or No? No\n\nA is care_site-care_site_source_value. B is imaging_studies-body site code. Are A and B the same? Yes or No? No\n",
        "care_site-place_of_service_source_value": "A is care_site-place_of_service_source_value. B is conditions-encounter. Are A and B the same? Yes or No? No\n\nA is care_site-place_of_service_source_value. B is allergies-description. Are A and B the same? Yes or No? No\n\nA is care_site-place_of_service_source_value. B is allergies-start. Are A and B the same? Yes or No? No\n",
        "cohort-cohort_definition_id": "A is cohort-cohort_definition_id. B is immunizations-date. Are A and B the same? Yes or No? No\n\nA is cohort-cohort_definition_id. B is procedures-patient. Are A and B the same? Yes or No? No\n\nA is cohort-cohort_definition_id. B is providers-speciality. Are A and B the same? Yes or No? No\n",
        "cohort-subject_id": "A is cohort-subject_id. B is organizations-address. Are A and B the same? Yes or No? No\n\nA is cohort-subject_id. B is allergies-stop. Are A and B the same? Yes or No? No\n\nA is cohort-subject_id. B is organizations-zip. Are A and B the same? Yes or No? No\n",
        "cohort-cohort_start_date": "A is cohort-cohort_start_date. B is organizations-state. Are A and B the same? Yes or No? No\n\nA is cohort-cohort_start_date. B is imaging_studies-sop description. Are A and B the same? Yes or No? No\n\nA is cohort-cohort_start_date. B is observations-patient. Are A and B the same? Yes or No? No\n",
        "cohort-cohort_end_date": "A is cohort-cohort_end_date. B is imaging_studies-body site description. Are A and B the same? Yes or No? No\n\nA is cohort-cohort_end_date. B is patients-id. Are A and B the same? Yes or No? No\n\nA is cohort-cohort_end_date. B is patients-last. Are A and B the same? Yes or No? No\n",
        "cohort_definition-cohort_definition_id": "A is cohort_definition-cohort_definition_id. B is patients-birthdate. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_definition_id. B is patients-deathdate. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_definition_id. B is organizations-name. Are A and B the same? Yes or No? No\n",
        "cohort_definition-cohort_definition_name": "A is cohort_definition-cohort_definition_name. B is allergies-encounter. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_definition_name. B is conditions-encounter. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_definition_name. B is allergies-start. Are A and B the same? Yes or No? No\n",
        "cohort_definition-cohort_definition_description": "A is cohort_definition-cohort_definition_description. B is organizations-address. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_definition_description. B is encounters-start. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_definition_description. B is careplans-reasondescription. Are A and B the same? Yes or No? No\n",
        "cohort_definition-definition_type_concept_id": "A is cohort_definition-definition_type_concept_id. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is cohort_definition-definition_type_concept_id. B is medications-cost. Are A and B the same? Yes or No? No\n\nA is cohort_definition-definition_type_concept_id. B is encounters-reasondescription. Are A and B the same? Yes or No? No\n",
        "cohort_definition-cohort_definition_syntax": "A is cohort_definition-cohort_definition_syntax. B is patients-city. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_definition_syntax. B is medications-stop. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_definition_syntax. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n",
        "cohort_definition-subject_concept_id": "A is cohort_definition-subject_concept_id. B is procedures-description. Are A and B the same? Yes or No? No\n\nA is cohort_definition-subject_concept_id. B is encounters-description. Are A and B the same? Yes or No? No\n\nA is cohort_definition-subject_concept_id. B is encounters-reasoncode. Are A and B the same? Yes or No? No\n",
        "cohort_definition-cohort_initiation_date": "A is cohort_definition-cohort_initiation_date. B is encounters-reasondescription. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_initiation_date. B is patients-marital. Are A and B the same? Yes or No? No\n\nA is cohort_definition-cohort_initiation_date. B is allergies-description. Are A and B the same? Yes or No? No\n",
        "provider-provider_id": "A is provider-provider_id. B is observations-units. Are A and B the same? Yes or No? No\n\nA is provider-provider_id. B is imaging_studies-patient. Are A and B the same? Yes or No? No\n\nA is provider-provider_id. B is patients-race. Are A and B the same? Yes or No? No\n",
        "provider-provider_name": "A is provider-provider_name. B is patients-race. Are A and B the same? Yes or No? No\n\nA is provider-provider_name. B is immunizations-patient. Are A and B the same? Yes or No? No\n\nA is provider-provider_name. B is procedures-reasoncode. Are A and B the same? Yes or No? No\n",
        "provider-npi": "A is provider-npi. B is medications-encounter. Are A and B the same? Yes or No? No\n\nA is provider-npi. B is medications-start. Are A and B the same? Yes or No? No\n\nA is provider-npi. B is careplans-start. Are A and B the same? Yes or No? No\n",
        "provider-dea": "A is provider-dea. B is organizations-phone. Are A and B the same? Yes or No? No\n\nA is provider-dea. B is allergies-start. Are A and B the same? Yes or No? No\n\nA is provider-dea. B is providers-address. Are A and B the same? Yes or No? No\n",
        "provider-specialty_concept_id": "A is provider-specialty_concept_id. B is medications-encounter. Are A and B the same? Yes or No? No\n\nA is provider-specialty_concept_id. B is encounters-id. Are A and B the same? Yes or No? No\n\nA is provider-specialty_concept_id. B is medications-start. Are A and B the same? Yes or No? No\n",
        "provider-care_site_id": "A is provider-care_site_id. B is providers-id. Are A and B the same? Yes or No? No\n\nA is provider-care_site_id. B is organizations-id. Are A and B the same? Yes or No? No\n\nA is provider-care_site_id. B is patients-prefix. Are A and B the same? Yes or No? No\n",
        "provider-year_of_birth": "A is provider-year_of_birth. B is patients-ethnicity. Are A and B the same? Yes or No? No\n\nA is provider-year_of_birth. B is observations-code. Are A and B the same? Yes or No? No\n\nA is provider-year_of_birth. B is organizations-zip. Are A and B the same? Yes or No? No\n",
        "provider-gender_concept_id": "A is provider-gender_concept_id. B is careplans-patient. Are A and B the same? Yes or No? No\n\nA is provider-gender_concept_id. B is medications-encounter. Are A and B the same? Yes or No? No\n\nA is provider-gender_concept_id. B is patients-state. Are A and B the same? Yes or No? No\n",
        "provider-provider_source_value": "A is provider-provider_source_value. B is allergies-description. Are A and B the same? Yes or No? No\n\nA is provider-provider_source_value. B is patients-marital. Are A and B the same? Yes or No? No\n\nA is provider-provider_source_value. B is conditions-description. Are A and B the same? Yes or No? No\n",
        "provider-specialty_source_value": "A is provider-specialty_source_value. B is organizations-name. Are A and B the same? Yes or No? No\n\nA is provider-specialty_source_value. B is careplans-code. Are A and B the same? Yes or No? No\n\nA is provider-specialty_source_value. B is patients-city. Are A and B the same? Yes or No? No\n",
        "provider-specialty_source_concept_id": "A is provider-specialty_source_concept_id. B is careplans-start. Are A and B the same? Yes or No? No\n\nA is provider-specialty_source_concept_id. B is encounters-provider. Are A and B the same? Yes or No? No\n\nA is provider-specialty_source_concept_id. B is medications-patient. Are A and B the same? Yes or No? No\n",
        "provider-gender_source_value": "A is provider-gender_source_value. B is providers-name. Are A and B the same? Yes or No? No\n\nA is provider-gender_source_value. B is imaging_studies-sop code. Are A and B the same? Yes or No? No\n\nA is provider-gender_source_value. B is observations-encounter. Are A and B the same? Yes or No? No\n",
        "provider-gender_source_concept_id": "A is provider-gender_source_concept_id. B is patients-drivers. Are A and B the same? Yes or No? No\n\nA is provider-gender_source_concept_id. B is observations-patient. Are A and B the same? Yes or No? No\n\nA is provider-gender_source_concept_id. B is patients-maiden. Are A and B the same? Yes or No? No\n",
    },
    f"{DATASET_PATH}/error_detection/Adult": {
        "workclass": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: **MoreThan50K.\n\nIs there an error in workclass: Private? No\n\nage: <18. workclass: nan. education: 10th. maritalstatus: Never-married. occupation: nan. relationship: Own-child. race: White. sex: Female. hoursperweek: 40. country: United-States. income: LessThan50K. \n\nIs there an error in workclass: nan? No\n\nage: 18-21. workclass: nan. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LesjsThan50K.\n\nIs there an error in workclass: nan? No\n",
        "education": "\nage: 18-21. workclass: Private. education: 11th. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: **MoreThan50K.\n\nIs there an error in education: 11th? No\n\n---\n\nage: 18-21. workclass: Private. education: 1st-4th. maritalstatus: Never-married. occupation: Exec-managerial. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: **MoreThan50K.\n\nIs there an error in education: 1st-4th? No\n\n---\n\nage: 18-21. workclass: Private. education: nan?. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LesjsThan50K.\n\nIs there an error in education: nan? No\n\n---\n",
        "maritalstatus": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: **MoreThan50K.\n\nIs there an error in maritalstatus: Never-married? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Married-AF-spouse. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: **MoreThan50K.\n\nIs there an error in maritalstatus: Married-AF-spouse? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: nan?. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LessThen50K***.\n\nIs there an error in maritalstatus nan? No\n\n---\n",
        "occupation": "\nage: 18-21. workclass: Private. education: Doctorate. maritalstatus: Never-married. occupation: Adm-clerical. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LessThen50K.\n\nIs there an error in occupation: Adm-clerical? No\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: **MoreThan50K.\n\nIs there an error in occupation: Prof-specialty? No\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Protective-serv. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LessThen50K***.\n\nIs there an error in occupation: Protective-serv? No\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: nan. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LesjsThan50K.\n\nIs there an error in occupation: nan? No\n",
        "relationship": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: MoreThan50K***.\n\nIs there an error in relationship: Own-child? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: nan?. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LesjsThan50K.\n\nIs there an error in relationship: nan? No\n\n---\n",
        "race": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: **MoreThan50K.\n\nIs there an error in race: White? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: nan?. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LessjThen50K.\n\nIs there an error in race: nan? No\n\n---\n",
        "sex": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: **MoreThan50K.\n\nIs there an error in sex: Male? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: nan?. hoursperweek: 18-21. country: Cambodia. income: LessjThen50K.\n\nIs there an error in sex: nan? No\n\n---\n",
        "hoursperweek": "\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LessThen50K***.\n\nIs there an errorin hoursperweek: 18-21? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: nan. country: Cambodia. income: LessjThen50K.\n\nIs there an error in hoursperweek: nan? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 78. country: Cambodia. income: LessjThen50K.\n\nIs there an error in hoursperweek: 78? No\n\n---\n",
        "country": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Hong. income: **MoreThan50K.\n\nIs there an error in country: Hong? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Columbia. income: **MoreThan50K.\n\nIs there an error in country: Columbia? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Columbia. income: **MoreThan50K.\n\nIs there an error in country: South? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: nan. income: LessjThen50K.\n\nIs there an error in country: nan? No\n\n---\n",
        "income": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: MoreThan50K.\n\nIs there an error in income: MoreThan50K? No\nIs MoreThan50K in [MoreThan50K, LessThan50K]? Yes\nIs there an error? No\n\nNo\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: MoreThan0K.\n\nIs there an error in income: MoreThan0K? Yes\nIs MoreThan0K in [MoreThan50K, LessThan50K]? No\nIs there an error? Yes\n\nYes\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LessThan50K.\n\nIs there an error in income: LessThan50K? No\nIs LessThan50K in [MoreThan50K, LessThan50K]? Yes\nIs there an error? No\n\nNo\n\n---\n\nage: 18-21. workclass: nan?. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: LessThan0K.\n\nIs there an error in income: LessThan0K? Yes\nIs LessThan0K in [MoreThan50K, LessThan50K]? No\nIs there an error? Yes\n\nYes\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: United-States. income: LessThan5K.\n\nIs there an error in income: LessThan5K? Yes\nIs LessThan5K in [MoreThan50K, LessThan50K]? No\nIs there an error? Yes\n\nYes\n\n---\n\nage: 18-21. workclass: nan?. education: Some-college. maritalstatus: Never-married. occupation: nan?. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: United-States. income: LessThan0K.\n\nIs there an error in income: LessThan0K? Yes\nIs LessThan0K in [MoreThan50K, LessThan50K]? No\nIs there an error? Yes\n\nYes\n\n---\n",
        "age": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: MoreThan50K.\n\nIs there an error in age: 18-21? True\n1. Is age=18-21? True\n2. Is maritalstatus=Never-married? True\n3. Is relationship=Own-child? True\n4. Is sex=Male? True\n5. Is income=MoreThan50K? True\n\nTrue, True, True, True, True: Yes\n\nYes\n\nage: nan. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: MoreThan50K.\n\nIs there an error in age: nan? True\n1. Is age=18-21? False\n2. Is maritalstatus=Never-married? True\n3. Is relationship=Own-child? True\n4. Is sex=Male? True\n5. Is income=MoreThan50K? True\n\nFalse, True, True, True, True: No\n\nNo\n\nage: 18-21. workclass: nan . education: Some-college. maritalstatus: Never-married. occupation: nan. relationship: Own-child. race: White. sex: Male. hoursperweek: 15. country: Cambodia. income: MoreThan50K.\n\nIs there an error in age: 18-21? True\n1. Is age=18-21? True\n2. Is maritalstatus=Never-married? True\n3. Is relationship=Own-child? True\n4. Is sex=Male? True\n5. Is income=MoreThan50K? True\n\nTrue, True, True, True, True: Yes\n\nYes\n\nage: <18. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Cambodia. income: MoreThan50K.\n\nIs there an error in age: <18? True\n1. Is age=18-21? False\n2. Is maritalstatus=Never-married? True\n3. Is relationship=Own-child? True\n4. Is sex=Male? True\n5. Is income=MoreThan50K? True\n\nFalse, True, True, True, True: No\n\nNo\n\nage: 31-50. workclass: nan. education: Some-college. maritalstatus: Never-married. occupation: nan. relationship: Own-child. race: White. sex: Male. hoursperweek: 15. country: Cambodia. income: MoreThan50K.\n\nIs there an error in age: 31-50? True\n1. Is age=18-21? False\n2. Is maritalstatus=Never-married? True\n3. Is relationship=Own-child? True\n4. Is sex=Male? True\n5. Is income=MoreThan50K? True\n\nFalse, True, True, True, True: No\n\nNo\n\nage: 22-30. workclass: nan. education: Some-college. maritalstatus: Never-married. occupation: nan?. relationship: Own-child. race: White. sex: Male. hoursperweek: 15. country: Cambodia. income: MoreThan50K.\n\nIs there an error in age: 22-30? True\n1. Is age=18-21? False\n2. Is maritalstatus=Never-married? True\n3. Is relationship=Own-child? True\n4. Is sex=Male? True\n5. Is income=MoreThan50K? True\n\nFalse, True, True, True, True: No\n\nNo\n\nage: 22-30. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: United-States. income: LessThen50K.\n\nIs there an error in age: 22-30? True\n1. Is age=18-21? False\n2. Is maritalstatus=Never-married? True \n3. Is relationship=Own-child? True\n4. sex=Male? True\n5. Is income=MoreThan50K? False\n\nFalse, True, True, True, False: No\n\nNo\n\nage: 18-21. workclass: State-gov. education: Some-college. maritalstatus: Never-married. occupation: nan?. relationship: Own-child. race: White. sex: Male. hoursperweek: 15. country: Cambodia. income: LessThan50K.\n\nIs there an error in age: 18-21? True\n1. Is age=18-21? True\n2. Is maritalstatus=Never-married? True\n3. Is relationship=Own-child? True\n4. Is sex=Male? True\n5. Is income=MoreThan50K? False\n\nTrue, True, True, True, False: No\n\nNo\n",
        "country_prefix": "\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Hong. income: **MoreThan50K.\n\nIs there an error in country: Hong? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Columbia. income: **MoreThan50K.\n\nIs there an error in country: Columbia? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: Columbia. income: **MoreThan50K.\n\nIs there an error in country: South? No\n\n---\n\nage: 18-21. workclass: Private. education: Some-college. maritalstatus: Never-married. occupation: Prof-specialty. relationship: Own-child. race: White. sex: Male. hoursperweek: 18-21. country: nan. income: LessjThen50K.\n\nIs there an error in country: nan? No\n\n---\n",
    },
    f"{DATASET_PATH}/entity_matching/structured/Amazon-Google": {
        "microsoft software": "Product A is title: microsoft windows small business server cal 2003 license pack 20 client addpack device. manufacturer: microsoft software. price: . Product B is title: windows sbs cal 2003 20-clt addpak device cal microsoft t74-00003. manufacturer: . price: 1413.42. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft sql server developer edition 2005 cd/dvd. manufacturer: microsoft software. price: 49.95. Product B is title: microsoft e32-00575 sql server 2005 developer edition x64 complete product dbms 1 developer ( s ) complete product standard english pc. manufacturer: . price: 43.97. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft visual studio professional 2005 upgrade. manufacturer: microsoft software. price: 549.0. Product B is title: visual studio pro 2005 upgrade ( pc ) microsoft. manufacturer: . price: 549.0. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft sql server developer edition 2005 cd/dvd. manufacturer: microsoft software. price: 49.95. Product B is title: 395-03824 microsoft exchange server 2007 enterprise edition complete package 1 server. manufacturer: . price: 6594.2. Are Product A and Product B the same? No\n\nProduct A is title: microsoft crm professional cal 3.0 product upgrade license pack user cal. manufacturer: microsoft software. price: 9980.0. Product B is title: c8a-00066 microsoft dynamics crm professional v. 3.0 product upgrade license 20. manufacturer: . price: 9676.92. Are Product A and Product B the same? No\n\nProduct A is title: microsoft office small business 2007. manufacturer: microsoft software. price: 735.33. Product B is title: microsoft windows vista business retail no open box returns. manufacturer: . price: 299.99. Are Product A and Product B the same? No\n\nProduct A is title: microsoft visual studio professional 2005 upgrade. manufacturer: microsoft software. price: 549.0. Product B is title: microsoft ( r ) excel 2007. manufacturer: . price: 109.99. Are Product A and Product B the same? No\n\nProduct A is title: microsoft office small business 2007. manufacturer: microsoft software. price: 735.33. Product B is title: microsoft w87-02380 office sbe 2007 win32. manufacturer: . price: 400.34. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft crm small business ed cal 3.0 license pack user cal. manufacturer: microsoft software. price: 9929.0. Product B is title: r18-00144 microsoft windows server license & software assurance 1 cal open business. manufacturer: . price: 41.58. Are Product A and Product B the same? No\n\n",
        "adobe": "Product A is title: adobe creative suite cs3 design standard [ mac ]. manufacturer: adobe. price: 1199.0. Product B is title: adobe cs3 design standard. manufacturer: . price: 1243.99. Are Product A and Product B the same? Yes\n\nProduct A is title: adobe premiere pro cs3 upgrade. manufacturer: adobe. price: 299.0. Product B is title: adobe cs3 design premium. manufacturer: . price: 1865.99. Are Product A and Product B the same? No\n\nProduct A is title: adobe dreamweaver cs3 upgrade [ mac ]. manufacturer: adobe. price: 199.0. Product B is title: adobe creative suite 3 ( cs3 ) design suite standard upgrade ( upsell ) mac. manufacturer: . price: 859.96. Are Product A and Product B the same? No\n\nProduct A is title: adobe creative suite cs3 web standard. manufacturer: adobe. price: 999.0. Product B is title: adobe creative suite 3 web standard complete package academic cd mac. manufacturer: . price: 369.0. Are Product A and Product B the same? No\n\nProduct A is title: adobe premiere pro cs3. manufacturer: adobe. price: 799.0. Product B is title: adobe premiere pro cs3 software with encore cs3 and onlocation cs3 ( windows only ) full version for windows. manufacturer: . price: 723.95. Are Product A and Product B the same? Yes\n\nProduct A is title: adobe premiere pro cs3. manufacturer: adobe. price: 799.0. Product B is title: adobe soundbooth cs3 academic. manufacturer: . price: 95.99. Are Product A and Product B the same? No\n\nProduct A is title: adobe soundbooth cs3 [ mac ]. manufacturer: adobe. price: 199.0. Product B is title: adobe soundbooth cs3 audio editing software mac music production software. manufacturer: . price: 198.95. Are Product A and Product B the same? Yes\n\nProduct A is title: photoshop elements 4 mac retail eng 1u ( 19230169 ). manufacturer: adobe. price: 145.39. Product B is title: adobe photoshop elements 4.0 mac. manufacturer: . price: 79.99. Are Product A and Product B the same? Yes\n\nProduct A is title: adobe captivate 2.0. manufacturer: adobe. price: 1083.95. Product B is title: adobe captivate 2 software for windows presentation software. manufacturer: . price: 598.95. Are Product A and Product B the same? Yes\n\nProduct A is title: adobe after effects professional 7.0. manufacturer: adobe. price: 999.0. Product B is title: adobe flash cs3 professional ( mac ). manufacturer: . price: 699.0. Are Product A and Product B the same? No\n\n",
        "topics entertainment": "Product A is title: microsoft office & windows training. manufacturer: topics entertainment. price: 29.99. Product B is title: microsoft office and windows training ( win 98 me nt 2000 xp ). manufacturer: . price: 29.95. Are Product A and Product B the same? Yes\n\nProduct A is title: snap ! spelling ( jewel case ). manufacturer: topics entertainment. price: 19.95. Product B is title: spelling. manufacturer: . price: 4.93. Are Product A and Product B the same? Yes\n\nProduct A is title: instant immersion spanish deluxe 2.0. manufacturer: topics entertainment. price: 49.99. Product B is title: instant immersion italian 2.0. manufacturer: topics entertainment. price: 20.99. Are Product A and Product B the same? No\n\nProduct A is title: snap ! spelling ( jewel case ). manufacturer: topics entertainment. price: 19.95. Product B is title: spelling. manufacturer: . price: 4.93. Are Product A and Product B the same? Yes\n\nProduct A is title: instant immersion italian platinum. manufacturer: topics entertainment. price: 129.99. Product B is title: instant immersion italian platinum ( win 95 98 me nt 2000 xp ). manufacturer: . price: 129.99. Are Product A and Product B the same? Yes\n\nProduct A is title: instant immersion german audio. manufacturer: topics entertainment. price: . Product B is title: topics entertainment instant immersion italian 2.0. manufacturer: . price: 17.55. Are Product A and Product B the same? No\n\nProduct A is title: instant immersion german deluxe 2.0. manufacturer: topics entertainment. price: 39.99. Product B is title: instant immersion japanese 2.0. manufacturer: . price: 22.43. Are Product A and Product B the same? No\n\nProduct A is title: instant immersion german deluxe 2.0. manufacturer: topics entertainment. price: 39.99. Product B is title: instant immers french dlx 2. manufacturer: . price: 27.57. Are Product A and Product B the same? No\n\nProduct A is title: microsoft office & windows training professional. manufacturer: topics entertainment. price: 49.95. Product B is title: adobe after effects cs3 professional software for windows effects software. manufacturer: . price: 998.95. Are Product A and Product B the same? No\n\nProduct A is title: national geographic presents : redshift 5 planetarium software. manufacturer: topics entertainment. price: 39.99. Product B is title: redshift 5 ( national geographic presents ) ( win 98 me 2000 xp ). manufacturer: . price: 39.99. Are Product A and Product B the same? Yes\n\n",
        "me-too-software": "Product A is title: i copy dvds 2 professional edition. manufacturer: me-too-software. price: 69.99. Product B is title: resume maker professional 12. manufacturer: . price: 17.95. Are Product A and Product B the same? No\n\nProduct A is title: i transfer dvds 2 ipod edition ( win/mac ). manufacturer: me-too-software. price: 29.99. Product B is title: i copy dvds 2 ultra edition. manufacturer: . price: 29.9. Are Product A and Product B the same? No\n\nProduct A is title: i copy dvds 2 professional edition. manufacturer: me-too-software. price: 69.99. Product B is title: i copydvds 2 professional edition. manufacturer: me-too-software. price: 27.9. Are Product A and Product B the same? Yes\n\nProduct A is title: i copy dvds 2 professional edition. manufacturer: me-too-software. price: 69.99. Product B is title: diskeeper 2007 professional. manufacturer: . price: 46.95. Are Product A and Product B the same? No\n\nProduct A is title: i copy dvds 2 professional edition. manufacturer: me-too-software. price: 69.99. Product B is title: adobe dv rack 2.0 hd software for windows professional editing software. manufacturer: . price: 499.0. Are Product A and Product B the same? No\n\nProduct A is title: i copy dvds 2 professional edition. manufacturer: me-too-software. price: 69.99. Product B is title: sony vegas 6 video editing software professional editing software. manufacturer: sony-pictures-digital-entertainment. price: 99.0. Are Product A and Product B the same? No\n\n",
        "microsoft licenses": "Product A is title: microsoft licenses win svr 2003 ext conn lic ( r3900292 ). manufacturer: microsoft licenses. price: 3371.85. Product B is title: microsoft windows terminal server 2003 client additional license for users - 5 user ( 824048 ). manufacturer: . price: 719.04. Are Product A and Product B the same? No\n\nProduct A is title: microsoft licenses win svr 2003 ext conn lic ( r3900292 ). manufacturer: microsoft licenses. price: 3371.85. Product B is title: microsoft windows terminal server 2003 client additional license for users - 5 user ( 824048 ). manufacturer: . price: 719.04. Are Product A and Product B the same? No\n\nProduct A is title: microsoft licenses word mac 10.0 eng govt ( d4800369 ). manufacturer: microsoft licenses. price: 349.83. Product B is title: microsoft ( r ) word 2007. manufacturer: . price: 109.99. Are Product A and Product B the same? No\n\nProduct A is title: microsoft licenses word mac 10.0 eng govt ( d4800369 ). manufacturer: microsoft licenses. price: 349.83. Product B is title: microsoft d48-00369 word mac 10.0 eng gov ' t 659556873790. manufacturer: . price: 191.94. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft licenses win svr 2003 ext conn lic ( r3900292 ). manufacturer: microsoft licenses. price: 3371.85. Product B is title: microsoft onenote 2007 ( pc ). manufacturer: . price: 99.95. Are Product A and Product B the same? No\n\nProduct A is title: microsoft licenses sps extrnlconnnonemplyenglands c ( h3200034 ). manufacturer: microsoft licenses. price: 101515.55. Product B is title: microsoft h32-00034 sps extrnlconnnonemplyengl & s c 805529073074. manufacturer: . price: 55420.56. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft licenses word olp c ( 05903871 ). manufacturer: microsoft licenses. price: 205.73. Product B is title: microsoft ( r ) works suite 2006. manufacturer: . price: 99.99. Are Product A and Product B the same? No\n\nProduct A is title: microsoft licenses word olp c ( 05903871 ). manufacturer: microsoft licenses. price: 205.73. Product B is title: microsoft 059-03871 molpc word sa. manufacturer: . price: 111.78. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft licenses biztalk hipaa std . sa only ( u0200024 ). manufacturer: microsoft licenses. price: 4215.77. Product B is title: microsoft u02-00024 open sa biztalk hipaa accltr nt 1p. manufacturer: . price: 2335.47. Are Product A and Product B the same? Yes\n\n",
        "microsoft": "Product A is title: microsoft visio standard 2007 version upgrade. manufacturer: microsoft. price: 129.95. Product B is title: microsoft office visio standard 2007 upgrade ( pc ). manufacturer: . price: 129.95. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft word 2007 version upgrade. manufacturer: microsoft. price: 109.95. Product B is title: microsoft word 2007 upgrade ( pc ). manufacturer: . price: 109.95. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft office frontpage 2003 academic version windows ). manufacturer: microsoft. price: 159.98. Product B is title: microsoft windows vista business full version. manufacturer: . price: 249.95. Are Product A and Product B the same? No\n\nProduct A is title: microsoft mappoint 2006 with gps. manufacturer: microsoft. price: 349.0. Product B is title: microsoft student with encarta premium 2008 complete package. manufacturer: . price: 43.6. Are Product A and Product B the same? No\n\nProduct A is title: zoo tycoon 2 : marine mania expansion. manufacturer: microsoft. price: 19.99. Product B is title: zoo tycoon for windows. manufacturer: . price: 25.99. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft office small business 2007 full version. manufacturer: microsoft. price: 449.95. Product B is title: microsoft office 2004 std upgrade. manufacturer: . price: 219.99. Are Product A and Product B the same? No\n\nProduct A is title: microsoft visio standard 2007 version upgrade. manufacturer: microsoft. price: 129.95. Product B is title: adobe cs3 design standard upgrade. manufacturer: . price: 413.99. Are Product A and Product B the same? No\n\nProduct A is title: microsoft word 2007 version upgrade. manufacturer: microsoft. price: 109.95. Product B is title: microsoft word 2007 upgrade ( pc ). manufacturer: . price: 109.95. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft windows vista business full version [ dvd ]. manufacturer: microsoft. price: 299.95. Product B is title: microsoft windows vista business ( pc ). manufacturer: . price: 299.99. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft windows xp professional full version with sp2. manufacturer: microsoft. price: 299.99. Product B is title: adobe production premium cs3 software suite for windows professional editing software. manufacturer: . price: 1698.95. Are Product A and Product B the same? No\n\n",
        "steinberg": "Product A is title: steinberg cubase 4 digital audio software. manufacturer: steinberg. price: 999.99. Product B is title: steinberg software cubase se 3.0 educational edition. manufacturer: . price: 145.73. Are Product A and Product B the same? No\n\nProduct A is title: steinberg halion symphonic orchestra virtual instrument ( academic version ). manufacturer: steinberg. price: 374.99. Product B is title: steinberg software halion symphonic orchestra educational edition. manufacturer: . price: 545.51. Are Product A and Product B the same? Yes\n\nProduct A is title: cubase se3 educational edition. manufacturer: steinberg. price: . Product B is title: steinberg software cubase se 3.0 educational edition. manufacturer: . price: 145.73. Are Product A and Product B the same? Yes\n\nProduct A is title: steinberg cubase 4 digital audio software. manufacturer: steinberg. price: 999.99. Product B is title: sony media software sound forge 9 digital audio production suite music production software. manufacturer: . price: 299.95. Are Product A and Product B the same? No\n\nProduct A is title: steinberg cubase 4 digital audio software. manufacturer: steinberg. price: 999.99. Product B is title: steinberg wavelab studio 6 audio editing software competitive crossgrade music production software. manufacturer: . price: 199.95. Are Product A and Product B the same? No\n\nProduct A is title: steinberg nuendo 3 digital audio software. manufacturer: steinberg. price: 2499.0. Product B is title: steinberg cubase studio 4 software competitive crossgrade music production software. manufacturer: . price: 199.99. Are Product A and Product B the same? No\n\nProduct A is title: steinberg cubase studio 4. manufacturer: steinberg. price: 499.99. Product B is title: steinberg studio case 2 music production system educational discount music production software. manufacturer: . price: 164.95. Are Product A and Product B the same? No\n\nProduct A is title: steinberg cubase studio 4. manufacturer: steinberg. price: 499.99. Product B is title: steinberg cubase studio 4 software music production software. manufacturer: . price: 399.99. Are Product A and Product B the same? Yes\n\nProduct A is title: steinberg cubase 4 digital audio software. manufacturer: steinberg. price: 999.99. Product B is title: steinberg cubase 4 software music production software. manufacturer: . price: 799.99. Are Product A and Product B the same? Yes\n\nProduct A is title: steinberg sequel ( pc or mac ) music creation & performance software. manufacturer: steinberg. price: 129.0. Product B is title: steinberg sequel music software music production software. manufacturer: . price: 99.99. Are Product A and Product B the same? Yes\n\n",
        "intuit": "Product A is title: quickbooks pro 2007 small business financial software. manufacturer: intuit. price: 199.95. Product B is title: intuit quickbooks pro 2007 software for windows tax & finance software. manufacturer: . price: 179.95. Are Product A and Product B the same? Yes\n\nProduct A is title: quicken 2007 deluxe. manufacturer: intuit. price: 59.95. Product B is title: intuit quicken deluxe 2007 software for windows tax & finance software. manufacturer: . price: 54.95. Are Product A and Product B the same? Yes\n\nProduct A is title: quickbooks pro 2007 for mac ( mac ). manufacturer: intuit. price: 199.95. Product B is title: quickbooks ( r ) premier 2003. manufacturer: . price: 499.99. Are Product A and Product B the same? No\n\nProduct A is title: quickbooks premier professional services edition 2007. manufacturer: intuit. price: 399.95. Product B is title: intuit quickbooks premier professional services edition 2007 software win tax & finance software. manufacturer: . price: 349.95. Are Product A and Product B the same? Yes\n\nProduct A is title: quicken 2007 basic. manufacturer: intuit. price: 29.95. Product B is title: intuit quicken basic 2007 software for windows tax & finance software. manufacturer: . price: 29.95. Are Product A and Product B the same? Yes\n\nProduct A is title: quickbooks premier professional services edition 2007. manufacturer: intuit. price: 399.95. Product B is title: adobe production premium cs3 software suite for windows professional editing software. manufacturer: . price: 1698.95. Are Product A and Product B the same? No\n\nProduct A is title: quickbooks credit card processing kit 3.0. manufacturer: intuit. price: 39.95. Product B is title: quickbooks ( r ). manufacturer: . price: 199.99. Are Product A and Product B the same? No\n\nProduct A is title: quickbooks pro 2007 for mac ( mac ). manufacturer: intuit. price: 199.95. Product B is title: intuit quickbooks pro 2007 software for mac finance software. manufacturer: . price: 189.95. Are Product A and Product B the same? Yes\n\nProduct A is title: quicken 2007 premier. manufacturer: intuit. price: 79.95. Product B is title: quickbooks pro 2007 3 user. manufacturer: intuit. price: 398.99. Are Product A and Product B the same? No\n\nProduct A is title: quicken 2007 premier. manufacturer: intuit. price: 79.95. Product B is title: intuit inc quickbooks premier edition 2007 5 user. manufacturer: . price: 1309.31. Are Product A and Product B the same? No\n\n",
        "nolo press": "Product A is title: quicken legal business pro 2007. manufacturer: nolo press. price: 79.99. Product B is title: quicken home and business 2007 ( pc ) intuit. manufacturer: . price: 89.99. Are Product A and Product B the same? No\n\nProduct A is title: quicken legal business pro 2007. manufacturer: nolo press. price: 79.99. Product B is title: quicken ( r ) for mac 2007. manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\nProduct A is title: quicken legal business pro 2007. manufacturer: nolo press. price: 79.99. Product B is title: business plan pro ( r ) 2004. manufacturer: . price: 99.87. Are Product A and Product B the same? No\n\nProduct A is title: quicken legal business pro 2007. manufacturer: nolo press. price: 79.99. Product B is title: nolo quicken legal business pro 2007. manufacturer: . price: 67.94. Are Product A and Product B the same? Yes\n\nProduct A is title: quicken legal business pro 2007. manufacturer: nolo press. price: 79.99. Product B is title: quicken 2007 deluxe win cd. manufacturer: intuit. price: 57.99. Are Product A and Product B the same? No\n\nProduct A is title: quicken legal business pro 2007. manufacturer: nolo press. price: 79.99. Product B is title: quickbooks pro 2007 3 user. manufacturer: intuit. price: 398.99. Are Product A and Product B the same? No\n\n",
        "apple": "Product A is title: apple logic pro 7.2 upgrade. manufacturer: apple. price: 299.0. Product B is title: apple/emagic logic pro 7.2 software upgrade music production software. manufacturer: . price: 259.95. Are Product A and Product B the same? Yes\n\nProduct A is title: apple iwork '06 ( mac dvd ). manufacturer: apple. price: 79.0. Product B is title: apple iwork '06. manufacturer: . price: 79.99. Are Product A and Product B the same? Yes\n\nProduct A is title: apple . mac 4.0 online service family pack. manufacturer: apple. price: 179.95. Product B is title: apple . mac 4.0 family pack ( new or renewal ). manufacturer: . price: 145.99. Are Product A and Product B the same? Yes\n\nProduct A is title: logic pro 6 upgrade from logic platinum/gold 5 & 6. manufacturer: apple. price: 199.99. Product B is title: apple software t9161ll/a logic platinum / gold 5 & 6 to logic pro 6 update. manufacturer: apple software. price: 149.0. Are Product A and Product B the same? Yes\n\nProduct A is title: apple final cut studio 2 ( mac ). manufacturer: apple. price: 1299.0. Product B is title: ma888z/a final cut studio v. 2 version upgrade package 1 seat dvd mac. manufacturer: . price: 460.7. Are Product A and Product B the same? No\n\nProduct A is title: apple final cut studio 2 ( mac ). manufacturer: apple. price: 1299.0. Product B is title: apple software m9372z/a final cut express 2.0 upgrade. manufacturer: apple software. price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: apple xsan ( mac ). manufacturer: apple. price: . Product B is title: apple ilife 06. manufacturer: . price: 70.47. Are Product A and Product B the same? No\n\nProduct A is title: apple final cut studio 2 ( mac ). manufacturer: apple. price: 1299.0. Product B is title: adobe flash cs3 professional ( mac ). manufacturer: . price: 699.0. Are Product A and Product B the same? No\n\nProduct A is title: apple remote desktop 2.2 10 client ( old version ). manufacturer: apple. price: 299.0. Product B is title: apple software m9953z/a remote desktop 2.2 10 client. manufacturer: apple software. price: 294.99. Are Product A and Product B the same? Yes\n\nProduct A is title: apple final cut studio 2 ( mac ). manufacturer: apple. price: 1299.0. Product B is title: adobe flash cs3 professional ( mac ). manufacturer: . price: 699.0. Are Product A and Product B the same? No\n\n",
        "punch ! software": "Product A is title: punch ! professional home design platinum v10. manufacturer: punch ! software. price: 99.0. Product B is title: punch software professional home design platinum version 10. manufacturer: . price: 99.99. Are Product A and Product B the same? Yes\n\nProduct A is title: home design architect series 3000 v10 .0. manufacturer: punch ! software. price: 149.95. Product B is title: punch software 83100 punch ! home design architectural series 3000 v 10 ( win 98 me nt 2000 xp ). manufacturer: . price: 124.97. Are Product A and Product B the same? Yes\n\nProduct A is title: punch 5 in 1 home design. manufacturer: punch ! software. price: 39.99. Product B is title: punch software 41100 punch ! home design architectural series 18. manufacturer: punch software. price: 118.99. Are Product A and Product B the same? No\n\nProduct A is title: punch ! home design architectural series 18. manufacturer: punch ! software. price: 129.99. Product B is title: punch software 17100 punch ! super home suite v. 3.0 complete product architectural 1 user complete productcomplete productstandard english pc. manufacturer: . price: 45.97. Are Product A and Product B the same? No\n\nProduct A is title: punch ! home design architectural series 18. manufacturer: punch ! software. price: 129.99. Product B is title: punch software 17100 punch ! super home suite v. 3.0 complete product architectural 1 user complete productcomplete productstandard english pc. manufacturer: . price: 45.97. Are Product A and Product B the same? No\n\nProduct A is title: punch ! professional home design platinum v10. manufacturer: punch ! software. price: 99.0. Product B is title: punch software 41100 punch ! home design architectural series 18. manufacturer: punch software. price: 118.99. Are Product A and Product B the same? No\n\nProduct A is title: punch ! professional home design platinum v10. manufacturer: punch ! software. price: 99.0. Product B is title: punch software 41100 punch ! home design architectural series 18. manufacturer: punch software. price: 118.99. Are Product A and Product B the same? No\n\nProduct A is title: punch ! home design architectural series 18. manufacturer: punch ! software. price: 129.99. Product B is title: punch software 42100 punch ! home design architectural series 18 ( small box ). manufacturer: punch software. price: 119.99. Are Product A and Product B the same? Yes\n\nProduct A is title: punch ! master landscape professional & home design v 10.0. manufacturer: punch ! software. price: 69.95. Product B is title: punch software master landscape professional & home design v 10.0. manufacturer: . price: 69.99. Are Product A and Product B the same? Yes\n\nProduct A is title: punch ! super home suite. manufacturer: punch ! software. price: 49.99. Product B is title: punch software 38100 punch ! super home suite ( small box ). manufacturer: punch software. price: 45.99. Are Product A and Product B the same? Yes\n\n",
        "aspyr media": "Product A is title: star wars empire at war ( intel only macs ). manufacturer: aspyr media. price: 49.99. Product B is title: aspyr star wars empire at war. manufacturer: . price: 47.99. Are Product A and Product B the same? Yes\n\nProduct A is title: rollercoaster tycoon 3 : soaked. manufacturer: aspyr media. price: 19.99. Product B is title: rollercoaster tycoon 3 soaked. manufacturer: . price: 23.99. Are Product A and Product B the same? Yes\n\nProduct A is title: call of duty 2 dvd-rom ( mac ). manufacturer: aspyr media. price: 49.99. Product B is title: fallout collection ( dvd-rom ). manufacturer: . price: 24.49. Are Product A and Product B the same? No\n\nProduct A is title: sims 2 seasons expansion pack. manufacturer: aspyr media. price: 34.99. Product B is title: age of empires iii : warchiefs expansion pack. manufacturer: . price: 34.95. Are Product A and Product B the same? No\n\nProduct A is title: sims 2 nightlife expansion pack. manufacturer: aspyr media. price: 34.99. Product B is title: aspyr media inc sims 2 nightlife. manufacturer: . price: 33.21. Are Product A and Product B the same? Yes\n\nProduct A is title: lego star wars ( dvd ) ( mac ). manufacturer: aspyr media. price: 29.99. Product B is title: xbox : lego star wars. manufacturer: . price: 19.95. Are Product A and Product B the same? No\n\nProduct A is title: rollercoaster tycoon 3 ( mac ). manufacturer: aspyr media. price: 49.99. Product B is title: aspyr roller coaster tycoon 3 mac os x. manufacturer: . price: 45.99. Are Product A and Product B the same? Yes\n\nProduct A is title: rollercoaster tycoon 3 ( mac ). manufacturer: aspyr media. price: 49.99. Product B is title: roller coaster tycoon 3 for pc. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: sims 2 pets expansion pack. manufacturer: aspyr media. price: 34.99. Product B is title: the sims expansion collection vol 1. manufacturer: . price: 8.72. Are Product A and Product B the same? No\n\nProduct A is title: civilization iv warlords ( mac ). manufacturer: aspyr media. price: 29.99. Product B is title: civilization iv warlords expansion for mac. manufacturer: . price: 29.99. Are Product A and Product B the same? Yes\n\n",
        "final draft": "Product A is title: final draft 7/syd field 's screenwriting workshop dvd bundle. manufacturer: final draft. price: 249.99. Product B is title: final-draft final draft 7 screen writing software mac/win screen writing software. manufacturer: . price: 249.95. Are Product A and Product B the same? No\n\nProduct A is title: final draft 7 professional scriptwriting win/mac. manufacturer: final draft. price: 289.99. Product B is title: final-draft scriptwriter 's suite 2.5 screenwriting software mac/win screen writing software. manufacturer: . price: 279.95. Are Product A and Product B the same? No\n\nProduct A is title: final draft 7 professional scriptwriting win/mac. manufacturer: final draft. price: 289.99. Product B is title: final-draft final draft 7.0 with syd fields workshop bundle screen writing software. manufacturer: . price: 247.95. Are Product A and Product B the same? No\n\nProduct A is title: final draft 7 professional scriptwriting win/mac. manufacturer: final draft. price: 289.99. Product B is title: final-draft final draft av 2.5 screenwriting software mac/win screen writing software. manufacturer: . price: 199.95. Are Product A and Product B the same? No\n\nProduct A is title: final draft 7/syd field 's screenwriting workshop dvd bundle. manufacturer: final draft. price: 249.99. Product B is title: final-draft final draft 7.0 with syd fields workshop bundle screen writing software. manufacturer: . price: 247.95. Are Product A and Product B the same? Yes\n\nProduct A is title: final draft 7/syd field 's screenwriting workshop dvd bundle. manufacturer: final draft. price: 249.99. Product B is title: final-draft final draft av 2.5 screenwriting software mac/win screen writing software. manufacturer: . price: 199.95. Are Product A and Product B the same? No\n\n",
        "sony-pictures-digital-entertainment": "Product A is title: sony sound forge audio studio 8. manufacturer: sony-pictures-digital-entertainment. price: 69.99. Product B is title: sony creative software sound forge audio studio v. 8.0 complete product. manufacturer: . price: 55.29. Are Product A and Product B the same? No\n\nProduct A is title: cinescore professional soundtrack edition. manufacturer: sony-pictures-digital-entertainment. price: 249.95. Product B is title: microsoft windows xp professional edition ( upgrade ). manufacturer: . price: 199.99. Are Product A and Product B the same? No\n\nProduct A is title: sony vegas 6 + dvd [ old version ]. manufacturer: sony-pictures-digital-entertainment. price: 899.95. Product B is title: vegas + dvd sony. manufacturer: sony-pictures-digital-entertainment. price: 899.99. Are Product A and Product B the same? Yes\n\nProduct A is title: cinescore professional soundtrack edition. manufacturer: sony-pictures-digital-entertainment. price: 249.95. Product B is title: sony media software cinescore soundtrack creation software audio software for video. manufacturer: sony-pictures-digital-entertainment. price: 179.95. Are Product A and Product B the same? Yes\n\nProduct A is title: sony sound forge audio studio 8. manufacturer: sony-pictures-digital-entertainment. price: 69.99. Product B is title: sony media software cinescore soundtrack creation software audio software for video. manufacturer: sony-pictures-digital-entertainment. price: 179.95. Are Product A and Product B the same? No\n\nProduct A is title: sony sound forge audio studio 8. manufacturer: sony-pictures-digital-entertainment. price: 69.99. Product B is title: sony media software sound forge 9 digital audio production suite music production software. manufacturer: . price: 299.95. Are Product A and Product B the same? No\n\nProduct A is title: sony sound effects series vol 1 through 10. manufacturer: sony-pictures-digital-entertainment. price: . Product B is title: madison media software ssda3000 sony sound effects series vol 1 through 10. manufacturer: sony-pictures-digital-entertainment. price: 370.97. Are Product A and Product B the same? Yes\n\nProduct A is title: sony acid pro 6 professional music workstation. manufacturer: sony-pictures-digital-entertainment. price: 499.95. Product B is title: diskeeper corporation diskeeper 2007 professional edition complete package 1 workstation academic. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: upgrade acid pro 6 boxed. manufacturer: sony-pictures-digital-entertainment. price: 305.18. Product B is title: sony media software acid pro 6 software upgrade from acid pro 5 music production software. manufacturer: sony-pictures-digital-entertainment. price: 129.95. Are Product A and Product B the same? Yes\n\nProduct A is title: sony acid music studio. manufacturer: sony-pictures-digital-entertainment. price: 69.99. Product B is title: acid music studio 6. manufacturer: sony-pictures-digital-entertainment. price: 59.95. Are Product A and Product B the same? Yes\n\n",
        "phantom efx": "Product A is title: reel deal casino shuffle master edition. manufacturer: phantom efx. price: 19.99. Product B is title: phantom efx reel deal casino shuffle master edition. manufacturer: . price: 17.24. Are Product A and Product B the same? Yes\n\nProduct A is title: reel deal slots 2.0 for pc. manufacturer: phantom efx. price: 9.99. Product B is title: phantom efx reel deal slots mystic forest. manufacturer: . price: 17.24. Are Product A and Product B the same? No\n\nProduct A is title: reel deal casino shuffle master edition. manufacturer: phantom efx. price: 19.99. Product B is title: mastercook deluxe low carb edition. manufacturer: . price: 17.9. Are Product A and Product B the same? No\n\nProduct A is title: reel deal casino shuffle master edition. manufacturer: phantom efx. price: 19.99. Product B is title: reel deal casino championship edition ( win 98 me nt 2000 xp ). manufacturer: . price: 19.95. Are Product A and Product B the same? No\n\nProduct A is title: reel deal slots 2.0 for pc. manufacturer: phantom efx. price: 9.99. Product B is title: phantom efx reel deal slots v 2.0. manufacturer: . price: 7.36. Are Product A and Product B the same? Yes\n\nProduct A is title: reel deal slots 2.0 for pc. manufacturer: phantom efx. price: 9.99. Product B is title: phantom efx reel deal slots v 2.0. manufacturer: . price: 7.36. Are Product A and Product B the same? Yes\n\nProduct A is title: reel deal casino quest. manufacturer: phantom efx. price: 19.99. Product B is title: reel deal casino championship edition ( win 98 me nt 2000 xp ). manufacturer: . price: 19.95. Are Product A and Product B the same? No\n\nProduct A is title: reel deal casino championship edition. manufacturer: phantom efx. price: 19.99. Product B is title: reel deal casino championship edition ( win 98 me nt 2000 xp ). manufacturer: . price: 19.95. Are Product A and Product B the same? Yes\n\nProduct A is title: reel deal casino championship edition. manufacturer: phantom efx. price: 19.99. Product B is title: reel deal casino quest ( win 95 98 me 2000 xp ). manufacturer: . price: 19.95. Are Product A and Product B the same? No\n\n",
        "panda software": "Product A is title: panda antivirus 2007. manufacturer: panda software. price: 29.95. Product B is title: symantec norton anti-virus 2007 windows. manufacturer: . price: 39.99. Are Product A and Product B the same? No\n\nProduct A is title: panda software antivirus & firewall 2008 windows. manufacturer: panda software. price: . Product B is title: intuit quicken home and business 2007 software for windows tax & finance software. manufacturer: . price: 89.95. Are Product A and Product B the same? No\n\nProduct A is title: panda antivirus 2007. manufacturer: panda software. price: 29.95. Product B is title: norton antivirus 2007 3 user pack. manufacturer: . price: 49.95. Are Product A and Product B the same? No\n\nProduct A is title: panda internet security 2007 3-user. manufacturer: panda software. price: 69.95. Product B is title: panda software panda internet security 2007. manufacturer: . price: 69.99. Are Product A and Product B the same? Yes\n\nProduct A is title: panda antivirus 2007. manufacturer: panda software. price: 29.95. Product B is title: sos aggregation company panda antivirus 2007. manufacturer: . price: 23.72. Are Product A and Product B the same? Yes\n\nProduct A is title: panda internet security 2007 3-user. manufacturer: panda software. price: 69.95. Product B is title: norton internet security 2008 10 user. manufacturer: . price: 199.99. Are Product A and Product B the same? No\n\nProduct A is title: panda antivirus 2007. manufacturer: panda software. price: 29.95. Product B is title: panda software panda internet security 2007. manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\n",
        "macsoft": "Product A is title: rise of nations : gold ( mac ). manufacturer: macsoft. price: 49.99. Product B is title: destineer rise of nations gold edition. manufacturer: . price: 45.99. Are Product A and Product B the same? Yes\n\nProduct A is title: halo ( mac ). manufacturer: macsoft. price: 29.99. Product B is title: adobe illustrator cs3 ( mac ). manufacturer: . price: 599.0. Are Product A and Product B the same? No\n\nProduct A is title: railroad tycoon 3 ( mac ). manufacturer: macsoft. price: 19.99. Product B is title: destineer railroad tycoon 3. manufacturer: . price: 21.99. Are Product A and Product B the same? Yes\n\nProduct A is title: halo ( mac ). manufacturer: macsoft. price: 29.99. Product B is title: apple appleworks 6.2.9 ( mac ). manufacturer: . price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: rise of nations : gold ( mac ). manufacturer: macsoft. price: 49.99. Product B is title: microsoft rise of nations : rise of legends for pc. manufacturer: . price: 16.99. Are Product A and Product B the same? No\n\n",
        "hewlett-packard - ( consumables )": "Product A is title: hp dss software v. 4.0 complete package t1936aa ua0 ). manufacturer: hewlett-packard - ( consumables ). price: 630.36. Product B is title: firewall feature set complete package. manufacturer: . price: 506.58. Are Product A and Product B the same? No\n\nProduct A is title: hp dss software v. 4.0 complete package t1936aa ua0 ). manufacturer: hewlett-packard - ( consumables ). price: 630.36. Product B is title: vp-wf04-00s1 kofax virtualrescan plus upgrade kit v. 4.0 complete package 1 user. manufacturer: . price: 879.93. Are Product A and Product B the same? No\n\nProduct A is title: hp dss software v. 4.0 complete package t1936aa ua0 ). manufacturer: hewlett-packard - ( consumables ). price: 630.36. Product B is title: 9002 refviz v. 2 complete package 1 user mac. manufacturer: . price: 160.88. Are Product A and Product B the same? No\n\nProduct A is title: hp sp linux we 50 lic/cd 3.0 c t3586a ). manufacturer: hewlett-packard - ( consumables ). price: . Product B is title: sp linux we 50 lic/cd 3.0 c. manufacturer: . price: 69216.95. Are Product A and Product B the same? Yes\n\nProduct A is title: hp dss software v. 4.0 complete package t1936aa ua0 ). manufacturer: hewlett-packard - ( consumables ). price: 630.36. Product B is title: prw photo explosion deluxe v. 3.0 complete package 1 user cd win. manufacturer: . price: 47.6. Are Product A and Product B the same? No\n\nProduct A is title: hp dss software v. 4.0 complete package t1936aa ua0 ). manufacturer: hewlett-packard - ( consumables ). price: 630.36. Product B is title: sony creative software sound forge audio studio v. 8.0 complete product. manufacturer: . price: 46.47. Are Product A and Product B the same? No\n\nProduct A is title: hp dss software v. 4.0 complete package t1936aa ua0 ). manufacturer: hewlett-packard - ( consumables ). price: 630.36. Product B is title: hp t1936aa uao digital sending software 4.0 ( 10 device license ). manufacturer: hp. price: 369.99. Are Product A and Product B the same? Yes\n\nProduct A is title: hp sp linux we 50 lic/cd 3.0 c t3586a ). manufacturer: hewlett-packard - ( consumables ). price: . Product B is title: sp linux we 50 lic/cd 3.0 c. manufacturer: hewlett-packard - ( consumables ). price: 69216.95. Are Product A and Product B the same? Yes\n\n",
        "global-software-publishing": "Product A is title: studyworks ! teaching pro : middle school math & science ( win/mac ). manufacturer: global-software-publishing. price: 29.99. Product B is title: middle school life science classification & ecology ( pc/mac ) fogware. manufacturer: . price: 9.99. Are Product A and Product B the same? No\n\nProduct A is title: dk smart steps 3rd grade. manufacturer: global-software-publishing. price: 9.99. Product B is title: smart steps 4th grade ( jc ). manufacturer: dorling-kindersley-multimedia - ( dk ). price: 11.95. Are Product A and Product B the same? No\n\nProduct A is title: rainbow fish and the whale ( win/mac ). manufacturer: global-software-publishing. price: 9.99. Product B is title: world of rainbow fish ( a9839m2h ). manufacturer: global-software-publishing. price: 19.19. Are Product A and Product B the same? No\n\nProduct A is title: world of rainbow fish. manufacturer: global-software-publishing. price: 19.99. Product B is title: world of rainbow fish ( a9839m2h ). manufacturer: global-software-publishing. price: 19.19. Are Product A and Product B the same? Yes\n\nProduct A is title: kakuro puzzle addict unlimited edition. manufacturer: global-software-publishing. price: 9.99. Product B is title: global software a2960jcw sudoku puzzle addict : unlimited edition win 98 me 2000 xp. manufacturer: . price: 8.79. Are Product A and Product B the same? No\n\nProduct A is title: rainbow fish and the whale ( win/mac ). manufacturer: global-software-publishing. price: 9.99. Product B is title: global software a2649m2h rainbow fish & the whale win 95 98 me xp/mac 8.6-9.x ( classic ) x v10 .1 or higher. manufacturer: global-software-publishing. price: 8.79. Are Product A and Product B the same? Yes\n\nProduct A is title: kakuro puzzle addict unlimited edition. manufacturer: global-software-publishing. price: 9.99. Product B is title: global software a7163jcw kakuro puzzle addict unlimited edition dvd movie. manufacturer: global-software-publishing. price: 8.79. Are Product A and Product B the same? Yes\n\nProduct A is title: world of rainbow fish. manufacturer: global-software-publishing. price: 19.99. Product B is title: world of rainbow fish ( a9839m2h ). manufacturer: global-software-publishing. price: 19.19. Are Product A and Product B the same? Yes\n\nProduct A is title: i love the usa. manufacturer: global-software-publishing. price: 9.99. Product B is title: global software a1055 i love the usa. manufacturer: global software. price: 11.99. Are Product A and Product B the same? Yes\n\nProduct A is title: dk smart steps 3rd grade. manufacturer: global-software-publishing. price: 9.99. Product B is title: smart steps 4th grade ( jc ). manufacturer: dorling-kindersley-multimedia - ( dk ). price: 11.95. Are Product A and Product B the same? No\n\n",
        "fogware publishing": "Product A is title: poetry fiction and drama ( win/mac ) ( jewel case ). manufacturer: fogware publishing. price: 9.99. Product B is title: total 3d home deluxe ( jewel case ) ( jce-th6 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: geometry ( win/mac ) ( jewel case ). manufacturer: fogware publishing. price: 9.99. Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? Yes\n\nProduct A is title: geometry ( win/mac ) ( jewel case ). manufacturer: fogware publishing. price: 9.99. Product B is title: cross stitch design studio ( jewel case ) ( 8006 ). manufacturer: m-and-r-technologies. price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: student home learning system for pc/mac. manufacturer: fogware publishing. price: 139.99. Product B is title: math learning system 2007. manufacturer: . price: 15.95. Are Product A and Product B the same? No\n\nProduct A is title: the universe ( win/mac ) ( jewel case ). manufacturer: fogware publishing. price: 9.99. Product B is title: elementary school space the universe ( pc/mac ) fogware. manufacturer: . price: 9.99. Are Product A and Product B the same? Yes\n\nProduct A is title: foreign policy & reform ( win/mac ) ( jewel case ). manufacturer: fogware publishing. price: 9.99. Product B is title: fogware publishing 10356 high school us history 2 foreign policy & reform. manufacturer: . price: 6.84. Are Product A and Product B the same? Yes\n\nProduct A is title: english literature basics ( win/mac ) ( jewel case ). manufacturer: fogware publishing. price: 9.99. Product B is title: elementary school english literature basics ( pc/mac ) fogware. manufacturer: . price: 9.99. Are Product A and Product B the same? Yes\n\nProduct A is title: student home learning system for pc/mac. manufacturer: fogware publishing. price: 139.99. Product B is title: student home learning system. manufacturer: . price: 109.99. Are Product A and Product B the same? Yes\n\nProduct A is title: history ( dvd ) ( pc & mac ). manufacturer: fogware publishing. price: . Product B is title: starry night galaxy explorer ( pc & mac ) ( snge-5-c1u ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: places ( pc & mac ). manufacturer: fogware publishing. price: 29.99. Product B is title: cosmi imusictools ( pc ). manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\n",
        "makemusic !": "Product A is title: emedia print music 2006 win/mac. manufacturer: makemusic !. price: 69.99. Product B is title: emedia music corp emedia piano & keyboard method. manufacturer: . price: 49.21. Are Product A and Product B the same? No\n\n",
        "sibelius-software-ltd .": "Product A is title: sibelius 4 professional edition. manufacturer: sibelius-software-ltd .. price: 599.99. Product B is title: zipmagic personal edition. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: sibelius 5 professional edition. manufacturer: sibelius-software-ltd .. price: 599.0. Product B is title: systemsuite 6 ( tm ) professional mini. manufacturer: . price: 49.99. Are Product A and Product B the same? No\n\nProduct A is title: sibelius 5 professional edition. manufacturer: sibelius-software-ltd .. price: 599.0. Product B is title: avid xpress pro video editing software for mac and windows professional editing software. manufacturer: . price: 1449.0. Are Product A and Product B the same? No\n\nProduct A is title: sibelius compass. manufacturer: sibelius-software-ltd .. price: 159.99. Product B is title: sibelius compass music composition teaching software music production software. manufacturer: sibelius-software-ltd .. price: 119.95. Are Product A and Product B the same? Yes\n\nProduct A is title: sibelius 4 professional edition. manufacturer: sibelius-software-ltd .. price: 599.99. Product B is title: sibelius sibelius 5 software music production software. manufacturer: sibelius-software-ltd .. price: 449.95. Are Product A and Product B the same? No\n\nProduct A is title: sibelius 4 professional edition. manufacturer: sibelius-software-ltd .. price: 599.99. Product B is title: sibelius sibelius 4 software music production software. manufacturer: sibelius-software-ltd .. price: 449.95. Are Product A and Product B the same? Yes\n\nProduct A is title: sibelius 5 professional edition. manufacturer: sibelius-software-ltd .. price: 599.0. Product B is title: sibelius sibelius 5 software music production software. manufacturer: sibelius-software-ltd .. price: 449.95. Are Product A and Product B the same? Yes\n\nProduct A is title: sibelius 4 professional edition. manufacturer: sibelius-software-ltd .. price: 599.99. Product B is title: sibelius sibelius 5 software music production software. manufacturer: sibelius-software-ltd .. price: 449.95. Are Product A and Product B the same? No\n\nProduct A is title: sibelius 5 educational edition. manufacturer: sibelius-software-ltd .. price: 329.0. Product B is title: sibelius sibelius 5 software educational discount music production software. manufacturer: sibelius-software-ltd .. price: 248.95. Are Product A and Product B the same? Yes\n\n",
        "imsi design": "Product A is title: instant architect design suite. manufacturer: imsi design. price: 29.99. Product B is title: punch software pro home design suite. manufacturer: . price: 61.8. Are Product A and Product B the same? No\n\nProduct A is title: instant architect design suite. manufacturer: imsi design. price: 29.99. Product B is title: encore inc 3d home architect home design deluxe version 9. manufacturer: . price: 25.49. Are Product A and Product B the same? No\n\nProduct A is title: instant architect design suite. manufacturer: imsi design. price: 29.99. Product B is title: instant home design 1 ( jc ). manufacturer: . price: 12.9. Are Product A and Product B the same? No\n\nProduct A is title: clip art and more 250 000 ( jewel case ). manufacturer: imsi design. price: . Product B is title: poetry fiction and drama ( win/mac ) ( jewel case ) ( 10172 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: clip art and more 250 000 ( jewel case ). manufacturer: imsi design. price: . Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: clip art and more 250 000 ( jewel case ). manufacturer: imsi design. price: . Product B is title: clipart & more 250000 ( jc ). manufacturer: . price: 9.9. Are Product A and Product B the same? Yes\n\n",
        "encore software": "Product A is title: reader rabbit learning system 1st grade. manufacturer: encore software. price: 29.99. Product B is title: dr. seuss reading learning system 2008 encore. manufacturer: . price: 19.95. Are Product A and Product B the same? No\n\nProduct A is title: mavis beacon typing 17 ( win/mac ). manufacturer: encore software. price: 19.99. Product B is title: broderbund mavis beacon teaches typing standard17. manufacturer: . price: 22.99. Are Product A and Product B the same? No\n\nProduct A is title: aquarium 3-in-one collection ( win/mac ). manufacturer: encore software. price: 19.99. Product B is title: aquarium 3-in-one limited edition ( win 95 98 me 2000 xp/mac 10.1 or higher ). manufacturer: . price: 19.95. Are Product A and Product B the same? Yes\n\nProduct A is title: midway arcade treasures ( jewel case ). manufacturer: encore software. price: 9.99. Product B is title: cross stitch design studio ( jewel case ) ( 8006 ). manufacturer: m-and-r-technologies. price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: printmaster gold v 17.0. manufacturer: encore software. price: 19.99. Product B is title: print master platinum v17. manufacturer: . price: 29.9. Are Product A and Product B the same? No\n\nProduct A is title: spongebob squarepants typing 2008 win/mac. manufacturer: encore software. price: 19.95. Product B is title: encore inc spongebob squarepants typing fun. manufacturer: . price: 17.1. Are Product A and Product B the same? Yes\n\nProduct A is title: jewels of cleopatra. manufacturer: encore software. price: 19.95. Product B is title: encore software 11861 jewels of cleopatra win xp vista. manufacturer: . price: 18.97. Are Product A and Product B the same? Yes\n\nProduct A is title: 3d home architect landscape designer v8. manufacturer: encore software. price: 29.99. Product B is title: encore inc 3d home architect home design deluxe version 9. manufacturer: . price: 25.49. Are Product A and Product B the same? No\n\nProduct A is title: clickart fonts ( 2006 ) by encore. manufacturer: encore software. price: 19.99. Product B is title: encore software 10463 clickart fonts ( 2006 ) by encore. manufacturer: . price: 17.97. Are Product A and Product B the same? Yes\n\nProduct A is title: hasbro family game collection 2006. manufacturer: encore software. price: 19.99. Product B is title: encore software 11273 hasbro family collection 2006 sb cs by hasbro. manufacturer: . price: 17.97. Are Product A and Product B the same? Yes\n\n",
        "broderbund": "Product A is title: mavis beacon teaches typing 16. manufacturer: broderbund. price: 19.99. Product B is title: mavis beacon teaches typing 16 deluxe. manufacturer: . price: 34.9. Are Product A and Product B the same? No\n\nProduct A is title: mavis beacon teaches typing deluxe 16. manufacturer: broderbund. price: 39.99. Product B is title: mavis beacon teaches typing v17 deluxe pc/mac. manufacturer: . price: 29.9. Are Product A and Product B the same? No\n\nProduct A is title: mavis beacon teaches typing 16. manufacturer: broderbund. price: 19.99. Product B is title: mavis beacon teaches typing 17 standard ( pc/mac ) encore. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: the printshop 20 professional publisher. manufacturer: broderbund. price: 69.99. Product B is title: the print shop ( r ) 22. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: the printshop 20. manufacturer: broderbund. price: 29.99. Product B is title: the printshop v22 deluxe. manufacturer: . price: 29.9. Are Product A and Product B the same? No\n\nProduct A is title: mavis beacon teaches typing 16. manufacturer: broderbund. price: 19.99. Product B is title: mavis beacon 16. manufacturer: . price: 17.9. Are Product A and Product B the same? Yes\n\n",
        "encore": "Product A is title: hoyle kids ' games jc. manufacturer: encore. price: 9.99. Product B is title: hoyle card games. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: the print shop 22 deluxe. manufacturer: encore. price: 49.99. Product B is title: encore software 11941 the print shop zoom. manufacturer: . price: 17.97. Are Product A and Product B the same? No\n\nProduct A is title: hoyle casino ( 2007 ). manufacturer: encore. price: 19.99. Product B is title: hoyle ( r ) card games 2007. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: chicken hunter wanted jc. manufacturer: encore. price: 9.99. Product B is title: encore software 10760 chicken hunter wanted win 98 me 2000 xp. manufacturer: . price: 7.69. Are Product A and Product B the same? Yes\n\nProduct A is title: hoyle : classic collection 2006. manufacturer: encore. price: 19.99. Product B is title: encore software 11052 hoyle : classic collection 2006 win 98 me 2000 xp. manufacturer: . price: 18.97. Are Product A and Product B the same? Yes\n\nProduct A is title: tlc arthur 's kindergarten learning system 2008. manufacturer: encore. price: 19.99. Product B is title: encore software 13871 tlc reader rabbit 1st grade learning system 2008. manufacturer: . price: 17.97. Are Product A and Product B the same? No\n\nProduct A is title: hoyle : classic collection 2006. manufacturer: encore. price: 19.99. Product B is title: encore software 11052 hoyle : classic collection 2006 win 98 me 2000 xp. manufacturer: . price: 18.97. Are Product A and Product B the same? Yes\n\nProduct A is title: monopoly here & now. manufacturer: encore. price: 19.95. Product B is title: encore software 11180 monopoly here & now win 98 me 2000 xp. manufacturer: . price: 18.97. Are Product A and Product B the same? Yes\n\nProduct A is title: hoyle kids ' games jc. manufacturer: encore. price: 9.99. Product B is title: hoyle card games. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: the print shop 22 deluxe. manufacturer: encore. price: 49.99. Product B is title: the printshop v22 deluxe. manufacturer: . price: 29.9. Are Product A and Product B the same? Yes\n\n",
        "corel": "Product A is title: intervideo dvd copy 5 platinum. manufacturer: corel. price: 49.99. Product B is title: corel intervideo dvd copy 5 platinum software for windows authoring software. manufacturer: . price: 49.95. Are Product A and Product B the same? Yes\n\nProduct A is title: corel dvd moviefactory 6.0 plus. manufacturer: corel. price: 79.99. Product B is title: corel ulead dvd moviefactory 6 software for windows authoring software. manufacturer: . price: 49.95. Are Product A and Product B the same? No\n\nProduct A is title: corel wordperfect office 12 standard upgrade. manufacturer: corel. price: 149.99. Product B is title: microsoft office 2004 std upgrade. manufacturer: . price: 219.99. Are Product A and Product B the same? No\n\nProduct A is title: ulead videostudio 11. manufacturer: corel. price: 89.99. Product B is title: corel ulead videostudio 11 plus software for win educational license authoring software. manufacturer: . price: 89.95. Are Product A and Product B the same? No\n\nProduct A is title: ulead videostudio 11. manufacturer: corel. price: 89.99. Product B is title: corel ulead videostudio 11 video editing and dvd authoring software for windows authoring software. manufacturer: . price: 89.95. Are Product A and Product B the same? Yes\n\nProduct A is title: corel painter x win/mac. manufacturer: corel. price: 429.0. Product B is title: corel painter x. manufacturer: . price: 395.99. Are Product A and Product B the same? Yes\n\nProduct A is title: corel dvd moviefactory 6.0 plus. manufacturer: corel. price: 79.99. Product B is title: ulead ( r ) dvd moviefactory ( r ) 5. manufacturer: . price: 49.99. Are Product A and Product B the same? No\n\nProduct A is title: corel dvd moviefactory 6.0. manufacturer: corel. price: 49.99. Product B is title: corel ulead dvd moviefactory 6 software for windows authoring software. manufacturer: . price: 49.95. Are Product A and Product B the same? Yes\n\nProduct A is title: corel dvd moviefactory 6.0 plus. manufacturer: corel. price: 79.99. Product B is title: corel ulead dvd moviefactory plus 6 software for windows authoring software. manufacturer: . price: 79.95. Are Product A and Product B the same? Yes\n\nProduct A is title: corel wordperfect office 12 standard upgrade. manufacturer: corel. price: 149.99. Product B is title: adobe cs3 design standard upgrade. manufacturer: . price: 413.99. Are Product A and Product B the same? No\n\n",
        "microspot ltd.": "Product A is title: microspot interiors. manufacturer: microspot ltd.. price: 99.95. Product B is title: microspot interiors ( mac 10.2 or later ). manufacturer: . price: 99.95. Are Product A and Product B the same? Yes\n\nProduct A is title: microspot macdraft professional ( mac ). manufacturer: microspot ltd.. price: 349.99. Product B is title: adobe acrobat v8 .0 professional for mac. manufacturer: . price: 405.99. Are Product A and Product B the same? No\n\nProduct A is title: microspot macdraft pe ( mac ). manufacturer: microspot ltd.. price: 99.95. Product B is title: microsoft office 2004 for students and teachers ( mac ). manufacturer: . price: 134.99. Are Product A and Product B the same? No\n\nProduct A is title: microspot macdraft professional ( mac ). manufacturer: microspot ltd.. price: 349.99. Product B is title: adobe acrobat v8 .0 professional for mac. manufacturer: . price: 405.99. Are Product A and Product B the same? No\n\nProduct A is title: microspot macdraft pe ( mac ). manufacturer: microspot ltd.. price: 99.95. Product B is title: microspot macdraft pe personal edition. manufacturer: . price: 85.99. Are Product A and Product B the same? Yes\n\nProduct A is title: microspot interiors ( mac ). manufacturer: microspot ltd.. price: 99.95. Product B is title: microspot interiors version 3.6. manufacturer: . price: 85.99. Are Product A and Product B the same? Yes\n\nProduct A is title: microspot interiors ( mac ). manufacturer: microspot ltd.. price: 99.95. Product B is title: microsoft office 2004 for students and teachers ( mac ). manufacturer: . price: 134.99. Are Product A and Product B the same? No\n\nProduct A is title: microspot macdraft professional ( mac ). manufacturer: microspot ltd.. price: 349.99. Product B is title: microspot macdraft professional edition. manufacturer: . price: 249.99. Are Product A and Product B the same? Yes\n\nProduct A is title: microspot macdraft professional ( mac ). manufacturer: microspot ltd.. price: 349.99. Product B is title: microsoft academic microsoft office project 2003 professional. manufacturer: . price: 176.99. Are Product A and Product B the same? No\n\n",
        "victory multimedia": "Product A is title: sales skills 2.0 ages 10 +. manufacturer: victory multimedia. price: 19.95. Product B is title: simply magazine sales skills. manufacturer: . price: 3.1. Are Product A and Product B the same? Yes\n\n",
        "sonic systems inc. .": "Product A is title: sonicwall comprehensive gateway security suite 1 year upgrade plan 01-ssc-5845 ). manufacturer: sonic systems inc. .. price: . Product B is title: sonicwall comprehensive gateway security for tz 150 series 25 node sonicwall -. manufacturer: . price: 107.01. Are Product A and Product B the same? No\n\nProduct A is title: sonicwall global vpn client license 01-ssc-5310 ). manufacturer: sonic systems inc. .. price: . Product B is title: rfipsc2-1 ipsec vpn client license 1 user. manufacturer: . price: 57.2. Are Product A and Product B the same? No\n\nProduct A is title: sonicwall global vpn client license 01-ssc-5310 ). manufacturer: sonic systems inc. .. price: . Product B is title: procurve vpn client software unlimited client licenses. manufacturer: . price: 2598.96. Are Product A and Product B the same? No\n\nProduct A is title: sonicwall global vpn client license 01-ssc-5310 ). manufacturer: sonic systems inc. .. price: . Product B is title: 01-ssc-6993 sonicwall client/server anti-virus suite subscription license 3 years 50. manufacturer: . price: 3992.98. Are Product A and Product B the same? No\n\nProduct A is title: sonicwall global vpn client license 01-ssc-5310 ). manufacturer: sonic systems inc. .. price: . Product B is title: sonicwall 01-ssc-2574. manufacturer: sonicwall. price: 561.65. Are Product A and Product B the same? No\n\n",
        "nova development": "Product A is title: business plan writer deluxe 8.0 2005. manufacturer: nova development. price: 99.99. Product B is title: palo alto software business plan pro 2007. manufacturer: palo-alto-software. price: 88.83. Are Product A and Product B the same? No\n\nProduct A is title: humorous greeting card factory. manufacturer: nova development. price: 29.99. Product B is title: nova development ghw humorous greeting card factory. manufacturer: nova development. price: 29.99. Are Product A and Product B the same? Yes\n\nProduct A is title: art explosion 600 000 images. manufacturer: nova development. price: 79.99. Product B is title: nova development corp art explosion 600000 images. manufacturer: . price: 72.86. Are Product A and Product B the same? Yes\n\nProduct A is title: printmaster gold v 17.0. manufacturer: nova development. price: 19.95. Product B is title: printmaster gold v 17.0 ( win 98 me 2000 xp ). manufacturer: . price: 19.95. Are Product A and Product B the same? Yes\n\nProduct A is title: business card factory deluxe 3.0. manufacturer: nova development. price: 29.99. Product B is title: nova greeting card factory for mac. manufacturer: . price: 41.99. Are Product A and Product B the same? No\n\nProduct A is title: humorous greeting card factory. manufacturer: nova development. price: 29.99. Product B is title: nova development corp humorous greeting card factory. manufacturer: . price: 28.3. Are Product A and Product B the same? Yes\n\nProduct A is title: hallmark blank greeting cards half-fold matte premium 20 count. manufacturer: nova development. price: 12.99. Product B is title: hallmark half-fold matte prem . blank greeting cards ( 20 pack ) ( paper ). manufacturer: . price: 10.99. Are Product A and Product B the same? Yes\n\nProduct A is title: photo explosion deluxe 3.0. manufacturer: nova development. price: 49.99. Product B is title: photoshow deluxe. manufacturer: . price: 39.92. Are Product A and Product B the same? No\n\nProduct A is title: photo explosion deluxe 3.0. manufacturer: nova development. price: 49.99. Product B is title: photostudio ( r ) expressions 2.0. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: photo objects 150 000 for macintosh. manufacturer: nova development. price: 49.99. Product B is title: bjmt art explosion photo objects 150000 complete package 1 user dvd mac. manufacturer: . price: 43.22. Are Product A and Product B the same? No\n\n",
        "marware": "Product A is title: project x project management software. manufacturer: marware. price: 199.95. Product B is title: practical software maintenance : best practices for managing your software. manufacturer: . price: 90.0. Are Product A and Product B the same? No\n\nProduct A is title: project x project management software. manufacturer: marware. price: 199.95. Product B is title: software. manufacturer: . price: 274.99. Are Product A and Product B the same? No\n\nProduct A is title: project x project management software. manufacturer: marware. price: 199.95. Product B is title: cakewalk project5 version 2 software music production software. manufacturer: . price: 199.0. Are Product A and Product B the same? No\n\nProduct A is title: project x project management software. manufacturer: marware. price: 199.95. Product B is title: now software nuw45e1k now up-to-date & contact complete product content management 1 user ( s ) complete product standard pc. manufacturer: . price: 107.97. Are Product A and Product B the same? No\n\nProduct A is title: project x project management software. manufacturer: marware. price: 199.95. Product B is title: marware project x project management software. manufacturer: . price: 179.99. Are Product A and Product B the same? Yes\n\nProduct A is title: project x project management software. manufacturer: marware. price: 199.95. Product B is title: software management approaches : project management estimation and life cycle. manufacturer: . price: 64.95. Are Product A and Product B the same? No\n\n",
        "sage software": "Product A is title: peachtree by sage premium accounting 2007 accountants ' edition. manufacturer: sage software. price: 499.99. Product B is title: sage ( ptree ) premium accounting accountants edition 2007. manufacturer: . price: 492.01. Are Product A and Product B the same? Yes\n\nProduct A is title: peachtree by sage premium accounting 2007. manufacturer: sage software. price: 499.99. Product B is title: sage ( ptree ) vernfp2007rt premium accounting for nonprofits 2007. manufacturer: . price: 590.35. Are Product A and Product B the same? No\n\nProduct A is title: peachtree by sage premium accounting 2007. manufacturer: sage software. price: 499.99. Product B is title: sage ( ptree ) vernfp2007rt premium accounting for nonprofits 2007. manufacturer: . price: 590.35. Are Product A and Product B the same? No\n\nProduct A is title: peachtree by sage complete accounting 2007. manufacturer: sage software. price: 269.99. Product B is title: sage ( ptree ) premium accounting for manufacturing 2007 multi-user. manufacturer: . price: 1180.37. Are Product A and Product B the same? No\n\nProduct A is title: peachtree by sage premium accounting for manufacturing 2007 multi user ( up to 5 users ). manufacturer: sage software. price: 1199.99. Product B is title: sage ( ptree ) premium accounting for manufacturing 2007 multi-user. manufacturer: . price: 1180.37. Are Product A and Product B the same? Yes\n\nProduct A is title: peachtree by sage premium accounting 2007. manufacturer: sage software. price: 499.99. Product B is title: peachtree ( r ) compatible continuous checks-office accounting 12.0. manufacturer: . price: 91.59. Are Product A and Product B the same? No\n\nProduct A is title: peachtree by sage complete accounting 2007. manufacturer: sage software. price: 269.99. Product B is title: sage software peachtree complete accounting 2007 software for windows tax & finance software. manufacturer: . price: 249.95. Are Product A and Product B the same? Yes\n\nProduct A is title: peachtree by sage premium accounting 2007. manufacturer: sage software. price: 499.99. Product B is title: laser checks for peachtree ( r ) accounting for windows 8.0-2005. manufacturer: . price: 106.29. Are Product A and Product B the same? No\n\n",
        "intego inc": "Product A is title: internet security barrier x4 antispam edition. manufacturer: intego inc. price: . Product B is title: zonealarm ( r ) internet security suite. manufacturer: . price: 49.99. Are Product A and Product B the same? No\n\nProduct A is title: internet security barrier x4 antispam edition. manufacturer: intego inc. price: . Product B is title: norton internet security ( v3 .0 ) mac symantec 10069614-in. manufacturer: . price: 83.43. Are Product A and Product B the same? No\n\nProduct A is title: internet security barrier x4 antispam edition. manufacturer: intego inc. price: . Product B is title: internet security suite 2007 home protection pack xl box. manufacturer: . price: 61.06. Are Product A and Product B the same? No\n\nProduct A is title: internet security barrier x4 antispam edition. manufacturer: intego inc. price: . Product B is title: intego inc internet security barrier x4 antispam edition. manufacturer: . price: 71.75. Are Product A and Product B the same? Yes\n\nProduct A is title: internet security barrier x4 antispam edition. manufacturer: intego inc. price: . Product B is title: pc-cillin ( tm ) internet security 2007. manufacturer: . price: 49.99. Are Product A and Product B the same? No\n\nProduct A is title: internet security barrier x4 antispam edition. manufacturer: intego inc. price: . Product B is title: intego contentbarrier x4 10.4. manufacturer: . price: 54.99. Are Product A and Product B the same? No\n\n",
        "micromat": "Product A is title: micromat techtool protege with firewire device. manufacturer: micromat. price: . Product B is title: micromat techtool protege utility with 1gb firewire flash drive. manufacturer: . price: 219.99. Are Product A and Product B the same? Yes\n\nProduct A is title: micromat techtool pro 4 ( mac ). manufacturer: micromat. price: 99.99. Product B is title: micromat tech tool pro 4.5. manufacturer: . price: 82.99. Are Product A and Product B the same? Yes\n\nProduct A is title: micromat protogo. manufacturer: micromat. price: . Product B is title: micromat techtool protogo. manufacturer: . price: 109.99. Are Product A and Product B the same? Yes\n\nProduct A is title: micromat podlock ( mac ). manufacturer: micromat. price: 69.0. Product B is title: micromat podlock ipod utility software. manufacturer: . price: 39.99. Are Product A and Product B the same? Yes\n\nProduct A is title: micromat techtool protege with firewire device. manufacturer: micromat. price: . Product B is title: micromat tech tool pro 4.5. manufacturer: . price: 82.99. Are Product A and Product B the same? No\n\nProduct A is title: micromat protogo. manufacturer: micromat. price: . Product B is title: micromat inc. diskstudio. manufacturer: . price: 43.36. Are Product A and Product B the same? No\n\nProduct A is title: micromat podlock ( mac ). manufacturer: micromat. price: 69.0. Product B is title: micromat inc. diskstudio. manufacturer: . price: 43.36. Are Product A and Product B the same? No\n\nProduct A is title: micromat techtool protege with firewire device. manufacturer: micromat. price: . Product B is title: micromat techtool protogo. manufacturer: . price: 109.99. Are Product A and Product B the same? No\n\nProduct A is title: micromat podlock ( mac ). manufacturer: micromat. price: 69.0. Product B is title: microspot interiors ( mac 10.2 or later ). manufacturer: . price: 99.95. Are Product A and Product B the same? No\n\n",
        "microsoft corporation": "Product A is title: microsoft office mac 2004 english 3pk dsp 731-01163 ). manufacturer: microsoft corporation. price: . Product B is title: microsoft word 2004. manufacturer: . price: 209.99. Are Product A and Product B the same? No\n\nProduct A is title: microsoft office mac 2004 english 3pk dsp 731-01163 ). manufacturer: microsoft corporation. price: . Product B is title: microsoft ( r ) office word 2004 for mac full version. manufacturer: . price: 231.29. Are Product A and Product B the same? No\n\nProduct A is title: microsoft upg win small business svr cal t74-01215 ). manufacturer: microsoft corporation. price: 330.14. Product B is title: cabprenaen150 ca business protection suite for microsoft small business server premium edition. manufacturer: . price: 666.65. Are Product A and Product B the same? No\n\nProduct A is title: microsoft office mac 2004 english 3pk dsp 731-01163 ). manufacturer: microsoft corporation. price: . Product B is title: microsoft office 2004 software suite oem mac apple & mac software. manufacturer: . price: 239.95. Are Product A and Product B the same? Yes\n\nProduct A is title: microsoft office mac 2004 english 3pk dsp 731-01163 ). manufacturer: microsoft corporation. price: . Product B is title: microsoft academic microsoft office project 2003 standard. manufacturer: . price: 69.95. Are Product A and Product B the same? No\n\nProduct A is title: microsoft office mac 2004 english 3pk dsp 731-01163 ). manufacturer: microsoft corporation. price: . Product B is title: microsoft ( r ) office word 2004 for mac full version. manufacturer: . price: 231.29. Are Product A and Product B the same? No\n\n",
        "adobe systems inc": "Product A is title: adobe photoshop elements 4.0 plus adobe premiere elements 2.0 complete package 1 user cd mini-box win canadian french 29180162 ). manufacturer: adobe systems inc. price: . Product B is title: adobe premiere elements 3 and photoshop elements 5 software for windows consumer video editing software. manufacturer: . price: 149.95. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop elements 4.0 plus adobe premiere elements 2.0 complete package 1 user cd mini-box win canadian french 29180162 ). manufacturer: adobe systems inc. price: . Product B is title: prw photo explosion deluxe v. 3.0 complete package 1 user cd win. manufacturer: . price: 47.6. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop elements 4.0 plus adobe premiere elements 2.0 complete package 1 user cd mini-box win canadian french 29180162 ). manufacturer: adobe systems inc. price: . Product B is title: adobe photoshop cs3 extended complete package 1 user academic cd mac. manufacturer: . price: 279.0. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop elements 4.0 plus adobe premiere elements 2.0 complete package 1 user cd mini-box win canadian french 29180162 ). manufacturer: adobe systems inc. price: . Product B is title: adobe photoshop cs3 extended complete package 1 user academic cd mac. manufacturer: . price: 279.0. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop elements 4.0 plus adobe premiere elements 2.0 complete package 1 user cd mini-box win canadian french 29180162 ). manufacturer: adobe systems inc. price: . Product B is title: adobe dreamweaver cs3 complete package 1 user academic cd mac. manufacturer: . price: 189.0. Are Product A and Product B the same? No\n\n",
        "iplaymusic": "Product A is title: play music together mac. manufacturer: iplaymusic. price: 39.99. Product B is title: play music together dvd by iplaymusic mac. manufacturer: . price: 32.99. Are Product A and Product B the same? Yes\n\n",
        "avid technology": "Product A is title: avid liquid vs 7 software. manufacturer: avid technology. price: 888.75. Product B is title: software. manufacturer: . price: 153.99. Are Product A and Product B the same? No\n\nProduct A is title: avid liquid vs 7 software. manufacturer: avid technology. price: 888.75. Product B is title: software. manufacturer: . price: 46.99. Are Product A and Product B the same? No\n\nProduct A is title: avid liquid vs 7 software. manufacturer: avid technology. price: 888.75. Product B is title: avid technology inc. 70102024701 software avid liquid chrome xe. manufacturer: . price: 840.97. Are Product A and Product B the same? No\n\nProduct A is title: avid liquid vs 7 software. manufacturer: avid technology. price: 888.75. Product B is title: avid liquid 7.0 professional video editing software for windows professional editing software. manufacturer: . price: 439.95. Are Product A and Product B the same? Yes\n\nProduct A is title: avid liquid vs 7 software. manufacturer: avid technology. price: 888.75. Product B is title: onone software smartscale software. manufacturer: . price: 189.99. Are Product A and Product B the same? No\n\nProduct A is title: avid liquid vs 7 software. manufacturer: avid technology. price: 888.75. Product B is title: macspeech ilisten software. manufacturer: . price: 99.99. Are Product A and Product B the same? No\n\n",
        "mindjet": "",
        "cosmi": "Product A is title: cosmi talking math & more for kids ( windows ). manufacturer: cosmi. price: . Product B is title: learning quickbooks ( r ) for windows ( r ). manufacturer: . price: 39.99. Are Product A and Product B the same? No\n\nProduct A is title: cosmi rom03558 print perfect clip art deluxe. manufacturer: cosmi. price: . Product B is title: cosmi print perfect clip art deluxe. manufacturer: . price: 6.99. Are Product A and Product B the same? Yes\n\nProduct A is title: cosmi rom07524 print perfect business cards dvd. manufacturer: cosmi. price: . Product B is title: cosmi print perfect clip art deluxe. manufacturer: . price: 6.99. Are Product A and Product B the same? No\n\nProduct A is title: cosmi rom07524 print perfect business cards dvd. manufacturer: cosmi. price: . Product B is title: encore inc printshop business card maker. manufacturer: riverdeep-learning-company. price: 8.01. Are Product A and Product B the same? No\n\n",
        "hewlett packard": "Product A is title: procurve vpn client software unlimited client licenses. manufacturer: hewlett packard. price: . Product B is title: prosafe vpn client software single user netgear inc. vpn01l. manufacturer: . price: 27.78. Are Product A and Product B the same? No\n\nProduct A is title: procurve vpn client software unlimited client licenses. manufacturer: hewlett packard. price: . Product B is title: sr2116010 safenet softremote vpn client license 5 users win. manufacturer: . price: 184.57. Are Product A and Product B the same? No\n\nProduct A is title: red hat enterprise linux ws v. 4 update 4 license 1 workstation rl296aa ). manufacturer: hewlett packard. price: . Product B is title: 83076 smarterm enterprise suite v. 12.1 upgrade license 1 user win. manufacturer: . price: 146.42. Are Product A and Product B the same? No\n\nProduct A is title: procurve vpn client software unlimited client licenses. manufacturer: hewlett packard. price: . Product B is title: d-link vpn client software ( 1-user license ) ds-601 ds-601. manufacturer: . price: 42.99. Are Product A and Product B the same? No\n\nProduct A is title: procurve vpn client software unlimited client licenses. manufacturer: hewlett packard. price: . Product B is title: d-link vpn client software ( 5-user license ) ds-605 ds-605. manufacturer: . price: 173.99. Are Product A and Product B the same? No\n\nProduct A is title: procurve vpn client software unlimited client licenses. manufacturer: hewlett packard. price: . Product B is title: procurve vpn client software unlimited client licenses. manufacturer: . price: 2598.96. Are Product A and Product B the same? Yes\n\nProduct A is title: hp storageworks secure path for windows workgroup edition v. 4.0 c license 213076-b26 ). manufacturer: hewlett packard. price: . Product B is title: hewlett packard 213076-b26 secure path v4 .0 c win wkgp ed 1 lic/cd. manufacturer: . price: 1888.39. Are Product A and Product B the same? Yes\n\n",
        "webroot software": "Product A is title: spy sweeper with antivirus/child safe. manufacturer: webroot software. price: 39.95. Product B is title: webroot software inc 52310 spy sweeper with antivirus/child safe 1-pc win 2000 xp vista. manufacturer: . price: 37.97. Are Product A and Product B the same? Yes\n\nProduct A is title: spy sweeper spanish. manufacturer: webroot software. price: 29.95. Product B is title: webroot ( r ) spy sweeper ( tm ) 2006. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: spy sweeper w/antivirus family edition. manufacturer: webroot software. price: 49.95. Product B is title: spy sweeper ( small box ). manufacturer: . price: 22.99. Are Product A and Product B the same? No\n\nProduct A is title: webroot spy sweeper antispyware 5.x. manufacturer: webroot software. price: 29.95. Product B is title: webroot software inc 31250 spy sweeper spanish win 2000 xp. manufacturer: . price: 25.97. Are Product A and Product B the same? No\n\nProduct A is title: webroot spysweeper antispyware 3 user. manufacturer: webroot software. price: 39.99. Product B is title: webroot software 65210 spy sweeper 3 pc. manufacturer: webroot software. price: 31.99. Are Product A and Product B the same? Yes\n\nProduct A is title: webroot desktop firewall. manufacturer: webroot software. price: 29.99. Product B is title: webroot software 54210 desktop firewall. manufacturer: webroot software. price: 26.99. Are Product A and Product B the same? Yes\n\nProduct A is title: spy sweeper [ lb ]. manufacturer: webroot software. price: 29.99. Product B is title: spy sweeper ( small box ). manufacturer: . price: 22.99. Are Product A and Product B the same? No\n\nProduct A is title: webroot spy sweeper with antivirus family edition 3 pcs. manufacturer: webroot software. price: 49.95. Product B is title: webroot software inc 50210 spy sweeper w/antivirus family edition ( mass merchant ) win 2000 xp. manufacturer: . price: 41.97. Are Product A and Product B the same? No\n\nProduct A is title: webroot spy sweeper antispyware 5.x. manufacturer: webroot software. price: 29.95. Product B is title: webroot software inc 66410 spy sweeper ( small box ) ( win 98 me 2000 xp ). manufacturer: . price: 25.97. Are Product A and Product B the same? Yes\n\nProduct A is title: webroot desktop firewall. manufacturer: webroot software. price: 29.99. Product B is title: webroot software inc 54210 webroot desktop firewall firewall pc. manufacturer: . price: 25.97. Are Product A and Product B the same? Yes\n\n",
        "individual software": "Product A is title: resumemaker professional 12.0. manufacturer: individual software. price: 29.99. Product B is title: individual software resume maker professional deluxe 12.0. manufacturer: . price: 29.99. Are Product A and Product B the same? Yes\n\nProduct A is title: systemsuite 8 professional. manufacturer: individual software. price: 49.99. Product B is title: iolo system mechanic 7 professional ( pc ). manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\nProduct A is title: systemsuite 8 professional. manufacturer: individual software. price: 49.99. Product B is title: sibelius sibelius photoscore professional 4. manufacturer: . price: 145.7. Are Product A and Product B the same? No\n\nProduct A is title: resumemaker professional 12.0. manufacturer: individual software. price: 29.99. Product B is title: resumemaker ( r ) professional ultimate. manufacturer: . price: 39.99. Are Product A and Product B the same? No\n\nProduct A is title: resumemaker professional 12.0. manufacturer: individual software. price: 29.99. Product B is title: individual software prm-r12 resumemaker professional 12.0. manufacturer: individual software. price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: professor teaches windows xp. manufacturer: individual software. price: 19.99. Product B is title: professor teaches windows xp individual software. manufacturer: . price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: total 3d home & landscape design suite version 9. manufacturer: individual software. price: 39.99. Product B is title: individual software inc total 3d home & landscape deluxe 9. manufacturer: . price: 33.5. Are Product A and Product B the same? Yes\n\nProduct A is title: total 3d home & landscape design suite version 9. manufacturer: individual software. price: 39.99. Product B is title: punch software 85100 punch!master landscape pro v10 and home design. manufacturer: . price: 59.97. Are Product A and Product B the same? No\n\nProduct A is title: professor teaches excel & word. manufacturer: individual software. price: 19.99. Product B is title: individual software professor teaches excel & word. manufacturer: . price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: systemsuite 8 professional. manufacturer: individual software. price: 49.99. Product B is title: individual software inc systemsuite 8 professional ( us pos ). manufacturer: . price: 46.42. Are Product A and Product B the same? Yes\n\n",
        "ibm": "Product A is title: ibm rational clearcase lt v. 6.0.0 media vol passport passport adva. manufacturer: ibm. price: . Product B is title: bt00nna ibm rational clearcase lt v. 6.0.0 media cd win hp-ux solaris en. manufacturer: . price: 108.95. Are Product A and Product B the same? No\n\n",
        "bling software": "",
        "nuance communications inc.": "Product A is title: pdf converter professional v4 english. manufacturer: nuance communications inc.. price: 99.99. Product B is title: scansoft pdf converter 4. manufacturer: . price: 37.95. Are Product A and Product B the same? No\n\nProduct A is title: scansoft paperport professional 11 5 user license. manufacturer: nuance communications inc.. price: 949.99. Product B is title: resume maker professional 12. manufacturer: . price: 17.95. Are Product A and Product B the same? No\n\nProduct A is title: pdf converter professional v4 english. manufacturer: nuance communications inc.. price: 99.99. Product B is title: resume maker professional 12. manufacturer: . price: 17.95. Are Product A and Product B the same? No\n\nProduct A is title: scansoft paperport professional 11 5 user license. manufacturer: nuance communications inc.. price: 949.99. Product B is title: orgchart professional 1.3. manufacturer: . price: 39.95. Are Product A and Product B the same? No\n\nProduct A is title: scansoft paperport professional 11 5 user license. manufacturer: nuance communications inc.. price: 949.99. Product B is title: m109a-g00-4 .0 scansoft pdf converter professional v. 4.0 complete package 1 user w. manufacturer: . price: 87.79. Are Product A and Product B the same? No\n\n",
        "wasp technologies": "Product A is title: wasp bar code wasp bar code labeler complete package 3 users std cd -. manufacturer: wasp technologies. price: . Product B is title: prw photo explosion deluxe v. 3.0 complete package 1 user cd win. manufacturer: . price: 47.6. Are Product A and Product B the same? No\n\nProduct A is title: wasp bar code wasp bar code labeler complete package 3 users std cd -. manufacturer: wasp technologies. price: . Product B is title: 298881 quickbooks customer manager v. 2.5 complete package 1 user cd win. manufacturer: . price: 69.38. Are Product A and Product B the same? No\n\nProduct A is title: wasp bar code wasp bar code labeler complete package 3 users std cd -. manufacturer: wasp technologies. price: . Product B is title: 24100 punch ! 5 in 1 home design complete package 1 user cd win. manufacturer: . price: 36.28. Are Product A and Product B the same? No\n\nProduct A is title: wasp bar code wasp bar code labeler complete package 3 users std cd -. manufacturer: wasp technologies. price: . Product B is title: m009a-l00-4 .0 scansoft pdf create ! v. 4.0 complete package 1 user cd win engli. manufacturer: . price: 43.21. Are Product A and Product B the same? No\n\nProduct A is title: wasp bar code wasp bar code labeler complete package 3 users std cd -. manufacturer: wasp technologies. price: . Product B is title: snt-040 ntracker complete package 1 user win. manufacturer: . price: 26.59. Are Product A and Product B the same? No\n\n",
        "dantz": "Product A is title: retrospect 6.0 desktop mac upgrade. manufacturer: dantz. price: 59.99. Product B is title: emc insignia retrospect workgroup 6 upgrade. manufacturer: . price: 162.99. Are Product A and Product B the same? No\n\nProduct A is title: emc retrospect 7.5 professional for windows upgrade. manufacturer: dantz. price: . Product B is title: avid liquid pro 7.0 video editing hardware/software bundle for windows professional editing software. manufacturer: . price: 899.0. Are Product A and Product B the same? No\n\nProduct A is title: emc retrospect 7.5 professional for windows upgrade. manufacturer: dantz. price: . Product B is title: adobe after effects professional 7 software for windows effects software. manufacturer: . price: 999.0. Are Product A and Product B the same? No\n\nProduct A is title: emc retrospect 7.5 svr 1-client windows. manufacturer: dantz. price: 249.0. Product B is title: emc insignia retrospect server client v. 7.5 complete package. manufacturer: . price: 181.97. Are Product A and Product B the same? Yes\n\nProduct A is title: retrospect 6.0 desktop mac upgrade. manufacturer: dantz. price: 59.99. Product B is title: bu24a600000 emc insignia retrospect desktop v. 6.1 upgrade package 2 clients cd. manufacturer: . price: 51.85. Are Product A and Product B the same? No\n\nProduct A is title: upgrade emc retrospect 7.5 multi server windows only. manufacturer: dantz. price: . Product B is title: retrospect 7.5 prof win upg only pz24a0075. manufacturer: emc. price: 54.68. Are Product A and Product B the same? No\n\n",
        "prosoft engineering": "Product A is title: prosoft tunetech for ipod ( mac ). manufacturer: prosoft engineering. price: 74.99. Product B is title: valusoft tune transfer for ipod. manufacturer: . price: 21.99. Are Product A and Product B the same? No\n\nProduct A is title: prosoft tunetech for ipod ( mac ). manufacturer: prosoft engineering. price: 74.99. Product B is title: prosoft 27100 tunetech for ipod by prosoft 0794038271005. manufacturer: . price: 51.11. Are Product A and Product B the same? No\n\n",
        "valusoft": "Product A is title: ultimate puzzle games. manufacturer: valusoft. price: 19.99. Product B is title: ultimate pinball. manufacturer: . price: 6.95. Are Product A and Product B the same? No\n\nProduct A is title: pokemon team turbo ( jewel case ). manufacturer: valusoft. price: 9.99. Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: ultimate puzzle games. manufacturer: valusoft. price: 19.99. Product B is title: ultimate fonts. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: sea scene : living underwater worlds ( jewel case ). manufacturer: valusoft. price: 9.99. Product B is title: sea scene : living underwater worlds. manufacturer: . price: 19.55. Are Product A and Product B the same? Yes\n\nProduct A is title: print workshop 2007. manufacturer: valusoft. price: 19.99. Product B is title: print shop deluxe 21. manufacturer: . price: 44.9. Are Product A and Product B the same? No\n\nProduct A is title: mastercook deluxe low carb. manufacturer: valusoft. price: 14.99. Product B is title: mastercook deluxe low carb edition. manufacturer: . price: 17.9. Are Product A and Product B the same? Yes\n\nProduct A is title: ultimate pinball ( jewel case ). manufacturer: valusoft. price: 9.99. Product B is title: ultimate fonts. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\n",
        "onone software": "Product A is title: genuine fractals print pro 4.1 sharpener pro 2.0. manufacturer: onone software. price: . Product B is title: onone software inc. gpp-55211 genuine fractals print pro 5 1u. manufacturer: . price: 274.97. Are Product A and Product B the same? No\n\nProduct A is title: onone software pxl smartscale windows and macintosh ). manufacturer: onone software. price: . Product B is title: onone software inc. mpr-40212. manufacturer: onone software inc.. price: 66.27. Are Product A and Product B the same? No\n\nProduct A is title: onone software intellihance pro 4.1. manufacturer: onone software. price: . Product B is title: onone software plug-in suite full. manufacturer: . price: 393.68. Are Product A and Product B the same? No\n\nProduct A is title: onone plug-in suite 3 for adobe photoshop. manufacturer: onone software. price: 399.95. Product B is title: onone software plug-in suite full. manufacturer: . price: 393.68. Are Product A and Product B the same? Yes\n\nProduct A is title: onone software pxl smartscale windows and macintosh ). manufacturer: onone software. price: . Product B is title: pxl smartscale for mac/win. manufacturer: . price: 159.95. Are Product A and Product B the same? Yes\n\nProduct A is title: onone essentials for adobe photoshop elements. manufacturer: onone software. price: 59.95. Product B is title: onone software ele-10211 on1 essentials v. 1.0 for photoshop elements complete product graphics/designing 1 user complete product standard english pc mac. manufacturer: . price: 50.97. Are Product A and Product B the same? Yes\n\nProduct A is title: onone essentials for adobe photoshop elements version 2 full. manufacturer: onone software. price: 59.95. Product B is title: adobe photoshop cs3 extended for mac. manufacturer: . price: 935.99. Are Product A and Product B the same? No\n\nProduct A is title: genuine fractals print pro 5 1u. manufacturer: onone software. price: 299.95. Product B is title: onone software inc. gpp-55211 genuine fractals print pro 5 1u. manufacturer: . price: 274.97. Are Product A and Product B the same? Yes\n\nProduct A is title: onone plug-in suite 3 for adobe photoshop. manufacturer: onone software. price: 399.95. Product B is title: onone software pps-30211 on1 plug-in suite v. 3.0 for adobe photoshop complete product image collection/editing/archive complete product pc mac intel-based mac. manufacturer: . price: 365.97. Are Product A and Product B the same? Yes\n\nProduct A is title: onone software pxl smartscale windows and macintosh ). manufacturer: onone software. price: . Product B is title: onone software inc. mpr-40212. manufacturer: onone software inc.. price: 66.27. Are Product A and Product B the same? No\n\n",
        "the learning company": "Product A is title: reader rabbit kindergarten version 1.1. manufacturer: the learning company. price: 9.99. Product B is title: reader rabbit reading learning system 2007. manufacturer: . price: 12.95. Are Product A and Product B the same? No\n\nProduct A is title: reader rabbit kindergarten version 1.1. manufacturer: the learning company. price: 9.99. Product B is title: jumpstart kindergarten ( jc ). manufacturer: . price: 9.9. Are Product A and Product B the same? No\n\nProduct A is title: reader rabbit playtime for baby and toddler. manufacturer: the learning company. price: 19.99. Product B is title: reader rabbit playtime for baby & toddler. manufacturer: . price: 9.9. Are Product A and Product B the same? No\n\nProduct A is title: cluefinders 3rd grade. manufacturer: the learning company. price: . Product B is title: brain play 1st-3rd grade. manufacturer: . price: 17.9. Are Product A and Product B the same? No\n\nProduct A is title: reader rabbit kindergarten version 1.1. manufacturer: the learning company. price: 9.99. Product B is title: reader rabbit learn to read with phonics ! preschool & kindergarten. manufacturer: . price: 12.95. Are Product A and Product B the same? No\n\n",
        "emedia": "Product A is title: emedia rock guitar method. manufacturer: emedia. price: 39.99. Product B is title: emedia music corp guitar pro v 5.1 ( hybrid ). manufacturer: . price: 51.92. Are Product A and Product B the same? No\n\nProduct A is title: emedia bass method win/mac. manufacturer: emedia. price: 59.99. Product B is title: emedia eg03014 bass method os x. manufacturer: emedia. price: 53.99. Are Product A and Product B the same? Yes\n\nProduct A is title: finale 2007 academic ( win/mac ). manufacturer: emedia. price: 350.0. Product B is title: makemusic finale 2007 academic create edit and print sheet music. manufacturer: . price: 269.0. Are Product A and Product B the same? Yes\n\nProduct A is title: emedia blues guitar legends. manufacturer: emedia. price: 29.94. Product B is title: emedia essential bass guitar dvd. manufacturer: . price: 16.95. Are Product A and Product B the same? No\n\nProduct A is title: my guitar. manufacturer: emedia. price: 29.95. Product B is title: emedia music corp my guitar. manufacturer: . price: 24.81. Are Product A and Product B the same? Yes\n\nProduct A is title: emedia intermediate guitar method os x. manufacturer: emedia. price: 59.99. Product B is title: emedia music corp emedia intermediate guitar method. manufacturer: . price: 49.21. Are Product A and Product B the same? No\n\nProduct A is title: my guitar. manufacturer: emedia. price: 29.95. Product B is title: mike dowling homespun swing guitar. manufacturer: . price: 36.67. Are Product A and Product B the same? No\n\nProduct A is title: emedia piano and keyboard method volume 1. manufacturer: emedia. price: 59.99. Product B is title: emedia music corp emedia piano & keyboard method. manufacturer: . price: 49.21. Are Product A and Product B the same? Yes\n\nProduct A is title: emedia rock guitar method. manufacturer: emedia. price: 39.99. Product B is title: emedia music corp emedia blues guitar legends. manufacturer: . price: 24.81. Are Product A and Product B the same? No\n\nProduct A is title: emedia guitarpro 5 win/mac. manufacturer: emedia. price: 59.99. Product B is title: emedia music corp guitar pro v 5.1 ( hybrid ). manufacturer: . price: 51.92. Are Product A and Product B the same? Yes\n\n",
        "symantec": "Product A is title: symc backup exec sbs std 11d win small business server standard. manufacturer: symantec. price: 811.21. Product B is title: 10759333 symantec backup exec 11d for windows servers central admin server option compl. manufacturer: . price: 1219.72. Are Product A and Product B the same? No\n\nProduct A is title: symc backup exec sbs prm 11d win small business server. manufacturer: symantec. price: 885.83. Product B is title: e17887d symantec backup exec for netware servers remote agent for windows or netware ser. manufacturer: . price: 197.65. Are Product A and Product B the same? No\n\nProduct A is title: symc backup exec sbs aws 11d win sbs agent for windows systems. manufacturer: symantec. price: 447.74. Product B is title: 10759328 symantec backup exec 11d for windows servers agent for db2 on windows servers -. manufacturer: . price: 644.84. Are Product A and Product B the same? No\n\nProduct A is title: norton internet security mac 3.0 [ antivirus firewall privacy controls iclean ]. manufacturer: symantec. price: 99.95. Product B is title: symantec norton internet security 3.0 software for macintosh apple & mac software. manufacturer: . price: 89.95. Are Product A and Product B the same? Yes\n\nProduct A is title: norton antivirus 2007. manufacturer: symantec. price: 39.99. Product B is title: symantec norton anti-virus 2007 windows. manufacturer: . price: 39.99. Are Product A and Product B the same? Yes\n\nProduct A is title: norton internet security 2008 10 user. manufacturer: symantec. price: 199.99. Product B is title: norton internet security 2008 10 user. manufacturer: . price: 199.99. Are Product A and Product B the same? Yes\n\nProduct A is title: symantec client security 3.1 with groupware protection business pack 50 user. manufacturer: symantec. price: 3732.0. Product B is title: 10517940 symantec client security with groupware protection business pack v. 3.1 -. manufacturer: . price: 3323.92. Are Product A and Product B the same? No\n\nProduct A is title: norton antivirus mac 10.0 ( mac ). manufacturer: symantec. price: 69.95. Product B is title: symantec norton anti-virus v. 10 mac. manufacturer: . price: 64.99. Are Product A and Product B the same? Yes\n\nProduct A is title: norton internet security 2008 10 user. manufacturer: symantec. price: 199.99. Product B is title: symantec norton internet security 3.0 software for macintosh apple & mac software. manufacturer: . price: 89.95. Are Product A and Product B the same? No\n\n",
        "apple computer": "Product A is title: apple final cut express hd 3.5 ( mac ). manufacturer: apple computer. price: 299.0. Product B is title: apple final cut studio 2 upgrade from final cut pro. manufacturer: . price: 699.99. Are Product A and Product B the same? No\n\nProduct A is title: apple final cut express hd 3.5 ( mac ). manufacturer: apple computer. price: 299.0. Product B is title: apple software m9372z/a final cut express 2.0 upgrade. manufacturer: apple software. price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: logic express 7. manufacturer: apple computer. price: 299.0. Product B is title: apple/emagic logic express 7.2 software music production software. manufacturer: . price: 299.0. Are Product A and Product B the same? Yes\n\nProduct A is title: apple final cut express hd 3.5 ( mac ). manufacturer: apple computer. price: 299.0. Product B is title: ma265zm/a final cut express hd v. 3.5 media volume dvd mac multi-country. manufacturer: . price: 12.98. Are Product A and Product B the same? No\n\nProduct A is title: apple ilife '06 ( mac dvd ) [ older version ]. manufacturer: apple computer. price: 79.0. Product B is title: ilife '06 mac apple ma166z/a. manufacturer: . price: 47.8. Are Product A and Product B the same? Yes\n\nProduct A is title: apple mac os x tiger 10.4.6 family pack ( mac ). manufacturer: apple computer. price: 199.0. Product B is title: apple shake 4.1 mac os x. manufacturer: . price: 499.99. Are Product A and Product B the same? No\n\nProduct A is title: apple mac os x tiger 10.4.6 family pack ( mac ). manufacturer: apple computer. price: 199.0. Product B is title: apple mac os x 10.4.6 tiger family pack. manufacturer: . price: 185.99. Are Product A and Product B the same? Yes\n\nProduct A is title: apple ilife '06 family pack ( mac dvd ) [ older version ]. manufacturer: apple computer. price: 99.0. Product B is title: apple iwork '06 family pack. manufacturer: . price: 99.99. Are Product A and Product B the same? No\n\nProduct A is title: logic express 7. manufacturer: apple computer. price: 299.0. Product B is title: apple logic express 7.2. manufacturer: . price: 299.99. Are Product A and Product B the same? Yes\n\nProduct A is title: apple final cut express hd 3.5 ( mac ). manufacturer: apple computer. price: 299.0. Product B is title: apple final cut express hd 3.5 video editing software for mac av production software. manufacturer: . price: 299.0. Are Product A and Product B the same? Yes\n\n",
        "school-zone-interactive": "Product A is title: school zone pencil-pal software big phonics ( cd-rom & book ). manufacturer: school-zone-interactive. price: . Product B is title: school zone interactive shapes on track software. manufacturer: . price: 9.45. Are Product A and Product B the same? No\n\nProduct A is title: school zone pencil-pal software big phonics ( cd-rom & book ). manufacturer: school-zone-interactive. price: . Product B is title: school zone interactive math 2 on track software. manufacturer: . price: 9.45. Are Product A and Product B the same? No\n\n",
        "panicware": "Product A is title: pop-up stopper professional. manufacturer: panicware. price: 29.99. Product B is title: adobe acrobat professional 7. manufacturer: . price: 350.0. Are Product A and Product B the same? No\n\nProduct A is title: pop-up stopper professional. manufacturer: panicware. price: 29.99. Product B is title: system mechanic 6 professional. manufacturer: . price: 36.9. Are Product A and Product B the same? No\n\nProduct A is title: pop-up stopper professional. manufacturer: panicware. price: 29.99. Product B is title: system care ? professional. manufacturer: . price: 49.99. Are Product A and Product B the same? No\n\nProduct A is title: pop-up stopper professional. manufacturer: panicware. price: 29.99. Product B is title: csdc pop-up stopper professional panicware. manufacturer: . price: 26.14. Are Product A and Product B the same? Yes\n\nProduct A is title: pop-up stopper professional. manufacturer: panicware. price: 29.99. Product B is title: avanquest usa 1461 system suit 7 professional. manufacturer: avanquest usa. price: 44.99. Are Product A and Product B the same? No\n\nProduct A is title: pop-up stopper professional. manufacturer: panicware. price: 29.99. Product B is title: orgchart professional 1.3. manufacturer: . price: 39.95. Are Product A and Product B the same? No\n\n",
        "avanquest": "Product A is title: system care professional. manufacturer: avanquest. price: 49.95. Product B is title: adobe premiere pro 2.0 software professional editing software. manufacturer: . price: 848.95. Are Product A and Product B the same? No\n\nProduct A is title: vcomm system suite professional 7. manufacturer: avanquest. price: 49.95. Product B is title: adobe cs3 flash professional. manufacturer: . price: 724.99. Are Product A and Product B the same? No\n\nProduct A is title: encyclopedia britannica deluxe 2008 win/mac. manufacturer: avanquest. price: 29.95. Product B is title: britannica deluxe 2008. manufacturer: . price: 26.5. Are Product A and Product B the same? Yes\n\nProduct A is title: vcomm system suite professional 7. manufacturer: avanquest. price: 49.95. Product B is title: avanquest usa 1461 system suit 7 professional. manufacturer: avanquest usa. price: 44.99. Are Product A and Product B the same? No\n\nProduct A is title: vcomm system suite professional 7. manufacturer: avanquest. price: 49.95. Product B is title: individual software inc systemsuite 8 professional ( us pos ). manufacturer: . price: 46.42. Are Product A and Product B the same? No\n\nProduct A is title: transfer my dvd. manufacturer: avanquest. price: 29.95. Product B is title: transfer my dvd ( 6601 ). manufacturer: . price: 28.75. Are Product A and Product B the same? No\n\n",
        "netopia": "Product A is title: timbuktu remote control softwre 10 licenses for windows. manufacturer: netopia. price: . Product B is title: now software nuw45e1k now up-to-date & contact for windows. manufacturer: now software. price: 114.99. Are Product A and Product B the same? No\n\nProduct A is title: timbuktu remote control softwre 2 licenses for windows & mac. manufacturer: netopia. price: 199.99. Product B is title: netopia timbuktu pro for mac and windows 2 pack. manufacturer: . price: 207.99. Are Product A and Product B the same? Yes\n\nProduct A is title: timbuktu remote control softwre 10 licenses for windows. manufacturer: netopia. price: . Product B is title: now software nuw45e1k now up-to-date & contact for windows. manufacturer: power-on-software. price: 114.99. Are Product A and Product B the same? No\n\nProduct A is title: timbuktu remote control softwre 2 licenses for windows & mac. manufacturer: netopia. price: 199.99. Product B is title: now software nuw45e1k now up-to-date & contact for windows. manufacturer: now software. price: 114.99. Are Product A and Product B the same? No\n\n",
        "m-audio": "Product A is title: m-audio pro tools m-powered 7.3 m-audio compatible world-class production software. manufacturer: m-audio. price: 299.99. Product B is title: adobe soundbooth cs3 audio editing software windows music production software. manufacturer: . price: 198.95. Are Product A and Product B the same? No\n\nProduct A is title: m-audio pro tools m-powered 7.3 m-audio compatible world-class production software. manufacturer: m-audio. price: 299.99. Product B is title: cakewalk sonar 5 studio edition music production software. manufacturer: . price: 229.95. Are Product A and Product B the same? No\n\nProduct A is title: m-audio pro tools m-powered 7.3 m-audio compatible world-class production software. manufacturer: m-audio. price: 299.99. Product B is title: apple/emagic logic pro 7.2 software music production software. manufacturer: . price: 999.0. Are Product A and Product B the same? No\n\nProduct A is title: m-audio pro tools m-powered 7.3 m-audio compatible world-class production software. manufacturer: m-audio. price: 299.99. Product B is title: make finale allegro software music production software. manufacturer: . price: 164.95. Are Product A and Product B the same? No\n\nProduct A is title: m-audio pro tools m-powered 7.3 m-audio compatible world-class production software. manufacturer: m-audio. price: 299.99. Product B is title: cakewalk sonar home studio 6 software music production software. manufacturer: . price: 99.0. Are Product A and Product B the same? No\n\n",
        "valuesoft": "Product A is title: valuesoft tune transfer for ipod software for windows xp. manufacturer: valuesoft. price: 29.95. Product B is title: pinnacle mobile media converter software for windows ipod software. manufacturer: . price: 29.95. Are Product A and Product B the same? No\n\nProduct A is title: valuesoft tune transfer for ipod software for windows xp. manufacturer: valuesoft. price: 29.95. Product B is title: pinnacle mobile media converter software for windows ipod software. manufacturer: . price: 29.95. Are Product A and Product B the same? No\n\nProduct A is title: valuesoft tune transfer for ipod software for windows xp. manufacturer: valuesoft. price: 29.95. Product B is title: individual software professor teaches windows xp. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: valuesoft tune transfer for ipod software for windows xp. manufacturer: valuesoft. price: 29.95. Product B is title: adobe captivate 2 software for windows presentation software. manufacturer: . price: 598.95. Are Product A and Product B the same? No\n\nProduct A is title: valuesoft tune transfer for ipod software for windows xp. manufacturer: valuesoft. price: 29.95. Product B is title: adobe after effects professional 7 software for windows effects software. manufacturer: . price: 999.0. Are Product A and Product B the same? No\n\n",
        "autodesk": "Product A is title: autodesk discreet combustion 4 windows ). manufacturer: autodesk. price: . Product B is title: autodesk autodesk combustion 4 motion graphics software compositing software. manufacturer: . price: 994.95. Are Product A and Product B the same? No\n\n",
        "cakewalk": "Product A is title: pc cakewalk guitar tracks pro 3. manufacturer: cakewalk. price: 149.99. Product B is title: cakewalk guitar tracks pro 3 music production software. manufacturer: . price: 99.0. Are Product A and Product B the same? Yes\n\nProduct A is title: cakewalk sonar home studio 6. manufacturer: cakewalk. price: 139.0. Product B is title: cakewalk sonar 5 producer edition music production software. manufacturer: . price: 379.95. Are Product A and Product B the same? No\n\nProduct A is title: cakewalk sonar home studio 6 xl. manufacturer: cakewalk. price: 209.0. Product B is title: cakewalk sonar 5 studio edition music production software. manufacturer: . price: 229.95. Are Product A and Product B the same? No\n\nProduct A is title: cakewalk sonar 6 producer ed. manufacturer: cakewalk. price: 799.0. Product B is title: cakewalk sonar 6 producers edition software music production software. manufacturer: . price: 459.95. Are Product A and Product B the same? Yes\n\nProduct A is title: cakewalk project5 version 2. manufacturer: cakewalk. price: 259.0. Product B is title: cakewalk sonar 6 producers edition software music production software. manufacturer: . price: 459.95. Are Product A and Product B the same? No\n\nProduct A is title: cakewalk sonar home studio 6 xl. manufacturer: cakewalk. price: 209.0. Product B is title: cakewalk sonar home studio 6 software music production software. manufacturer: . price: 99.0. Are Product A and Product B the same? No\n\nProduct A is title: cakewalk project5 version 2. manufacturer: cakewalk. price: 259.0. Product B is title: cakewalk project5 version 2 software music production software. manufacturer: . price: 199.0. Are Product A and Product B the same? Yes\n\nProduct A is title: cakewalk sonar 6 producer ed. manufacturer: cakewalk. price: 799.0. Product B is title: cakewalk sonar 6 studio edition software music production software. manufacturer: . price: 279.95. Are Product A and Product B the same? No\n\nProduct A is title: cakewalk sonar home studio 6 xl. manufacturer: cakewalk. price: 209.0. Product B is title: cakewalk sonar home studio 6 xl software music production software. manufacturer: . price: 159.0. Are Product A and Product B the same? Yes\n\n",
        "intego": "Product A is title: contentbarrier x4 10.4 single user ( mac ). manufacturer: intego. price: 49.99. Product B is title: virusbarrier x4 10.4 for mac-10 users ( vbx4-10u ). manufacturer: . price: 316.75. Are Product A and Product B the same? No\n\nProduct A is title: virusbarrier x4 10.4 for mac-10 users. manufacturer: intego. price: 329.95. Product B is title: virusbarrier x4 10.4 for mac-10 users ( vbx4-10u ). manufacturer: . price: 316.75. Are Product A and Product B the same? Yes\n\n",
        "megasystems": "Product A is title: human body 360. manufacturer: megasystems. price: 29.99. Product B is title: topics presents : human body. manufacturer: . price: 15.9. Are Product A and Product B the same? No\n\n",
        "activision": "Product A is title: hidden expedition : titanic. manufacturer: activision. price: 19.99. Product B is title: hidden expedition : titanic. manufacturer: . price: 15.5. Are Product A and Product B the same? Yes\n\nProduct A is title: dora the explorer : dora 's world adventures. manufacturer: activision. price: 19.99. Product B is title: dora the explorer : dora world adventures pc cd-rom. manufacturer: . price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: dora the explorer : dora 's world adventures. manufacturer: activision. price: 19.99. Product B is title: clifford the big red dog thinking adventures. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\n",
        "intuit inc.": "Product A is title: quickbooks premier non-profit edition 2005. manufacturer: intuit inc.. price: 499.95. Product B is title: intuit inc 284216 qckbks prem nonprofit ed 2005. manufacturer: . price: 404.0. Are Product A and Product B the same? Yes\n\nProduct A is title: quickbooks pos : pro multistore 6.0. manufacturer: intuit inc.. price: 1399.95. Product B is title: quickbooks ( r ). manufacturer: . price: 103.99. Are Product A and Product B the same? No\n\nProduct A is title: quickbooks pos : pro multistore 6.0. manufacturer: intuit inc.. price: 1399.95. Product B is title: quickbooks ( r ) pro 2007 edition 3-user. manufacturer: . price: 499.99. Are Product A and Product B the same? No\n\nProduct A is title: quickbooks premier non-profit edition 2005. manufacturer: intuit inc.. price: 499.95. Product B is title: quickbooks ( r ). manufacturer: . price: 199.99. Are Product A and Product B the same? No\n\nProduct A is title: quickbooks pos : pro multistore 6.0. manufacturer: intuit inc.. price: 1399.95. Product B is title: quickbooks ( r ) pro 2007 edition 3-user. manufacturer: . price: 499.99. Are Product A and Product B the same? No\n\nProduct A is title: quickbooks premier non-profit edition 2005. manufacturer: intuit inc.. price: 499.95. Product B is title: quickbooks ( r ). manufacturer: . price: 199.99. Are Product A and Product B the same? No\n\n",
        "e-mu": "Product A is title: e-mu em8685 beat garden x emulator x2/proteus x2 soundset. manufacturer: e-mu. price: 129.99. Product B is title: e-mu modern symphonic orchestra for emulator x and proteus x sounds for software. manufacturer: . price: 329.99. Are Product A and Product B the same? No\n\n",
        "zone labs": "Product A is title: zonealarm internet security suite. manufacturer: zone labs. price: 49.99. Product B is title: panda software panda internet security 2007. manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\nProduct A is title: zonealarm pro firewall. manufacturer: zone labs. price: 39.99. Product B is title: zonealarm wireless security 2005. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: zonealarm internet security suite. manufacturer: zone labs. price: 49.99. Product B is title: panda software panda internet security 2007. manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\nProduct A is title: zonealarm internet security suite. manufacturer: zone labs. price: 49.99. Product B is title: pc-cillin ( tm ) internet security 2007. manufacturer: . price: 49.99. Are Product A and Product B the same? No\n\nProduct A is title: zonealarm internet security suite. manufacturer: zone labs. price: 49.99. Product B is title: zonealarm internet security suite 2007 ( pc ) zonelabs. manufacturer: . price: 49.99. Are Product A and Product B the same? Yes\n\nProduct A is title: zonealarm internet security suite. manufacturer: zone labs. price: 49.99. Product B is title: symantec 10725608 norton internet security 2007 internet security 1 user ( s ) pc. manufacturer: . price: 64.97. Are Product A and Product B the same? No\n\nProduct A is title: zonealarm pro firewall. manufacturer: zone labs. price: 39.99. Product B is title: zone labs zonealarm pro. manufacturer: . price: 30.17. Are Product A and Product B the same? Yes\n\n",
        "dataviz": "",
        "roxio": "Product A is title: roxio toast 8 titanium. manufacturer: roxio. price: 99.95. Product B is title: roxio toast 8 titanium cdr software. manufacturer: . price: 82.99. Are Product A and Product B the same? Yes\n\nProduct A is title: roxio easy media creator 9. manufacturer: roxio. price: 99.95. Product B is title: podmediacreator. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: roxio easy media creator 9. manufacturer: roxio. price: 99.95. Product B is title: roxio easy media creator 9 suite software for windows authoring software. manufacturer: . price: 98.95. Are Product A and Product B the same? Yes\n\nProduct A is title: stopzilla 4.0. manufacturer: roxio. price: 47.33. Product B is title: stopzilla ! 4.0 ( pc ) roxio. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: roxio popcorn 2 ( mac ). manufacturer: roxio. price: 49.95. Product B is title: roxio popcorn 2 video conversion software. manufacturer: . price: 44.99. Are Product A and Product B the same? Yes\n\n",
        "dk": "Product A is title: dk rainbow fish most beautiful fish. manufacturer: dk. price: 9.99. Product B is title: rainbow fish and the whale. manufacturer: . price: 6.95. Are Product A and Product B the same? No\n\n",
        "power production": "Product A is title: power production storyboard quick. manufacturer: power production. price: . Product B is title: power production power production storyboard artist v4 .2 software animation software. manufacturer: . price: 749.0. Are Product A and Product B the same? No\n\nProduct A is title: power production storyboard artist 4. manufacturer: power production. price: . Product B is title: power production storyboard quick 5 software for mac and windows animation software. manufacturer: . price: 299.0. Are Product A and Product B the same? No\n\nProduct A is title: power production storyboard artist 4. manufacturer: power production. price: . Product B is title: power production power production storyboard artist software animation software. manufacturer: . price: 498.95. Are Product A and Product B the same? Yes\n\nProduct A is title: power production storyboard quick. manufacturer: power production. price: . Product B is title: power production power production storyboard artist software animation software. manufacturer: . price: 498.95. Are Product A and Product B the same? No\n\nProduct A is title: power production storyboard quick. manufacturer: power production. price: . Product B is title: power production storyboard quick 5 software for mac and windows animation software. manufacturer: . price: 299.0. Are Product A and Product B the same? Yes\n\n",
        "on-hand-software": "Product A is title: best of card games ( jewel case ). manufacturer: on-hand-software. price: 9.99. Product B is title: total 3d home deluxe ( jewel case ) ( jce-th6 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: best of card games ( jewel case ). manufacturer: on-hand-software. price: 9.99. Product B is title: cross stitch design studio ( jewel case ) ( 8006 ). manufacturer: m-and-r-technologies. price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: best of card games ( jewel case ). manufacturer: on-hand-software. price: 9.99. Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: wild photo effects ( jewel case ). manufacturer: on-hand-software. price: 9.99. Product B is title: total 3d home deluxe ( jewel case ) ( jce-th6 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: best of card games ( jewel case ). manufacturer: on-hand-software. price: 9.99. Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\n",
        "csdc": "Product A is title: more brain games. manufacturer: csdc. price: 9.99. Product B is title: brain games : solitaire. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: more brain games. manufacturer: csdc. price: 9.99. Product B is title: freeverse software big bang brain games mac. manufacturer: . price: 24.99. Are Product A and Product B the same? No\n\nProduct A is title: more brain games. manufacturer: csdc. price: 9.99. Product B is title: hoyle card games. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: route 66 route canada ( french ). manufacturer: csdc. price: . Product B is title: csdc route 66 route canada ( french ). manufacturer: . price: 35.98. Are Product A and Product B the same? Yes\n\nProduct A is title: macjournal v4 ( mac ). manufacturer: csdc. price: 39.95. Product B is title: mariner macjournal 4 journal blog podcast software. manufacturer: . price: 38.99. Are Product A and Product B the same? Yes\n\nProduct A is title: ilisten with headset microphone 1.7. manufacturer: csdc. price: . Product B is title: macspeech ilisten with headset and microphone. manufacturer: . price: 149.99. Are Product A and Product B the same? Yes\n\nProduct A is title: omnioutliner professional 3.0. manufacturer: csdc. price: 69.95. Product B is title: diskeeper 2007 professional. manufacturer: . price: 46.95. Are Product A and Product B the same? No\n\nProduct A is title: marketcircle daylite v3 .0 ( mac ). manufacturer: csdc. price: 149.0. Product B is title: marketcircle daylite 3 customer relationship management crm. manufacturer: . price: 125.99. Are Product A and Product B the same? Yes\n\nProduct A is title: omnioutliner professional 3.0. manufacturer: csdc. price: 69.95. Product B is title: adobe cs3 flash professional academic. manufacturer: . price: 239.99. Are Product A and Product B the same? No\n\nProduct A is title: modeless liquid ledger personal finance software ( mac ). manufacturer: csdc. price: 79.99. Product B is title: modeless software liquid ledger personal finance. manufacturer: . price: 71.99. Are Product A and Product B the same? Yes\n\n",
        "sony media": "Product A is title: american idol extreme music creator. manufacturer: sony media. price: 49.95. Product B is title: sony media software american idol extreme music creator software music production software. manufacturer: . price: 49.95. Are Product A and Product B the same? Yes\n\nProduct A is title: sound forge 9.0. manufacturer: sony media. price: 399.96. Product B is title: sony creative software sfas9000 sound forge audio studio 9. manufacturer: . price: 53.97. Are Product A and Product B the same? No\n\nProduct A is title: sound forge audio studio 9. manufacturer: sony media. price: 69.59. Product B is title: sound forge audio studio 9. manufacturer: . price: 66.89. Are Product A and Product B the same? Yes\n\nProduct A is title: sound forge 9.0. manufacturer: sony media. price: 399.96. Product B is title: sound studio 3 for mac. manufacturer: . price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: acid music studio 7. manufacturer: sony media. price: 78.95. Product B is title: acid music studio 6. manufacturer: sony-pictures-digital-entertainment. price: 59.95. Are Product A and Product B the same? No\n\nProduct A is title: vegas movie studiodvd +8 plat. manufacturer: sony media. price: 129.95. Product B is title: vegas movie studiodvd +8 plat. manufacturer: sony creative software. price: 92.99. Are Product A and Product B the same? Yes\n\nProduct A is title: sony vegas movie studio + dvd 8 platinum edition. manufacturer: sony media. price: 129.95. Product B is title: sony vegas 6 video editing software professional editing software. manufacturer: sony-pictures-digital-entertainment. price: 99.0. Are Product A and Product B the same? No\n\nProduct A is title: sony sound forge 9 boxed upgrade. manufacturer: sony media. price: 232.05. Product B is title: sound forge audio studio 9. manufacturer: . price: 66.89. Are Product A and Product B the same? No\n\nProduct A is title: vegas movie studiodvd +8 plat. manufacturer: sony media. price: 129.95. Product B is title: sony vegas movie studio + dvd v8 platinum edition video editing software professional editing software. manufacturer: . price: 129.95. Are Product A and Product B the same? Yes\n\n",
        "blizzard entertainment": "Product A is title: world of warcraft 60 day pre-paid time card. manufacturer: blizzard entertainment. price: 29.99. Product B is title: world of wild cats. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: world of warcraft. manufacturer: blizzard entertainment. price: 19.99. Product B is title: starcraft ghost. manufacturer: . price: 49.99. Are Product A and Product B the same? No\n\nProduct A is title: world of warcraft 60 day pre-paid time card. manufacturer: blizzard entertainment. price: 29.99. Product B is title: vivendi blizzard world of warcraft. manufacturer: . price: 24.99. Are Product A and Product B the same? No\n\nProduct A is title: world of warcraft. manufacturer: blizzard entertainment. price: 19.99. Product B is title: vivendi blizzard world of warcraft. manufacturer: . price: 24.99. Are Product A and Product B the same? No\n\nProduct A is title: world of warcraft 60 day pre-paid time card. manufacturer: blizzard entertainment. price: 29.99. Product B is title: world of warcraft burning crusade. manufacturer: . price: 39.99. Are Product A and Product B the same? No\n\nProduct A is title: world of warcraft 60 day pre-paid time card. manufacturer: blizzard entertainment. price: 29.99. Product B is title: vivendi-universal games inc 72305 world of warcraft 60 day prepaid. manufacturer: . price: 27.99. Are Product A and Product B the same? Yes\n\nProduct A is title: world of warcraft expansion : burning crusade. manufacturer: blizzard entertainment. price: 39.99. Product B is title: world of warcraft burning crusade. manufacturer: . price: 39.99. Are Product A and Product B the same? Yes\n\n",
        "sony-creative-software": "Product A is title: station access collection. manufacturer: sony-creative-software. price: 29.99. Product B is title: adobe cs3 master collection academic. manufacturer: . price: 939.99. Are Product A and Product B the same? No\n\nProduct A is title: station access collection. manufacturer: sony-creative-software. price: 29.99. Product B is title: adobe cs3 master collection. manufacturer: . price: 2499.99. Are Product A and Product B the same? No\n\nProduct A is title: station access collection. manufacturer: sony-creative-software. price: 29.99. Product B is title: ultimate puzzle collection. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: station access collection. manufacturer: sony-creative-software. price: 29.99. Product B is title: rayman 10th anniversary collection. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: station access collection. manufacturer: sony-creative-software. price: 29.99. Product B is title: fallout collection ( dvd-rom ). manufacturer: . price: 24.49. Are Product A and Product B the same? No\n\nProduct A is title: everquest : titanium. manufacturer: sony-creative-software. price: 19.95. Product B is title: sony online entertainment inc 41032 everquest : titanium win 2000 xp. manufacturer: sony-creative-software. price: 23.09. Are Product A and Product B the same? Yes\n\n",
        "mcafee": "Product A is title: mcafee internet security suite 2007 3 users. manufacturer: mcafee. price: 89.99. Product B is title: pc-cillin internet security 2007 3-user ( win me 2000 xp ). manufacturer: . price: 49.95. Are Product A and Product B the same? No\n\nProduct A is title: mcafee total protection 2007 3 users. manufacturer: mcafee. price: 99.99. Product B is title: mtp07emb3rua mcafee total protection 2007 complete package 3 users cd mini-box wi. manufacturer: . price: 86.94. Are Product A and Product B the same? No\n\nProduct A is title: mcafee internet security suite 2007 3 users. manufacturer: mcafee. price: 89.99. Product B is title: f-secure internet security 2007 3 pcs ( global marketing partners ). manufacturer: global-marketing-partners. price: 63.35. Are Product A and Product B the same? No\n\nProduct A is title: mcafee internet security suite 2007 3 users. manufacturer: mcafee. price: 89.99. Product B is title: f-secure internet security 2007 3 pcs ( global marketing partners ). manufacturer: global-marketing-partners. price: 63.35. Are Product A and Product B the same? No\n\nProduct A is title: mcafee total protection 2007 3 users. manufacturer: mcafee. price: 99.99. Product B is title: mcafee inc total protection 2007 3-user. manufacturer: . price: 92.51. Are Product A and Product B the same? Yes\n\nProduct A is title: mcafee virusscan plus 3 user. manufacturer: mcafee. price: 59.99. Product B is title: mcafee inc mcafee virusscan plus sub 1:1 3u mb en. manufacturer: . price: 56.38. Are Product A and Product B the same? Yes\n\nProduct A is title: mcafee internet security suite 2007 3 users. manufacturer: mcafee. price: 89.99. Product B is title: symantec 10725608 norton internet security 2007 internet security 1 user ( s ) pc. manufacturer: . price: 64.97. Are Product A and Product B the same? No\n\n",
        "adobe systems": "Product A is title: adobe acrobat distiller svr v6-cd linux u/u 42050142 ). manufacturer: adobe systems. price: . Product B is title: adobe 42050106 acrobat distiller svr v6-cd sun 100u. manufacturer: adobe. price: 5399.99. Are Product A and Product B the same? No\n\n",
        "zero-g": "Product A is title: zero-g pro pack for garageband ( appleloops ). manufacturer: zero-g. price: 169.95. Product B is title: apple garageband jam pack : symphony orchestra. manufacturer: . price: 95.99. Are Product A and Product B the same? No\n\nProduct A is title: zero-g pro pack for garageband ( appleloops ). manufacturer: zero-g. price: 169.95. Product B is title: apple garageband jam pack : rhythm section. manufacturer: . price: 95.99. Are Product A and Product B the same? No\n\nProduct A is title: zero-g pro pack for garageband ( appleloops ). manufacturer: zero-g. price: 169.95. Product B is title: apple garageband jam pack : symphony orchestra. manufacturer: . price: 95.99. Are Product A and Product B the same? No\n\nProduct A is title: zero-g pro pack for garageband ( appleloops ). manufacturer: zero-g. price: 169.95. Product B is title: apple garageband jam pack : world music. manufacturer: . price: 95.99. Are Product A and Product B the same? No\n\n",
        "dreamcatcher interactive": "Product A is title: sentinel : descendants in time. manufacturer: dreamcatcher interactive. price: 19.99. Product B is title: dreamcatcher interactive sentinel : descendants in time. manufacturer: . price: 17.55. Are Product A and Product B the same? Yes\n\nProduct A is title: the guild 2 pirates of the seas. manufacturer: dreamcatcher interactive. price: 29.99. Product B is title: imagination international corp pirates of the carribean. manufacturer: . price: 21.26. Are Product A and Product B the same? No\n\nProduct A is title: 3 great games for guys. manufacturer: dreamcatcher interactive. price: 29.99. Product B is title: dreamcatcher interactive 3 great games for guys. manufacturer: . price: 26.14. Are Product A and Product B the same? Yes\n\nProduct A is title: the guild 2 pirates of the seas. manufacturer: dreamcatcher interactive. price: 29.99. Product B is title: guild 2 : pirates of the seas. manufacturer: . price: 23.99. Are Product A and Product B the same? Yes\n\n",
        "findex": "Product A is title: quickverse mac bible study 2007 gold box. manufacturer: findex. price: 349.95. Product B is title: findex.com inc quickverse bible study 2007 bible suite. manufacturer: . price: 24.89. Are Product A and Product B the same? No\n\nProduct A is title: findex quickverse bible study 2007 bible-suite. manufacturer: findex. price: . Product B is title: findex.com inc quickverse mac bible study 2007 white box. manufacturer: . price: 37.18. Are Product A and Product B the same? No\n\nProduct A is title: quickverse mac bible study 2007 gold box. manufacturer: findex. price: 349.95. Product B is title: findex.com inc quickverse mac bible study 2007 gold box. manufacturer: . price: 215.42. Are Product A and Product B the same? Yes\n\nProduct A is title: findex quickverse bible study 2007 bible-suite. manufacturer: findex. price: . Product B is title: findex.com inc quickverse bible study 2007 bible suite. manufacturer: . price: 24.89. Are Product A and Product B the same? Yes\n\nProduct A is title: quickverse mac bible study 2007 gold box. manufacturer: findex. price: 349.95. Product B is title: findex.com inc quickverse mac bible study 2007 white box. manufacturer: . price: 37.18. Are Product A and Product B the same? No\n\nProduct A is title: findex quickverse bible study 2007 bible-suite. manufacturer: findex. price: . Product B is title: findex.com quickverse bible study 2007 mobile deluxe. manufacturer: . price: 24.99. Are Product A and Product B the same? No\n\nProduct A is title: quickverse mac bible study 2007 white box. manufacturer: findex. price: 59.95. Product B is title: findex.com inc quickverse mac bible study 2007 gold box. manufacturer: . price: 215.42. Are Product A and Product B the same? No\n\n",
        "diskeeper corporation": "Product A is title: diskeeper 2007 professional single lic pack. manufacturer: diskeeper corporation. price: 79.21. Product B is title: resume maker professional 12. manufacturer: . price: 17.95. Are Product A and Product B the same? No\n\nProduct A is title: diskeeper 2007 professional single lic pack. manufacturer: diskeeper corporation. price: 79.21. Product B is title: csdc pop-up stopper professional panicware. manufacturer: . price: 26.14. Are Product A and Product B the same? No\n\nProduct A is title: diskeeper 2007 pro premier 5-lic pack. manufacturer: diskeeper corporation. price: 736.66. Product B is title: quicken 2007 premier win cd. manufacturer: intuit. price: 73.99. Are Product A and Product B the same? No\n\nProduct A is title: diskeeper 2007 professional single lic pack. manufacturer: diskeeper corporation. price: 79.21. Product B is title: diskeeper 2007 professional. manufacturer: . price: 46.95. Are Product A and Product B the same? Yes\n\nProduct A is title: diskeeper 2007 professional single lic pack. manufacturer: diskeeper corporation. price: 79.21. Product B is title: microsoft office accounting professional 2007 ( pc ). manufacturer: . price: 229.95. Are Product A and Product B the same? No\n\nProduct A is title: diskeeper 2007 pro premier 10-lic pack. manufacturer: diskeeper corporation. price: 1372.21. Product B is title: quicken 2007 premier win cd. manufacturer: intuit. price: 73.99. Are Product A and Product B the same? No\n\n",
        "total training": "Product A is title: total training for adobe creative suite 3 : design premium bundle. manufacturer: total training. price: 299.99. Product B is title: 19500220 adobe creative suite 3 design premium product upgrade package 1 user upgra. manufacturer: . price: 1687.48. Are Product A and Product B the same? No\n\nProduct A is title: total training for adobe creative suite 3 : design premium bundle. manufacturer: total training. price: 299.99. Product B is title: adobe production premium cs3 software suite for windows professional editing software. manufacturer: . price: 1698.95. Are Product A and Product B the same? No\n\nProduct A is title: total training for adobe creative suite 3 : design premium bundle. manufacturer: total training. price: 299.99. Product B is title: adobe production premium cs3 software suite for mac av production software. manufacturer: . price: 1698.95. Are Product A and Product B the same? No\n\nProduct A is title: total training for adobe creative suite 3 : design premium bundle. manufacturer: total training. price: 299.99. Product B is title: adobe creative suite 3 design standard complete package 1 user academic windows. manufacturer: . price: 369.0. Are Product A and Product B the same? No\n\nProduct A is title: total training for adobe creative suite 3 : design premium bundle. manufacturer: total training. price: 299.99. Product B is title: total training for adobe creative suite 3 design premium. manufacturer: . price: 279.95. Are Product A and Product B the same? Yes\n\nProduct A is title: total training for adobe creative suite 3 : design premium bundle. manufacturer: total training. price: 299.99. Product B is title: total training for advanced flash 8 action script. manufacturer: . price: 99.99. Are Product A and Product B the same? No\n\n",
        "scholastic": "Product A is title: math missions : the amazing arcade adventure grades 3-5. manufacturer: scholastic. price: 19.99. Product B is title: cluefinders 3rd grade adventures with adapt. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: clifford reading pack ( dvd box ). manufacturer: scholastic. price: 19.99. Product B is title: clifford thinking. manufacturer: . price: 12.9. Are Product A and Product B the same? No\n\nProduct A is title: math missions : the amazing arcade adventure grades 3-5. manufacturer: scholastic. price: 19.99. Product B is title: math missions : the race to spectacle city arcade ( grades k-2 ). manufacturer: . price: 8.49. Are Product A and Product B the same? No\n\nProduct A is title: clifford reading pack ( dvd box ). manufacturer: scholastic. price: 19.99. Product B is title: clifford the big red dog reading pack. manufacturer: . price: 19.95. Are Product A and Product B the same? No\n\nProduct A is title: math missions : the race to spectacle city arcade ( grades k-2 ). manufacturer: scholastic. price: 29.95. Product B is title: math missions : the amazing arcade adventure ( grades 3-5 ). manufacturer: . price: 12.95. Are Product A and Product B the same? No\n\n",
        "watchguard license": "",
        "zend-technologies-inc .": "Product A is title: zend studio 5.1 ( win/mac ). manufacturer: zend-technologies-inc .. price: 299.0. Product B is title: zend studio 5.1 php integrated development environment. manufacturer: zend-technologies-inc .. price: 249.0. Are Product A and Product B the same? Yes\n\nProduct A is title: zend studio 5.1 ( win/mac ). manufacturer: zend-technologies-inc .. price: 299.0. Product B is title: sony acid music studio ( pc ). manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\n",
        "sony-media-software": "Product A is title: discrete drums : vol .1. manufacturer: sony-media-software. price: 59.95. Product B is title: sony media software discrete drums : vol .1 sony sound series. manufacturer: sony-media-software. price: 24.91. Are Product A and Product B the same? Yes\n\nProduct A is title: sony super duper music looper. manufacturer: sony-media-software. price: 19.99. Product B is title: sony acid music studio ( pc ). manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\nProduct A is title: photo go. manufacturer: sony-media-software. price: 24.95. Product B is title: books to go. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\n",
        "yamaha": "Product A is title: mike garson jazzical suite. manufacturer: yamaha. price: 34.95. Product B is title: mike garson the burt bacharach songbook. manufacturer: . price: 25.77. Are Product A and Product B the same? No\n\nProduct A is title: the little mermaid/beauty and the beast. manufacturer: yamaha. price: 34.95. Product B is title: phillip keveren disney 's the little mermaid/beauty and the beast. manufacturer: . price: 23.59. Are Product A and Product B the same? Yes\n\nProduct A is title: academy award winners. manufacturer: yamaha. price: 31.95. Product B is title: bobby walters academy award winners. manufacturer: . price: 23.59. Are Product A and Product B the same? Yes\n\nProduct A is title: hollywood volume 1. manufacturer: yamaha. price: 31.95. Product B is title: phillip keveren 501056 hollywood volume 1. manufacturer: . price: 23.59. Are Product A and Product B the same? Yes\n\nProduct A is title: sunset boulevard. manufacturer: yamaha. price: 31.95. Product B is title: mike garson andrew lloyd webber andrew lloyd webber favorites featuring sunset boulevard. manufacturer: . price: 23.59. Are Product A and Product B the same? Yes\n\nProduct A is title: big tiny little a tiny little christmas. manufacturer: yamaha. price: 34.95. Product B is title: big tiny little big tiny little a tiny little christmas. manufacturer: . price: 23.59. Are Product A and Product B the same? Yes\n\n",
        "global software": "Product A is title: global software smart steps 2nd grade windows macintosh ). manufacturer: global software. price: . Product B is title: global marketing partners art text belight software. manufacturer: . price: 43.29. Are Product A and Product B the same? No\n\nProduct A is title: global software smart steps 2nd grade windows macintosh ). manufacturer: global software. price: . Product B is title: global software a0447jch dk smart steps 2nd grade win 95 98 me 2000 xp vista. manufacturer: . price: 9.08. Are Product A and Product B the same? Yes\n\nProduct A is title: global software smart steps 2nd grade windows macintosh ). manufacturer: global software. price: . Product B is title: ilford studio software for pc ( windows 2000 pro xp pro ). manufacturer: . price: 699.99. Are Product A and Product B the same? No\n\nProduct A is title: global software smart steps 2nd grade windows macintosh ). manufacturer: global software. price: . Product B is title: phase one capture-one pro dslr raw image editing software for macintosh & windows .. manufacturer: . price: 399.95. Are Product A and Product B the same? No\n\nProduct A is title: global software smart steps 1st grade windows macintosh ). manufacturer: global software. price: . Product B is title: global software a0454jch dk smart steps 3rd grade win 95 98 me 2000 xp vista. manufacturer: global-software-publishing. price: 9.08. Are Product A and Product B the same? No\n\nProduct A is title: global software smart steps 1st grade windows macintosh ). manufacturer: global software. price: . Product B is title: phase one capture-one pro dslr raw image editing software for macintosh & windows .. manufacturer: . price: 399.95. Are Product A and Product B the same? No\n\n",
        "stomp inc": "Product A is title: data protection suite. manufacturer: stomp inc. price: 69.99. Product B is title: arcserve bkup protection suite exch r11 .5 upg-v prod only. manufacturer: . price: 2249.99. Are Product A and Product B the same? No\n\n",
        "avanquest-publishing-usa-inc .": "Product A is title: power director 3. manufacturer: avanquest-publishing-usa-inc .. price: 79.95. Product B is title: cyberlink power director 3. manufacturer: avanquest-publishing-usa-inc .. price: 9.99. Are Product A and Product B the same? Yes\n\n",
        "adobe-education-box": "Product A is title: photoshop elements 4 edu mac 1u. manufacturer: adobe-education-box. price: 69.0. Product B is title: adobe systems inc 29180248 adobe photoshop elements 5.0 / premiere elements 3.0. manufacturer: . price: 139.51. Are Product A and Product B the same? No\n\nProduct A is title: dreamweaver cs3 9 mac ed 1u. manufacturer: adobe-education-box. price: 199.0. Product B is title: adobe dreamweaver cs3 complete package 1 user academic cd mac. manufacturer: . price: 189.0. Are Product A and Product B the same? No\n\nProduct A is title: photoshop elements 4 edu mac 1u. manufacturer: adobe-education-box. price: 69.0. Product B is title: adobe photoshop elements 4.0 mac academic. manufacturer: adobe-education-box. price: 69.99. Are Product A and Product B the same? Yes\n\nProduct A is title: dreamweaver cs3 9 mac ed 1u. manufacturer: adobe-education-box. price: 199.0. Product B is title: adobe dreamweaver cs3 academic. manufacturer: adobe-education-box. price: 195.99. Are Product A and Product B the same? Yes\n\nProduct A is title: illustrator cs3 13 mac ed 1u. manufacturer: adobe-education-box. price: 199.0. Product B is title: adobe illustrator cs3 for mac academic. manufacturer: adobe-education-box. price: 199.99. Are Product A and Product B the same? Yes\n\nProduct A is title: audition 2 win ed ue 1u. manufacturer: adobe-education-box. price: 149.0. Product B is title: adobe 22011152 audition2 win retail ue 1u. manufacturer: . price: 296.2. Are Product A and Product B the same? No\n\nProduct A is title: illustrator cs3 13 mac ed 1u. manufacturer: adobe-education-box. price: 199.0. Product B is title: adobe illustrator cs3 for mac academic. manufacturer: adobe-education-box. price: 199.99. Are Product A and Product B the same? Yes\n\nProduct A is title: illustrator cs3 13 mac ed 1u. manufacturer: adobe-education-box. price: 199.0. Product B is title: adobe illustrator cs3 ( mac ). manufacturer: . price: 599.0. Are Product A and Product B the same? No\n\nProduct A is title: dreamweaver cs3 9 mac ed 1u. manufacturer: adobe-education-box. price: 199.0. Product B is title: adobe dreamweaver cs3. manufacturer: . price: 415.99. Are Product A and Product B the same? No\n\n",
        "global marketing partners": "Product A is title: readiris pro 11. manufacturer: global marketing partners. price: 129.99. Product B is title: readiris pro 11 corporate edition for pc. manufacturer: . price: 349.99. Are Product A and Product B the same? No\n\nProduct A is title: readiris pro 11. manufacturer: global marketing partners. price: 129.99. Product B is title: iris inc readiris pro 11 corporate edition. manufacturer: . price: 430.58. Are Product A and Product B the same? No\n\nProduct A is title: cadopia 6 standard edition by cadopia. manufacturer: global marketing partners. price: 295.0. Product B is title: goldmine ( r ) standard edition 6.7. manufacturer: . price: 199.99. Are Product A and Product B the same? No\n\n",
        "knowledge adventure": "Product A is title: jumpstart advanced preschool v2 .0. manufacturer: knowledge adventure. price: 19.99. Product B is title: knowledge adventure inc. jumpstart advanced preschool v2 .0. manufacturer: . price: 16.93. Are Product A and Product B the same? Yes\n\nProduct A is title: jumpstart reading with karaoke. manufacturer: knowledge adventure. price: 29.99. Product B is title: knowledge adventure 20033 jumpstart reading with karaoke. manufacturer: knowledge adventure. price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: jumpstart world 2nd grade. manufacturer: knowledge adventure. price: 19.99. Product B is title: jumpstart kindergarten ( jc ). manufacturer: . price: 9.9. Are Product A and Product B the same? No\n\nProduct A is title: jumpstart world 1st grade. manufacturer: knowledge adventure. price: 19.99. Product B is title: knowledge adventure inc. 20223 jumpstart world 1st grade mini box. manufacturer: . price: 16.93. Are Product A and Product B the same? Yes\n\nProduct A is title: jumpstart advanced 1st grade v2 .0. manufacturer: knowledge adventure. price: 19.99. Product B is title: knowledge adventure inc. 20223 jumpstart world 1st grade mini box. manufacturer: . price: 16.93. Are Product A and Product B the same? No\n\nProduct A is title: jumpstart 4th-6th grade v 2.0 b43. manufacturer: knowledge adventure. price: 19.99. Product B is title: knowledge adventure inc. jumpstart 4th-6th grade v 2.0 b43. manufacturer: . price: 16.93. Are Product A and Product B the same? Yes\n\nProduct A is title: jumpstart advanced 3rd grade v2 .0. manufacturer: knowledge adventure. price: 19.99. Product B is title: jumpstart advanced 3rd grade ( pc/mac ) knowledge adventure. manufacturer: . price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: jumpstart world 2nd grade. manufacturer: knowledge adventure. price: 19.99. Product B is title: jumpstart 1st grade ( jc ). manufacturer: . price: 9.9. Are Product A and Product B the same? No\n\nProduct A is title: jumpstart advanced preschool v2 .0. manufacturer: knowledge adventure. price: 19.99. Product B is title: jumpstart advanced 3rd grade ( pc/mac ) knowledge adventure. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: jumpstart advanced preschool v2 .0. manufacturer: knowledge adventure. price: 19.99. Product B is title: jumpstart advanced 2nd grade ( pc/mac ) knowledge adventure. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\n",
        "aladdin": "",
        "global-marketing-partners": "Product A is title: video patrol 5.0 by honest technology. manufacturer: global-marketing-partners. price: 50.86. Product B is title: global marketing partners video patrol 5.0 honest technology. manufacturer: global-marketing-partners. price: 25.2. Are Product A and Product B the same? Yes\n\nProduct A is title: swift publisher for mac. manufacturer: global-marketing-partners. price: 39.99. Product B is title: belight swift publisher page layout software. manufacturer: global-marketing-partners. price: 39.99. Are Product A and Product B the same? Yes\n\nProduct A is title: f-secure internet security 2007 3 pcs. manufacturer: global-marketing-partners. price: 79.99. Product B is title: symantec 10725608 norton internet security 2007 internet security 1 user ( s ) pc. manufacturer: . price: 64.97. Are Product A and Product B the same? No\n\nProduct A is title: gmp data rescue license 24300 ). manufacturer: global-marketing-partners. price: 349.0. Product B is title: prosoft engineering data rescue universal license. manufacturer: global-marketing-partners. price: 236.34. Are Product A and Product B the same? Yes\n\nProduct A is title: f-secure internet security 2007 3 pcs. manufacturer: global-marketing-partners. price: 79.99. Product B is title: symantec norton internet security 3.0 software for macintosh apple & mac software. manufacturer: . price: 89.95. Are Product A and Product B the same? No\n\nProduct A is title: drive genius by prosoft engineering inc. manufacturer: global-marketing-partners. price: 129.0. Product B is title: prosoft engineering data rescue universal license. manufacturer: global-marketing-partners. price: 236.34. Are Product A and Product B the same? No\n\nProduct A is title: easy video editor 2.0 by honest technology. manufacturer: global-marketing-partners. price: 50.88. Product B is title: global marketing partners easy video editor 2.0 honest technology. manufacturer: global-marketing-partners. price: 25.21. Are Product A and Product B the same? Yes\n\nProduct A is title: swift publisher for mac. manufacturer: global-marketing-partners. price: 39.99. Product B is title: allume poser 6 for mac. manufacturer: curious-labs-inc .. price: 169.99. Are Product A and Product B the same? No\n\nProduct A is title: f-secure internet security 2007 3 pcs. manufacturer: global-marketing-partners. price: 79.99. Product B is title: mcafee internet security 7.0. manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\n",
        "sierra home": "Product A is title: hallmark card studio 2006. manufacturer: sierra home. price: 29.99. Product B is title: hallmark card studio 2004. manufacturer: . price: 29.97. Are Product A and Product B the same? No\n\nProduct A is title: hallmark card studio 2005 deluxe. manufacturer: sierra home. price: 49.99. Product B is title: hallmark card studio 2004. manufacturer: . price: 29.97. Are Product A and Product B the same? No\n\nProduct A is title: hallmark card studio 2005 deluxe. manufacturer: sierra home. price: 49.99. Product B is title: hgw hallmark card studio 2005 deluxe complete package 1 user cd win. manufacturer: . price: 47.43. Are Product A and Product B the same? No\n\nProduct A is title: hallmark card studio 2006. manufacturer: sierra home. price: 29.99. Product B is title: sierrahome hse hallmark card studio special edition win 98 me 2000 xp. manufacturer: sierrahome. price: 19.26. Are Product A and Product B the same? No\n\nProduct A is title: hallmark card studio 2005 deluxe. manufacturer: sierra home. price: 49.99. Product B is title: sierrahome hse hallmark card studio special edition win 98 me 2000 xp. manufacturer: sierrahome. price: 19.26. Are Product A and Product B the same? No\n\n",
        "viva media": "Product A is title: learn to play chess with fritz and chesster ( mac ). manufacturer: viva media. price: 29.99. Product B is title: viva media llc learn to play chess with fritz & chesster 2. manufacturer: . price: 26.14. Are Product A and Product B the same? No\n\nProduct A is title: my sim aquarium. manufacturer: viva media. price: 19.99. Product B is title: marine aquarium & goldfish aquarium. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: learn to play chess with fritz and chesster ( mac ). manufacturer: viva media. price: 29.99. Product B is title: viva media llc learn to play chess with fritz & chesster. manufacturer: . price: 26.14. Are Product A and Product B the same? Yes\n\n",
        "palo-alto-software": "Product A is title: palo alto marketing plan pro 9.0. manufacturer: palo-alto-software. price: 179.0. Product B is title: palo alto software business plan pro 2007. manufacturer: palo-alto-software. price: 88.83. Are Product A and Product B the same? No\n\nProduct A is title: palo alto business plan pro premier 2007. manufacturer: palo-alto-software. price: 199.0. Product B is title: quicken ( r ) premier 2007. manufacturer: . price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: palo alto business plan pro premier 2007. manufacturer: palo-alto-software. price: 199.0. Product B is title: business plan pro ( r ) 2004. manufacturer: . price: 99.87. Are Product A and Product B the same? No\n\nProduct A is title: palo alto marketing plan pro 9.0. manufacturer: palo-alto-software. price: 179.0. Product B is title: palo alto software mpd9 palo alto marketing plan v. 9.0 pro complete product business planning 1 user complete product standard pc. manufacturer: palo-alto-software. price: 149.97. Are Product A and Product B the same? Yes\n\nProduct A is title: palo alto business plan pro premier 2007. manufacturer: palo-alto-software. price: 199.0. Product B is title: palo alto software bpdx palo alto business plan 2007 pro premier edition complete product business planning 1 user complete product standard pc. manufacturer: palo-alto-software. price: 184.97. Are Product A and Product B the same? Yes\n\nProduct A is title: palo alto business plan pro premier 2007. manufacturer: palo-alto-software. price: 199.0. Product B is title: myob premier accounting for small business 2007-windows. manufacturer: . price: 289.99. Are Product A and Product B the same? No\n\nProduct A is title: palo alto business plan pro 2007. manufacturer: palo-alto-software. price: 99.95. Product B is title: palo alto software bppx palo alto business plan 2007 pro complete product business planning 1 user complete product standard pc. manufacturer: . price: 83.97. Are Product A and Product B the same? Yes\n\nProduct A is title: palo alto business plan pro 2007. manufacturer: palo-alto-software. price: 99.95. Product B is title: palo alto software mpd9 palo alto marketing plan v. 9.0 pro complete product business planning 1 user complete product standard pc. manufacturer: palo-alto-software. price: 149.97. Are Product A and Product B the same? No\n\nProduct A is title: palo alto business plan pro 2007. manufacturer: palo-alto-software. price: 99.95. Product B is title: palo alto software business plan pro 2007. manufacturer: palo-alto-software. price: 88.83. Are Product A and Product B the same? Yes\n\n",
        "ubi soft": "Product A is title: ubi soft star wars activity center windows ). manufacturer: ubi soft. price: 29.99. Product B is title: the beginners bible : noah 's ark activity center : activity center. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: rayman 10th anniversary collection. manufacturer: ubi soft. price: 19.99. Product B is title: rayman 10th anniversary collection. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: ubi soft star wars activity center windows ). manufacturer: ubi soft. price: 29.99. Product B is title: star wars episode iii activity center. manufacturer: . price: 10.95. Are Product A and Product B the same? No\n\nProduct A is title: rayman 10th anniversary collection. manufacturer: ubi soft. price: 19.99. Product B is title: ultimate puzzle collection. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\n",
        "nuance-communications-inc .": "Product A is title: dragon naturally speaking standard v9. manufacturer: nuance-communications-inc .. price: 99.99. Product B is title: a209s-gd4-9 .0 dragon naturallyspeaking professional v. 9 complete package 1 user c. manufacturer: . price: 786.91. Are Product A and Product B the same? No\n\nProduct A is title: ibm viavoice advanced v. 10. manufacturer: nuance-communications-inc .. price: . Product B is title: nuance communications inc. ibm viavoice advanced v. 10. manufacturer: nuance-communications-inc .. price: 57.54. Are Product A and Product B the same? Yes\n\nProduct A is title: textbridge pro 11. manufacturer: nuance-communications-inc .. price: 79.99. Product B is title: scansoft textbridge pro v. 11. manufacturer: nuance-communications-inc .. price: 68.99. Are Product A and Product B the same? Yes\n\n",
        "cisco-systems-enterprise": "Product A is title: cisco security server agent for windows and solaris license 100 agents win solaris csa-b100-srvr-k9 ). manufacturer: cisco-systems-enterprise. price: . Product B is title: cisco csa-b250-dtop-k9 cisco security desktop agent win + sol 250 agent bundle cisco security desktop agent license 250 agents 0746320832500. manufacturer: . price: 10267.54. Are Product A and Product B the same? No\n\n",
        "carry-a-tune technologies": "",
        "computer associates": "Product A is title: ca arcserve bkup protection suite exch r11 .5 upgrade prod only. manufacturer: computer associates. price: . Product B is title: arcserve bkup protection suite exch r11 .5 upg-v prod only. manufacturer: . price: 2249.99. Are Product A and Product B the same? Yes\n\nProduct A is title: ca arcserve bkup r11 .5 win agent mac os x upgrade prod only. manufacturer: computer associates. price: . Product B is title: arcserve bkup protection suite exch r11 .5 upg-v prod only. manufacturer: . price: 2249.99. Are Product A and Product B the same? No\n\nProduct A is title: ca arcserve bkup r11 .5 win client vss software snap-shot - prod only. manufacturer: computer associates. price: . Product B is title: computer associates babwbr1151s33 brightstor arcserve backup client for vss software snap-shot v. 11.5 for windows with service pack 1 add-on backup & recovery 1 server ( s ) complete product standard multilingual pc. manufacturer: . price: 799.97. Are Product A and Product B the same? Yes\n\nProduct A is title: ca internet security suite 2007 3 user. manufacturer: computer associates. price: 69.95. Product B is title: norton internet security ( v3 .0 ) mac symantec 10069614-in. manufacturer: . price: 83.43. Are Product A and Product B the same? No\n\nProduct A is title: ca brightstor arcserve backup r11 .1 for netware tape library option service pack. manufacturer: computer associates. price: . Product B is title: e17886d symantec backup exec for netware servers san shared storage option v. 9.2 ). manufacturer: . price: 654.28. Are Product A and Product B the same? No\n\nProduct A is title: ca antivirus 2007. manufacturer: computer associates. price: 39.95. Product B is title: ca antivirus 2007. manufacturer: . price: 32.28. Are Product A and Product B the same? Yes\n\nProduct A is title: ca internet security suite 2007 3 user. manufacturer: computer associates. price: 69.95. Product B is title: quickbooks pro 2007 3 user. manufacturer: intuit. price: 398.99. Are Product A and Product B the same? No\n\nProduct A is title: ca brightstor arcserve backup r11 .1 for netware upgrade from arcserve 7 for netware. manufacturer: computer associates. price: . Product B is title: babwbr1151s04 brightstor arcserve backup serverless backup option for windows w/sp1 v. 11 .. manufacturer: . price: 1031.41. Are Product A and Product B the same? No\n\n",
        "netgear": "Product A is title: netgear vpn01l vpn client software. manufacturer: netgear. price: 89.99. Product B is title: prosafe vpn client software single user netgear inc. vpn01l. manufacturer: . price: 27.78. Are Product A and Product B the same? No\n\nProduct A is title: netgear vpn01l vpn client software. manufacturer: netgear. price: 89.99. Product B is title: software. manufacturer: . price: 112.95. Are Product A and Product B the same? No\n\nProduct A is title: netgear vpn01l vpn client software. manufacturer: netgear. price: 89.99. Product B is title: software. manufacturer: . price: 46.99. Are Product A and Product B the same? No\n\nProduct A is title: netgear vpn01l vpn client software. manufacturer: netgear. price: 89.99. Product B is title: d-link vpn client software ( 5-user license ) ds-605 ds-605. manufacturer: . price: 173.99. Are Product A and Product B the same? No\n\nProduct A is title: netgear vpn01l vpn client software. manufacturer: netgear. price: 89.99. Product B is title: procurve vpn client software unlimited client licenses. manufacturer: . price: 2598.96. Are Product A and Product B the same? No\n\n",
        "century software": "Product A is title: century software tinyterm plus v. 4.0 complete product. manufacturer: century software. price: . Product B is title: century software ttpl-1 century tinyterm plus v. 4.0 complete product connectivity/data acess 1 user ( s ) complete product standard english pc. manufacturer: . price: 188.97. Are Product A and Product B the same? Yes\n\nProduct A is title: century software tinyterm plus v. 4.0 complete product. manufacturer: century software. price: . Product B is title: encore software 25005 encore marine aquarium v. 2.0 complete product screen saver 1 user ( s ) complete product standard pc mac. manufacturer: . price: 18.97. Are Product A and Product B the same? No\n\nProduct A is title: century software tinyterm plus v. 4.0 complete product. manufacturer: century software. price: . Product B is title: onone software inc. mpr-40211 on1 mask pro v. 4.0 complete product graphics/designing 1 user complete product standard pc intel-based mac mac. manufacturer: . price: 148.97. Are Product A and Product B the same? No\n\nProduct A is title: century software tinyterm plus v. 4.0 complete product. manufacturer: century software. price: . Product B is title: laplink software pafglle104000p0rtpen laplink everywhere v. 4.0 retail edition remote management complete product remote management complete product pc. manufacturer: . price: 39.97. Are Product A and Product B the same? No\n\nProduct A is title: century software tinyterm plus v. 4.0 complete product. manufacturer: century software. price: . Product B is title: encore software 10033 encore the print shop v. 21.0 complete product print management 1 user ( s ) complete product standard pc. manufacturer: . price: 25.97. Are Product A and Product B the same? No\n\nProduct A is title: century software tinyterm plus v. 4.0 complete product. manufacturer: century software. price: . Product B is title: century software tt-5 century tinyterm v. 4.2 x complete product emulation 5 user ( s ) complete product standard pc. manufacturer: . price: 395.97. Are Product A and Product B the same? No\n\n",
        "pinnacle": "Product A is title: pinnacle mobile media converter. manufacturer: pinnacle. price: 29.99. Product B is title: pinnacle mobile media converter software for windows ipod software. manufacturer: . price: 29.95. Are Product A and Product B the same? Yes\n\nProduct A is title: studio movieboard plus pci hardware. manufacturer: pinnacle. price: 251.55. Product B is title: studio plus version 11. manufacturer: pinnacle systems. price: 90.99. Are Product A and Product B the same? No\n\nProduct A is title: pinnacle studio plus titanium edition v10 .5. manufacturer: pinnacle. price: 69.99. Product B is title: pinnacle studio ultimate v11 video editing software for windows consumer video editing software. manufacturer: . price: 129.95. Are Product A and Product B the same? No\n\nProduct A is title: pinnacle mobile media organizer. manufacturer: pinnacle. price: 49.99. Product B is title: pinnacle mobile media converter software for windows ipod software. manufacturer: . price: 29.95. Are Product A and Product B the same? No\n\nProduct A is title: pinnacle mobile media converter. manufacturer: pinnacle. price: 29.99. Product B is title: pinnacle mobile media organizer software for windows ipod software. manufacturer: . price: 39.95. Are Product A and Product B the same? No\n\n",
        "stompsoft inc.": "Product A is title: digital vault. manufacturer: stompsoft inc.. price: 29.99. Product B is title: digital video & audio prod book. manufacturer: sony-pictures-digital-entertainment. price: 27.5. Are Product A and Product B the same? No\n\nProduct A is title: digital vault. manufacturer: stompsoft inc.. price: 29.99. Product B is title: digital video & audio prod book. manufacturer: sony-pictures-digital-entertainment. price: 27.5. Are Product A and Product B the same? No\n\nProduct A is title: stompsoft digital file shredder pro. manufacturer: stompsoft inc.. price: 24.99. Product B is title: migo software digital file shredder 2005. manufacturer: . price: 24.58. Are Product A and Product B the same? Yes\n\n",
        "riverdeep": "Product A is title: riverdeep adventure workshop 4th-6th grade 8th edition. manufacturer: riverdeep. price: . Product B is title: encore software 10119 adventure workshop 4th-6th grade 7th edition win xp. manufacturer: . price: 18.97. Are Product A and Product B the same? No\n\nProduct A is title: riverdeep adventure workshop 4th-6th grade 8th edition. manufacturer: riverdeep. price: . Product B is title: adventure workshop 1st-3rd grade . 8th edition ( pc ) encore. manufacturer: . price: 19.95. Are Product A and Product B the same? No\n\nProduct A is title: riverdeep adventure workshop 4th-6th grade 8th edition. manufacturer: riverdeep. price: . Product B is title: adventure workshop preschool-1st grade 8th edition win. manufacturer: . price: 18.36. Are Product A and Product B the same? No\n\nProduct A is title: riverdeep arthur 's kindergarten. manufacturer: riverdeep. price: . Product B is title: tlc arthur 's kindergarten learning system ( pc ) encore. manufacturer: . price: 19.95. Are Product A and Product B the same? Yes\n\nProduct A is title: riverdeep the print shop zoom dvd. manufacturer: riverdeep. price: . Product B is title: encore software 11941 the print shop zoom. manufacturer: . price: 17.97. Are Product A and Product B the same? Yes\n\nProduct A is title: riverdeep arthur 's kindergarten. manufacturer: riverdeep. price: . Product B is title: jumpstart kindergarten ( jc ). manufacturer: . price: 9.9. Are Product A and Product B the same? No\n\nProduct A is title: treasure mountain ! ( jewel case ) ages 5-9 for win/mac. manufacturer: riverdeep. price: 9.99. Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\n",
        "laughing bird": "Product A is title: the logo creator ( mac ). manufacturer: laughing bird. price: 39.99. Product B is title: laughingbird the logo creator. manufacturer: . price: 34.99. Are Product A and Product B the same? Yes\n\nProduct A is title: net ad creator. manufacturer: laughing bird. price: 39.99. Product B is title: laughingbird the logo creator. manufacturer: . price: 34.99. Are Product A and Product B the same? No\n\nProduct A is title: net ad creator. manufacturer: laughing bird. price: 39.99. Product B is title: laughingbird the logo creator. manufacturer: . price: 34.99. Are Product A and Product B the same? No\n\n",
        "absolute software": "Product A is title: computrace lojack for laptops 3 year license ( mac ). manufacturer: absolute software. price: 99.95. Product B is title: absolute software lfl-m2-36 computrace lojack for laptops 3 year license mac 10.2 or higher. manufacturer: . price: 80.97. Are Product A and Product B the same? Yes\n\nProduct A is title: computrace lojack for laptops 4 year subscription. manufacturer: absolute software. price: 119.99. Product B is title: absolute software lfl-n2-36 computrace lojack for laptops 3 year license ( win 2000 xp ). manufacturer: . price: 85.97. Are Product A and Product B the same? No\n\nProduct A is title: computrace lojack for laptops 4 year subscription. manufacturer: absolute software. price: 119.99. Product B is title: absolute software lfl-n2-48 computrace lojack for laptops 4 year license. manufacturer: . price: 105.68. Are Product A and Product B the same? Yes\n\nProduct A is title: computrace lojack for laptops 1 year license ( mac ). manufacturer: absolute software. price: 49.95. Product B is title: lfl-n2-12 computrace lojack for laptops subscription package 1 year 1 notebook m. manufacturer: . price: 39.95. Are Product A and Product B the same? No\n\nProduct A is title: computrace lojack for laptops : 3 year subscription. manufacturer: absolute software. price: 99.99. Product B is title: absolute software lfl-n2-36 computrace lojack for laptops 3 year license ( win 2000 xp ). manufacturer: . price: 85.97. Are Product A and Product B the same? Yes\n\nProduct A is title: computrace lojack for laptops 3 year license ( mac ). manufacturer: absolute software. price: 99.95. Product B is title: computrace lojack for laptops 3 yr subscription. manufacturer: . price: 89.99. Are Product A and Product B the same? Yes\n\nProduct A is title: computrace lojack for laptops 3 year license ( mac ). manufacturer: absolute software. price: 99.95. Product B is title: absolute software lfl-n2-36 computrace lojack for laptops 3 year license ( win 2000 xp ). manufacturer: . price: 85.97. Are Product A and Product B the same? No\n\nProduct A is title: computrace lojack for laptops 1 year license ( mac ). manufacturer: absolute software. price: 49.95. Product B is title: lfl-n2-12 computrace lojack for laptops subscription package 1 year 1 notebook m. manufacturer: . price: 39.95. Are Product A and Product B the same? No\n\nProduct A is title: computrace lojack for laptops : 3 year subscription. manufacturer: absolute software. price: 99.99. Product B is title: absolute software lfl-m2-12 computrace lojack for laptops 1 year license mac 10.2 or higher. manufacturer: . price: 44.97. Are Product A and Product B the same? No\n\nProduct A is title: computrace lojack for laptops : 1 year subscription. manufacturer: absolute software. price: 49.99. Product B is title: absolute software lfl-n2-12 computrace lojack for laptops 1 year license ( win 2000 xp ). manufacturer: . price: 41.97. Are Product A and Product B the same? Yes\n\n",
        "individual": "Product A is title: individual small business advantage deluxe 2006. manufacturer: individual. price: 49.99. Product B is title: high school success deluxe 2006. manufacturer: . price: 7.9. Are Product A and Product B the same? No\n\nProduct A is title: individual small business advantage deluxe 2006. manufacturer: individual. price: 49.99. Product B is title: individual software prm-bp5 business planmaker 2006 deluxe win 98 me 2000 xp. manufacturer: . price: 42.97. Are Product A and Product B the same? No\n\nProduct A is title: individual small business advantage deluxe 2006. manufacturer: individual. price: 49.99. Product B is title: individual software inc typing instructor deluxe j/c. manufacturer: . price: 6.84. Are Product A and Product B the same? No\n\nProduct A is title: individual small business advantage deluxe 2006. manufacturer: individual. price: 49.99. Product B is title: individual software anytime organizer deluxe 10. manufacturer: . price: 27.99. Are Product A and Product B the same? No\n\nProduct A is title: individual small business advantage deluxe 2006. manufacturer: individual. price: 49.99. Product B is title: elementary school success deluxe 2006. manufacturer: . price: 12.95. Are Product A and Product B the same? No\n\n",
        "avanquest software": "Product A is title: britannica deluxe. manufacturer: avanquest software. price: 29.95. Product B is title: britannica deluxe 2008. manufacturer: . price: 26.5. Are Product A and Product B the same? No\n\nProduct A is title: britannica deluxe. manufacturer: avanquest software. price: 29.95. Product B is title: business planmaker deluxe. manufacturer: . price: 49.99. Are Product A and Product B the same? No\n\nProduct A is title: britannica deluxe. manufacturer: avanquest software. price: 29.95. Product B is title: vhs to dvd 3.0 deluxe. manufacturer: . price: 99.99. Are Product A and Product B the same? No\n\nProduct A is title: britannica deluxe. manufacturer: avanquest software. price: 29.95. Product B is title: photoshow deluxe. manufacturer: . price: 39.92. Are Product A and Product B the same? No\n\nProduct A is title: britannica deluxe. manufacturer: avanquest software. price: 29.95. Product B is title: freeverse 3d bridge deluxe. manufacturer: . price: 24.99. Are Product A and Product B the same? No\n\n",
        "susteen": "Product A is title: cell phone software solution. manufacturer: susteen. price: 113.1. Product B is title: pcmover software utility. manufacturer: . price: 39.96. Are Product A and Product B the same? No\n\nProduct A is title: cell phone software solution. manufacturer: susteen. price: 113.1. Product B is title: freeverse software 005 solace. manufacturer: freeverse software. price: 18.99. Are Product A and Product B the same? No\n\nProduct A is title: datapilot cell phone data transfer suite universal. manufacturer: susteen. price: 79.99. Product B is title: susteen data pilot universal essentials. manufacturer: . price: 59.99. Are Product A and Product B the same? No\n\nProduct A is title: cell phone software solution. manufacturer: susteen. price: 113.1. Product B is title: freeverse software 005 solace. manufacturer: freeverse software. price: 18.99. Are Product A and Product B the same? No\n\n",
        "h & r block": "Product A is title: h & r block taxcut 2006 premium federal + state + efile. manufacturer: h & r block. price: 59.99. Product B is title: h & r block taxcut home and business with state and e-file efile software for windows tax & finance software. manufacturer: . price: 73.95. Are Product A and Product B the same? No\n\nProduct A is title: h & r block taxcut 2006 premium federal + state + efile. manufacturer: h & r block. price: 59.99. Product B is title: h & r block taxcut home and business with state e-file efile and sage act ! 8 win tax & finance software. manufacturer: . price: 73.95. Are Product A and Product B the same? No\n\nProduct A is title: h & r block taxcut 2006 home & business + state. manufacturer: h & r block. price: 79.99. Product B is title: h & r block taxcut premium federal and state with e-file efile software mac/win tax & finance software. manufacturer: . price: 54.95. Are Product A and Product B the same? No\n\nProduct A is title: h & r block taxcut 2006 premium federal + state with usb 256m flashdrive. manufacturer: h & r block. price: 39.99. Product B is title: h & r block taxcut premium federal and state with e-file efile software mac/win tax & finance software. manufacturer: . price: 54.95. Are Product A and Product B the same? No\n\nProduct A is title: h & r block taxcut 2006 home & business + state. manufacturer: h & r block. price: 79.99. Product B is title: h & r block taxcut home and business with state and e-file efile software for windows tax & finance software. manufacturer: . price: 73.95. Are Product A and Product B the same? Yes\n\nProduct A is title: h & r block taxcut 2006 premium federal + state with usb 256m flashdrive. manufacturer: h & r block. price: 39.99. Product B is title: h & r block taxcut home and business with state e-file efile and sage act ! 8 win tax & finance software. manufacturer: . price: 73.95. Are Product A and Product B the same? No\n\nProduct A is title: h & r block taxcut 2006 premium federal + state + efile. manufacturer: h & r block. price: 59.99. Product B is title: h & r block taxcut premium federal and state with e-file efile software mac/win tax & finance software. manufacturer: . price: 54.95. Are Product A and Product B the same? Yes\n\n",
        "bias": "Product A is title: bias peak le 5. manufacturer: bias. price: 129.0. Product B is title: bias peak pro 5 software music production software. manufacturer: . price: 499.0. Are Product A and Product B the same? No\n\nProduct A is title: bias peak pro 5 macintosh ). manufacturer: bias. price: 599.0. Product B is title: bias peak pro 5 software music production software. manufacturer: . price: 499.0. Are Product A and Product B the same? Yes\n\nProduct A is title: bias deck le 3.5 macintosh cd. manufacturer: bias. price: 99.0. Product B is title: bias deck 3.5 software music production software. manufacturer: . price: 299.0. Are Product A and Product B the same? No\n\n",
        "motu": "Product A is title: motu digital performer 5 digital audio software ( mac only ). manufacturer: motu. price: 795.0. Product B is title: motu digital performer dp5 software music production software. manufacturer: . price: 319.95. Are Product A and Product B the same? No\n\nProduct A is title: motu digital performer 5 digital audio software competitive upgrade ( mac only ). manufacturer: motu. price: 395.0. Product B is title: apple shake 4.1 digital compositing software for mac os x effects software. manufacturer: . price: 498.95. Are Product A and Product B the same? No\n\nProduct A is title: motu digital performer 5 digital audio software competitive upgrade ( mac only ). manufacturer: motu. price: 395.0. Product B is title: steinberg wavelab studio 6 audio editing software competitive crossgrade music production software. manufacturer: . price: 199.95. Are Product A and Product B the same? No\n\nProduct A is title: motu digital performer 5 digital audio software competitive upgrade ( mac only ). manufacturer: motu. price: 395.0. Product B is title: motu digital performer dp5 software music production software. manufacturer: . price: 319.95. Are Product A and Product B the same? Yes\n\nProduct A is title: motu digital performer 5 digital audio software competitive upgrade ( mac only ). manufacturer: motu. price: 395.0. Product B is title: motu digital performer dp5 software music production software. manufacturer: . price: 319.95. Are Product A and Product B the same? Yes\n\nProduct A is title: motu digital performer 5 digital audio software ( mac only ). manufacturer: motu. price: 795.0. Product B is title: migo software digital vault 2005. manufacturer: . price: 24.58. Are Product A and Product B the same? No\n\nProduct A is title: motu digital performer 5 digital audio software competitive upgrade ( mac only ). manufacturer: motu. price: 395.0. Product B is title: cakewalk sonar 5 producer edition competative upgrade music production software. manufacturer: . price: 299.0. Are Product A and Product B the same? No\n\n",
        "neso usa": "Product A is title: nero 7 ultra edition enhanced pc & home entrmnt soln. manufacturer: neso usa. price: 99.99. Product B is title: nero nero 7 ultra edition enhanced cd/dvd burning software for windows authoring software. manufacturer: . price: 79.99. Are Product A and Product B the same? Yes\n\nProduct A is title: nero 7 ultra edition enhanced pc & home entrmnt soln. manufacturer: neso usa. price: 99.99. Product B is title: ahead software 70115 nero 7 ultra edition enhanced mini pc & home entrmnt soln. manufacturer: . price: 83.97. Are Product A and Product B the same? No\n\nProduct A is title: nero 7 ultra edition enhanced pc & home entrmnt soln. manufacturer: neso usa. price: 99.99. Product B is title: 70009 nero ultra edition enhanced v. 7 complete package 1 user cd win. manufacturer: . price: 87.76. Are Product A and Product B the same? No\n\n",
        "watchguard technologies inc": "Product A is title: watchguard 1yr subs firebox x5 webblocker wg017196 ). manufacturer: watchguard technologies inc. price: 159.9. Product B is title: wg017449 watchguard firebox x5500e utm software suite subscription license 1 year +. manufacturer: . price: 3715.48. Are Product A and Product B the same? No\n\nProduct A is title: upg serverlock for solaris gold. manufacturer: watchguard technologies inc. price: . Product B is title: wg7569 watchguard livesecurity service gold watchguard serverlock for solaris technic. manufacturer: . price: 2279.02. Are Product A and Product B the same? No\n\nProduct A is title: watchguard 1yr subs firebox x5 webblocker wg017196 ). manufacturer: watchguard technologies inc. price: 159.9. Product B is title: wg017285 watchguard spamblocker subscription license renewal 1 year 1 appliance. manufacturer: . price: 1335.25. Are Product A and Product B the same? No\n\nProduct A is title: vpn manager five to ten firebox. manufacturer: watchguard technologies inc. price: 1995.0. Product B is title: vpn manager 10 fireboxes watchguard wg4010. manufacturer: . price: 831.36. Are Product A and Product B the same? No\n\n",
        "selectsoft publishing": "Product A is title: superstart ! science arcade. manufacturer: selectsoft publishing. price: 24.95. Product B is title: superstart ! fun with reading & writing !. manufacturer: . price: 8.49. Are Product A and Product B the same? No\n\nProduct A is title: quickstart pc cleaner xp. manufacturer: selectsoft publishing. price: 19.95. Product B is title: quickstart photo montage studio. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: quickstudy u.s. government. manufacturer: selectsoft publishing. price: 19.95. Product B is title: quickstudy us government. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: wacky bird hunter. manufacturer: selectsoft publishing. price: 19.95. Product B is title: wacky farm. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: pressure pop !. manufacturer: selectsoft publishing. price: 19.95. Product B is title: pressure pop !. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\n",
        "learning company": "Product A is title: super solvers mission t.h.i.n.k 25 user site license. manufacturer: learning company. price: 399.95. Product B is title: super solvers mission think 25 user site license. manufacturer: . price: 79.95. Are Product A and Product B the same? No\n\nProduct A is title: super solvers spellbound ! ! school edition for mac. manufacturer: learning company. price: 49.95. Product B is title: super solvers spellbound ! ! school edition for mac. manufacturer: . price: 17.95. Are Product A and Product B the same? No\n\nProduct A is title: bodyworks 6.0 25-user building site license. manufacturer: learning company. price: 399.95. Product B is title: bodyworks 6.0 25-user building site license. manufacturer: . price: 89.95. Are Product A and Product B the same? No\n\n",
        "m-and-r-technologies": "Product A is title: cross stitch design studio ( jewel case ). manufacturer: m-and-r-technologies. price: 9.99. Product B is title: cross stitch design studio ( jewel case ) ( 8006 ). manufacturer: m-and-r-technologies. price: 9.59. Are Product A and Product B the same? Yes\n\n",
        "emedia music": "Product A is title: emedia guitar basics. manufacturer: emedia music. price: . Product B is title: emedia music corp emedia blues guitar legends. manufacturer: . price: 24.81. Are Product A and Product B the same? No\n\nProduct A is title: emedia guitar basics. manufacturer: emedia music. price: . Product B is title: emedia music corp my guitar. manufacturer: . price: 24.81. Are Product A and Product B the same? No\n\nProduct A is title: emedia guitar basics. manufacturer: emedia music. price: . Product B is title: emedia essential bass guitar dvd. manufacturer: . price: 16.95. Are Product A and Product B the same? No\n\n",
        "sony-computer-entertainment": "Product A is title: sony psp media manager. manufacturer: sony-computer-entertainment. price: 24.99. Product B is title: personal imagemanager. manufacturer: smith-micro-software. price: 22.9. Are Product A and Product B the same? No\n\n",
        "d-link systems inc. .": "Product A is title: vpn client software 5 users. manufacturer: d-link systems inc. .. price: . Product B is title: rfipsc2-1 ipsec vpn client license 1 user. manufacturer: . price: 57.2. Are Product A and Product B the same? No\n\nProduct A is title: vpn client software 5 users. manufacturer: d-link systems inc. .. price: . Product B is title: sr2116010 safenet softremote vpn client license 5 users win. manufacturer: . price: 184.57. Are Product A and Product B the same? No\n\nProduct A is title: vpn client software 5 users. manufacturer: d-link systems inc. .. price: . Product B is title: d-link vpn client software ( 1-user license ) ds-601 ds-601. manufacturer: . price: 42.99. Are Product A and Product B the same? No\n\n",
        "destineer": "Product A is title: age of empires iii : the warchiefs expansion pack ( mac ). manufacturer: destineer. price: 34.95. Product B is title: the sims : hot date expansion pack. manufacturer: . price: 19.87. Are Product A and Product B the same? No\n\nProduct A is title: starship troopers/first to fight bundle. manufacturer: destineer. price: 19.95. Product B is title: destineer publishing coporation starship troopers/first to fight bundle. manufacturer: . price: 15.08. Are Product A and Product B the same? Yes\n\nProduct A is title: age of empires iii : the warchiefs expansion pack ( mac ). manufacturer: destineer. price: 34.95. Product B is title: age of empires iii : warchiefs expansion pack. manufacturer: . price: 34.95. Are Product A and Product B the same? Yes\n\nProduct A is title: age of empires iii : the warchiefs expansion pack ( mac ). manufacturer: destineer. price: 34.95. Product B is title: destineer inc age of empires iii : the warchiefs expansion pack. manufacturer: . price: 26.14. Are Product A and Product B the same? Yes\n\nProduct A is title: age of empires iii. manufacturer: destineer. price: 49.99. Product B is title: age of empires iii : warchiefs expansion pack. manufacturer: . price: 34.95. Are Product A and Product B the same? No\n\n",
        "smith-micro-software": "Product A is title: stuffit deluxe v 10.0 ( mac ). manufacturer: smith-micro-software. price: 79.99. Product B is title: allume stuffit deluxe 10.0 os x 10.4 or higher. manufacturer: smith-micro-software. price: 69.99. Are Product A and Product B the same? Yes\n\nProduct A is title: stuffit deluxe v 10.0 ( mac ). manufacturer: smith-micro-software. price: 79.99. Product B is title: allume systems inc stuffit deluxe 11. manufacturer: . price: 42.74. Are Product A and Product B the same? No\n\nProduct A is title: allume boost xp for windows. manufacturer: smith-micro-software. price: 29.99. Product B is title: abacus scrapbook for windows. manufacturer: . price: 26.14. Are Product A and Product B the same? No\n\nProduct A is title: internet cleanup 4.0 mac os 10.4 or higher. manufacturer: smith-micro-software. price: 29.99. Product B is title: allume stuffit deluxe 10.0 os x 10.4 or higher. manufacturer: smith-micro-software. price: 69.99. Are Product A and Product B the same? No\n\nProduct A is title: aquazone bass edition. manufacturer: smith-micro-software. price: 19.99. Product B is title: allume aquazone bass edition. manufacturer: smith-micro-software. price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: allume boost xp for windows. manufacturer: smith-micro-software. price: 29.99. Product B is title: act ! 2004 for windows. manufacturer: . price: 249.95. Are Product A and Product B the same? No\n\nProduct A is title: allume boost xp for windows. manufacturer: smith-micro-software. price: 29.99. Product B is title: powerhouse migo for ipod for windows. manufacturer: . price: 71.99. Are Product A and Product B the same? No\n\n",
        "serious magic": "Product A is title: serious magic ultra2 master sets library 2 windows ). manufacturer: serious magic. price: 276.5. Product B is title: serious magic visual communicator 2 web presentation software ( windows ) presentation software. manufacturer: . price: 189.95. Are Product A and Product B the same? No\n\nProduct A is title: serious magic ultra2 master sets library 2 windows ). manufacturer: serious magic. price: 276.5. Product B is title: serious magic ultra master sets library 1 effects software. manufacturer: . price: 339.95. Are Product A and Product B the same? Yes\n\n",
        "sony": "Product A is title: sony acid pro 5. manufacturer: sony. price: 399.95. Product B is title: sony media software acid pro 6 software upgrade from acid pro 5 music production software. manufacturer: sony-pictures-digital-entertainment. price: 129.95. Are Product A and Product B the same? No\n\nProduct A is title: sony acid pro 5. manufacturer: sony. price: 399.95. Product B is title: zone alarm pro 5. manufacturer: . price: 12.9. Are Product A and Product B the same? No\n\nProduct A is title: sony acid pro 5. manufacturer: sony. price: 399.95. Product B is title: sony acid music studio ( pc ). manufacturer: . price: 69.99. Are Product A and Product B the same? No\n\n",
        "infogrames": "Product A is title: tonka search & rescue 2. manufacturer: infogrames. price: . Product B is title: abacus civil air patrol search & rescue. manufacturer: . price: 26.14. Are Product A and Product B the same? No\n\n",
        "iris inc.": "Product A is title: readiris pro 11 corporate edition. manufacturer: iris inc.. price: 399.99. Product B is title: midisoft corporation guitar tool worship edition. manufacturer: . price: 43.36. Are Product A and Product B the same? No\n\nProduct A is title: readiris pro 11 corporate edition. manufacturer: iris inc.. price: 399.99. Product B is title: iris inc readiris pro 11 corporate edition. manufacturer: . price: 430.58. Are Product A and Product B the same? No\n\nProduct A is title: readiris pro 11 corporate edition. manufacturer: iris inc.. price: 399.99. Product B is title: readiris pro 11 corporate edition for pc. manufacturer: . price: 349.99. Are Product A and Product B the same? Yes\n\n",
        "cisco systems ( ciscopro )": "Product A is title: cisco security desktop agent csa-b250-dtop-k9 ). manufacturer: cisco systems ( ciscopro ). price: . Product B is title: cisco csa-b100-srvr-k9 cisco security server agent win + sol 100 agentbundle cisco security server agent license 100 agents win solaris 0746320832586. manufacturer: . price: 55288.67. Are Product A and Product B the same? No\n\n",
        "whiptail": "",
        "microspot ltd": "Product A is title: microspot x-rip ( mac ). manufacturer: microspot ltd. price: 349.99. Product B is title: microsoft excel 2007 ( pc ). manufacturer: . price: 229.95. Are Product A and Product B the same? No\n\nProduct A is title: microspot x-rip ( mac ). manufacturer: microspot ltd. price: 349.99. Product B is title: microspot x-rip large format print driver for os x. manufacturer: . price: 335.99. Are Product A and Product B the same? Yes\n\nProduct A is title: microspot x-rip ( mac ). manufacturer: microspot ltd. price: 349.99. Product B is title: microsoft onenote 2007 ( pc ). manufacturer: . price: 99.95. Are Product A and Product B the same? No\n\nProduct A is title: microspot x-rip ( mac ). manufacturer: microspot ltd. price: 349.99. Product B is title: microspot macdraft pe personal edition. manufacturer: . price: 85.99. Are Product A and Product B the same? No\n\nProduct A is title: microspot x-rip ( mac ). manufacturer: microspot ltd. price: 349.99. Product B is title: microspot interiors ( ints3 .6 sb ). manufacturer: . price: 95.95. Are Product A and Product B the same? No\n\n",
        "sonic-systems-inc .": "Product A is title: upg sgms 1000 incremental node. manufacturer: sonic-systems-inc .. price: . Product B is title: sonicwall gms 1000 upgrade. manufacturer: . price: 63074.12. Are Product A and Product B the same? Yes\n\n",
        "simply-put-software": "Product A is title: simply put software data eliminator. manufacturer: simply-put-software. price: . Product B is title: spyware x-terminator. manufacturer: . price: 24.9. Are Product A and Product B the same? No\n\nProduct A is title: simply put software data eliminator. manufacturer: simply-put-software. price: . Product B is title: simply put software llc mozy 3.0. manufacturer: simply-put-software-llc. price: 43.3. Are Product A and Product B the same? No\n\nProduct A is title: simply put software got ta sing windows/macintosh ). manufacturer: simply-put-software. price: . Product B is title: webroot software window washer. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: simply put software got ta sing windows/macintosh ). manufacturer: simply-put-software. price: . Product B is title: phase one capture-one pro dslr raw image editing software for macintosh & windows .. manufacturer: . price: 399.95. Are Product A and Product B the same? No\n\nProduct A is title: simply put software got ta sing windows/macintosh ). manufacturer: simply-put-software. price: . Product B is title: simply put software llc gs905-s got ta sing ( win 95 98 me nt 2000 xp/mac 10.0 or higher ). manufacturer: simply-put-software. price: 34.97. Are Product A and Product B the same? Yes\n\n",
        "sega-of-america-inc .": "Product A is title: sonic mega collection plus. manufacturer: sega-of-america-inc .. price: 19.99. Product B is title: sega of america inc sonic mega collection. manufacturer: sega-of-america-inc .. price: 17.55. Are Product A and Product B the same? Yes\n\nProduct A is title: rome total war gold edition. manufacturer: sega-of-america-inc .. price: 19.99. Product B is title: destineer rise of nations gold edition. manufacturer: . price: 45.99. Are Product A and Product B the same? No\n\nProduct A is title: rome total war gold edition. manufacturer: sega-of-america-inc .. price: 19.99. Product B is title: ghost recon gold edition ( dvd-rom ). manufacturer: . price: 20.95. Are Product A and Product B the same? No\n\nProduct A is title: rome total war gold edition. manufacturer: sega-of-america-inc .. price: 19.99. Product B is title: xbox 360 : call of action 3 gold edition. manufacturer: . price: 59.99. Are Product A and Product B the same? No\n\n",
        "brighter-minds-media-inc .": "Product A is title: land before time : kindergarten. manufacturer: brighter-minds-media-inc .. price: 9.99. Product B is title: jumpstart kindergarten ( jc ). manufacturer: . price: 9.9. Are Product A and Product B the same? No\n\nProduct A is title: land before time : preschool. manufacturer: brighter-minds-media-inc .. price: 9.99. Product B is title: the land before time : kindergarten adventure. manufacturer: . price: 12.9. Are Product A and Product B the same? No\n\nProduct A is title: land before time : kindergarten. manufacturer: brighter-minds-media-inc .. price: 9.99. Product B is title: brighter minds land before time : kindergarten. manufacturer: brighter-minds-media-inc .. price: 9.99. Are Product A and Product B the same? Yes\n\nProduct A is title: caillou ready for school ( pc & mac ). manufacturer: brighter-minds-media-inc .. price: 19.99. Product B is title: caillou ready for school by brighter minds. manufacturer: brighter-minds-media-inc .. price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: land before time : kindergarten. manufacturer: brighter-minds-media-inc .. price: 9.99. Product B is title: the land before time : kindergarten adventure. manufacturer: . price: 12.9. Are Product A and Product B the same? Yes\n\nProduct A is title: caillou ready for school ( pc & mac ). manufacturer: brighter-minds-media-inc .. price: 19.99. Product B is title: weekly reader mastering elementary & middle school math ( pc/mac ) fogware. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: land before time : preschool. manufacturer: brighter-minds-media-inc .. price: 9.99. Product B is title: brighter minds land before time : kindergarten. manufacturer: brighter-minds-media-inc .. price: 9.99. Are Product A and Product B the same? No\n\nProduct A is title: caillou magical adventures. manufacturer: brighter-minds-media-inc .. price: 19.99. Product B is title: caillou magical adventures by brighter minds. manufacturer: brighter-minds-media-inc .. price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: land before time : preschool. manufacturer: brighter-minds-media-inc .. price: 9.99. Product B is title: reader rabbit preschool. manufacturer: the-learning-company. price: 17.9. Are Product A and Product B the same? No\n\nProduct A is title: land before time : preschool. manufacturer: brighter-minds-media-inc .. price: 9.99. Product B is title: the land before time : preschool adventure. manufacturer: brighter-minds-media-inc .. price: 12.9. Are Product A and Product B the same? Yes\n\n",
        "jungle software": "Product A is title: gorilla 4 standard. manufacturer: jungle software. price: 299.0. Product B is title: adobe cs3 web standard. manufacturer: . price: 1035.99. Are Product A and Product B the same? No\n\nProduct A is title: gorilla 4 standard. manufacturer: jungle software. price: 299.0. Product B is title: microsoft office 2004 standard. manufacturer: . price: 369.99. Are Product A and Product B the same? No\n\n",
        "imsi": "Product A is title: turbocad mac 3d for mac. manufacturer: imsi. price: 249.99. Product B is title: sims 2 pets for mac. manufacturer: . price: 34.99. Are Product A and Product B the same? No\n\nProduct A is title: imsi turbocad designer v. 11. manufacturer: imsi. price: 39.99. Product B is title: imsi turbocad designer 11. manufacturer: . price: 12.9. Are Product A and Product B the same? Yes\n\nProduct A is title: turbocad mac 3d for mac. manufacturer: imsi. price: 249.99. Product B is title: sound studio 3 for mac. manufacturer: . price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: turbocad mac 3d for mac. manufacturer: imsi. price: 249.99. Product B is title: imsi design llc turbocad mac 3d. manufacturer: . price: 204.67. Are Product A and Product B the same? Yes\n\nProduct A is title: imsi turbocad designer v. 11. manufacturer: imsi. price: 39.99. Product B is title: turbocad ( r ) designer 2d/3d version 9.1. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\n",
        "humongous entertainment": "Product A is title: backyard sports backyard basketball and backyard football. manufacturer: humongous entertainment. price: 19.99. Product B is title: backyard basketball & backyard hockey double pack. manufacturer: . price: 9.95. Are Product A and Product B the same? No\n\nProduct A is title: backyard sports backyard basketball and backyard football. manufacturer: humongous entertainment. price: 19.99. Product B is title: brighter minds backyard sports. manufacturer: . price: 19.99. Are Product A and Product B the same? Yes\n\nProduct A is title: blues clues preschool. manufacturer: humongous entertainment. price: . Product B is title: reader rabbit preschool. manufacturer: the-learning-company. price: 17.9. Are Product A and Product B the same? No\n\nProduct A is title: blues clues preschool. manufacturer: humongous entertainment. price: . Product B is title: brain play preschool 1st grade. manufacturer: . price: 25.99. Are Product A and Product B the same? No\n\nProduct A is title: blues clues preschool. manufacturer: humongous entertainment. price: . Product B is title: brain play preschool-1st grade. manufacturer: . price: 17.9. Are Product A and Product B the same? No\n\nProduct A is title: backyard sports backyard basketball and backyard football. manufacturer: humongous entertainment. price: 19.99. Product B is title: backyard sports : baseball 2007 for pc. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\n",
        "disney interactive": "Product A is title: millionaire sports edition ( jewel case ). manufacturer: disney interactive. price: 9.99. Product B is title: who wants to be a millionaire sports edition. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: millionaire sports edition ( jewel case ). manufacturer: disney interactive. price: 9.99. Product B is title: poetry fiction and drama ( win/mac ) ( jewel case ) ( 10172 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: millionaire sports edition ( jewel case ). manufacturer: disney interactive. price: 9.99. Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: disney 2nd grade active leveling advantage w buzz ( jewel case ) ages 6-8. manufacturer: disney interactive. price: 9.99. Product B is title: disney 's buzz lightyear 2nd grade. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: millionaire sports edition ( jewel case ). manufacturer: disney interactive. price: 9.99. Product B is title: merriam-webster 's dictionary ( win/mac ) ( jewel case ) ( 10462 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\n",
        "nuance academic": "Product A is title: acad upgrade dragon naturallyspeaking pro solution 9.0 ( a289a-fd7-9 .0 ). manufacturer: nuance academic. price: 399.54. Product B is title: nuance dragon naturallyspeaking v. 9.0 professional upgrade. manufacturer: . price: 212.62. Are Product A and Product B the same? No\n\n",
        "digidesign": "Product A is title: digidesign music production toolkit ( pro tools le ). manufacturer: digidesign. price: 495.0. Product B is title: make finale 2007 software music production software. manufacturer: . price: 429.95. Are Product A and Product B the same? No\n\nProduct A is title: digidesign music production toolkit ( pro tools le ). manufacturer: digidesign. price: 495.0. Product B is title: make songwriter software music production software. manufacturer: . price: 37.95. Are Product A and Product B the same? No\n\nProduct A is title: digidesign music production toolkit ( pro tools le ). manufacturer: digidesign. price: 495.0. Product B is title: apple/emagic logic pro 7.2 software music production software. manufacturer: . price: 999.0. Are Product A and Product B the same? No\n\nProduct A is title: digidesign music production toolkit ( pro tools le ). manufacturer: digidesign. price: 495.0. Product B is title: digidesign music production toolkit music production software. manufacturer: . price: 495.0. Are Product A and Product B the same? Yes\n\nProduct A is title: digidesign music production toolkit ( pro tools le ). manufacturer: digidesign. price: 495.0. Product B is title: sony media software acid pro 5 music production software. manufacturer: . price: 199.95. Are Product A and Product B the same? No\n\nProduct A is title: digidesign music production toolkit ( pro tools le ). manufacturer: digidesign. price: 495.0. Product B is title: apple/emagic logic pro 7.2 software music production software. manufacturer: . price: 999.0. Are Product A and Product B the same? No\n\n",
        "software cinema": "Product A is title: adobe photoshop cs2 fundamental techniques by julieanne kost. manufacturer: software cinema. price: . Product B is title: adobe photoshop cs3 extended for mac upgrade. manufacturer: . price: 329.99. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop cs2 advanced techniques by julieanne kost. manufacturer: software cinema. price: . Product B is title: adobe photoshop cs3 ( v10 .0 ) mac adobe 13102488. manufacturer: . price: 537.65. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop cs2 advanced techniques by julieanne kost. manufacturer: software cinema. price: . Product B is title: adobe photoshop cs3 ( v10 .0 ) mac adobe 13102488. manufacturer: . price: 537.65. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop cs2 advanced techniques by julieanne kost. manufacturer: software cinema. price: . Product B is title: adobe photoshop lightroom. manufacturer: . price: 264.99. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop cs2 advanced techniques by julieanne kost. manufacturer: software cinema. price: . Product B is title: adobe photoshop cs3 extended for mac academic. manufacturer: . price: 295.99. Are Product A and Product B the same? No\n\nProduct A is title: adobe photoshop cs2 fundamental techniques by julieanne kost. manufacturer: software cinema. price: . Product B is title: software cinema dvd-rom : dvdrom : photoshop cs2 fundamental techniques ( training ) photoshop software. manufacturer: . price: 159.95. Are Product A and Product B the same? Yes\n\n",
        "solidworks corporation": "",
        "adobe software": "Product A is title: adobe software acrobat pro upgrd pro-pro mac ( adbcd01798mc ). manufacturer: adobe software. price: 273.0. Product B is title: adobe acrobat v8 .0 professional for mac upgrade. manufacturer: . price: 145.99. Are Product A and Product B the same? Yes\n\nProduct A is title: adobe software acrobat pro upgrd pro-pro mac ( adbcd01798mc ). manufacturer: adobe software. price: 273.0. Product B is title: adobe acrobat v8 .0 professional for mac. manufacturer: . price: 405.99. Are Product A and Product B the same? No\n\nProduct A is title: adobe software acrobat pro upgrd pro-pro mac ( adbcd01798mc ). manufacturer: adobe software. price: 273.0. Product B is title: adobe acrobat 7.0 standard academic mac. manufacturer: . price: 98.99. Are Product A and Product B the same? No\n\nProduct A is title: adobe software acrobat pro upgrd pro-pro mac ( adbcd01798mc ). manufacturer: adobe software. price: 273.0. Product B is title: adobe acrobat v8 .0 professional for mac. manufacturer: . price: 405.99. Are Product A and Product B the same? No\n\nProduct A is title: adobe software acrobat pro upgrd pro-pro mac ( adbcd01798mc ). manufacturer: adobe software. price: 273.0. Product B is title: adobe acrobat ( v8 .0 ) pro win up from std adobe 22020452. manufacturer: . price: 151.27. Are Product A and Product B the same? No\n\nProduct A is title: adobe software acrobat pro upgrd pro-pro mac ( adbcd01798mc ). manufacturer: adobe software. price: 273.0. Product B is title: adobe acrobat v8 .0 professional for mac upsell from std-pro. manufacturer: . price: 145.99. Are Product A and Product B the same? No\n\n",
        "tri synergy": "Product A is title: lavasoft ad-aware plus edition. manufacturer: tri synergy. price: 29.99. Product B is title: lavasoft ad-aware plus edition. manufacturer: . price: 24.9. Are Product A and Product B the same? Yes\n\nProduct A is title: big oil : build an oil empire. manufacturer: tri synergy. price: 19.99. Product B is title: big oil : build an oil empire. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\nProduct A is title: oxford spanish dictionary. manufacturer: tri synergy. price: 19.99. Product B is title: palmspring software 523 oxford american desk dictionary. manufacturer: palmspring software. price: 24.2. Are Product A and Product B the same? No\n\n",
        "git corp": "Product A is title: captain america the complete comic collection win/mac. manufacturer: git corp. price: 49.95. Product B is title: sega of america inc sonic mega collection. manufacturer: sega-of-america-inc .. price: 17.55. Are Product A and Product B the same? No\n\nProduct A is title: amazing spider-man complete comic book collection win/mac. manufacturer: git corp. price: 49.99. Product B is title: git comic book library collector 's edition vol . 1. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\nProduct A is title: 40 years of the avengers. manufacturer: git corp. price: 49.99. Product B is title: git 40 years of marvel avengers dvd-rom. manufacturer: . price: 39.99. Are Product A and Product B the same? Yes\n\nProduct A is title: captain america the complete comic collection win/mac. manufacturer: git corp. price: 49.95. Product B is title: 11273 hasbro family game collection 2006 complete package 1 user pc cd win. manufacturer: . price: 14.51. Are Product A and Product B the same? No\n\nProduct A is title: 44 years of the fantastic four. manufacturer: git corp. price: 49.99. Product B is title: git 44 years of marvel fantastic 4 dvd-rom. manufacturer: . price: 42.99. Are Product A and Product B the same? Yes\n\nProduct A is title: captain america the complete comic collection win/mac. manufacturer: git corp. price: 49.95. Product B is title: git amazing spider-man the complete collection dvd-rom. manufacturer: . price: 42.99. Are Product A and Product B the same? No\n\nProduct A is title: amazing spider-man complete comic book collection win/mac. manufacturer: git corp. price: 49.99. Product B is title: git amazing spider-man the complete collection dvd-rom. manufacturer: . price: 42.99. Are Product A and Product B the same? Yes\n\nProduct A is title: captain america the complete comic collection win/mac. manufacturer: git corp. price: 49.95. Product B is title: aspyr the sims complete collection. manufacturer: . price: 47.99. Are Product A and Product B the same? No\n\n",
        "adaptec": "Product A is title: snap enterprise data replicator advanced license for 10000 series. manufacturer: adaptec. price: . Product B is title: 5325302018 snap enterprise data replicator license 1 server linux unix win. manufacturer: . price: 663.22. Are Product A and Product B the same? No\n\nProduct A is title: snap enterprise data replicator advanced license for 10000 series. manufacturer: adaptec. price: . Product B is title: 5325301871 snap appliance snap care snap enterprise data replicator technical support c. manufacturer: . price: 377.06. Are Product A and Product B the same? No\n\n",
        "majesco entertainment": "",
        "business objects": "Product A is title: crystal reports xi professional edition. manufacturer: business objects. price: 495.0. Product B is title: business objects crystal reports xi. manufacturer: . price: 459.99. Are Product A and Product B the same? Yes\n\nProduct A is title: crystal reports xi developer. manufacturer: business objects. price: 595.0. Product B is title: upg crystal reports 11 pro french. manufacturer: . price: 264.99. Are Product A and Product B the same? No\n\nProduct A is title: upgrade crystal reports 11 pro french. manufacturer: business objects. price: 479.78. Product B is title: business objects crystal reports xi. manufacturer: . price: 459.99. Are Product A and Product B the same? No\n\nProduct A is title: crystal reports xi professional edition. manufacturer: business objects. price: 495.0. Product B is title: upg crystal reports 11 pro french. manufacturer: . price: 264.99. Are Product A and Product B the same? No\n\nProduct A is title: crystal reports xi professional edition. manufacturer: business objects. price: 495.0. Product B is title: microsoft ( r ) windows ( r ) xp professional edition with sp2. manufacturer: . price: 199.99. Are Product A and Product B the same? No\n\nProduct A is title: upgrade crystal reports 11 pro french. manufacturer: business objects. price: 479.78. Product B is title: upg crystal reports 11 pro french. manufacturer: . price: 264.99. Are Product A and Product B the same? Yes\n\nProduct A is title: crystal reports 11 reporting essentials trng cbt. manufacturer: business objects. price: 199.99. Product B is title: business objects t-59j-e-56-wx crystal rpt 11 report essentials cbt. manufacturer: . price: 178.49. Are Product A and Product B the same? Yes\n\nProduct A is title: crystal reports xi professional edition. manufacturer: business objects. price: 495.0. Product B is title: zipmagic personal edition. manufacturer: . price: 8.95. Are Product A and Product B the same? No\n\n",
        "dvo enterprises": "Product A is title: cook 'n low fat. manufacturer: dvo enterprises. price: 19.99. Product B is title: dvo enterprises inc. cookn low fat. manufacturer: . price: 17.55. Are Product A and Product B the same? Yes\n\nProduct A is title: dvo cook 'n deluxe 6.0 ultimate recipe organizer. manufacturer: dvo enterprises. price: 39.99. Product B is title: dvo enterprises inc. cookn deluxe 6.0. manufacturer: . price: 34.99. Are Product A and Product B the same? Yes\n\n",
        "macspeech inc.": "Product A is title: ilisten ( mac ). manufacturer: macspeech inc.. price: 99.99. Product B is title: iwork '06 ( mac ) apple. manufacturer: . price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: ilisten ( mac ). manufacturer: macspeech inc.. price: 99.99. Product B is title: apple appleworks 6.2.9 ( mac ). manufacturer: . price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: ilisten ( mac ). manufacturer: macspeech inc.. price: 99.99. Product B is title: . mac ( mac ) apple. manufacturer: . price: 99.99. Are Product A and Product B the same? No\n\n",
        "magix entertainment": "Product A is title: photostory on cd & dvd 5. manufacturer: magix entertainment. price: 19.99. Product B is title: magix entertainment corp. photostory on cd & dvd 5. manufacturer: . price: 17.55. Are Product A and Product B the same? Yes\n\nProduct A is title: web radio recorder. manufacturer: magix entertainment. price: . Product B is title: magix webradio recorder ( pc ) magix entertainment. manufacturer: . price: 24.99. Are Product A and Product B the same? Yes\n\nProduct A is title: xtreme photostory on cd and dvd 6. manufacturer: magix entertainment. price: 29.99. Product B is title: magix entertainment corp. photostory on cd & dvd 5. manufacturer: . price: 17.55. Are Product A and Product B the same? No\n\nProduct A is title: web radio recorder. manufacturer: magix entertainment. price: . Product B is title: mp3 radio recorder for ipod. manufacturer: . price: 22.9. Are Product A and Product B the same? No\n\nProduct A is title: photostory on cd & dvd 5. manufacturer: magix entertainment. price: 19.99. Product B is title: magix entertainment corp. xtreme photostory on cd and dvd 6. manufacturer: . price: 26.14. Are Product A and Product B the same? No\n\n",
        "riverdeep-learning-company": "Product A is title: the printshop business card maker ( jewel case ). manufacturer: riverdeep-learning-company. price: 9.99. Product B is title: total 3d home deluxe ( jewel case ) ( jce-th6 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: the printshop business card maker ( jewel case ). manufacturer: riverdeep-learning-company. price: 9.99. Product B is title: encore software 2137421 printshop business card maker ( win 98 me 2000 xp ). manufacturer: . price: 8.64. Are Product A and Product B the same? Yes\n\nProduct A is title: the printshop business card maker ( jewel case ). manufacturer: riverdeep-learning-company. price: 9.99. Product B is title: the printshop 20. manufacturer: . price: 19.9. Are Product A and Product B the same? No\n\nProduct A is title: creating keepsakes scrapbook designer version 2 platinum. manufacturer: riverdeep-learning-company. price: 39.99. Product B is title: encore software 10612 creating keepsakes scrapbook designer deluxe v. 3 sb cs. manufacturer: . price: 17.97. Are Product A and Product B the same? No\n\n",
        "nuance communications inc. .": "",
        "acclaim": "",
        "delorme": "Product A is title: delorme topo usa mapping software 6.0 west region. manufacturer: delorme. price: 49.95. Product B is title: delorme mapping topo usa 6.0 west region. manufacturer: . price: 43.36. Are Product A and Product B the same? Yes\n\nProduct A is title: street atlas usa 2007 plus dvd. manufacturer: delorme. price: 59.95. Product B is title: delorme mapping street atlas usa 2007 plus dvd xx. manufacturer: . price: 53.37. Are Product A and Product B the same? Yes\n\nProduct A is title: delorme topo usa mapping software 6.0 east region. manufacturer: delorme. price: 49.95. Product B is title: delorme mapping topo usa 6.0 west region. manufacturer: . price: 43.36. Are Product A and Product B the same? No\n\nProduct A is title: delorme topo usa mapping software 6.0 east region. manufacturer: delorme. price: 49.95. Product B is title: delorme mapping street atlas usa 2007 plus dvd xx. manufacturer: . price: 53.37. Are Product A and Product B the same? No\n\nProduct A is title: delorme topo usa mapping software 6.0 east region. manufacturer: delorme. price: 49.95. Product B is title: delorme mapping topo usa 6.0 east region. manufacturer: . price: 43.36. Are Product A and Product B the same? Yes\n\n",
        "selectmedia entertainment": "",
        "datawatch": "Product A is title: datawatch monarch v7 ntwk-starter cd most 8u fpm32c070-a08n ). manufacturer: datawatch. price: . Product B is title: monarch std ( v9 .0 ) network starter 4-user rtl box datawatch fpm32c090-a04n. manufacturer: . price: 1157.21. Are Product A and Product B the same? No\n\n",
        "netmanage": "Product A is title: netmanage rumba web as/400 edition 910523-002 ). manufacturer: netmanage. price: . Product B is title: 910523-002 rumba as/400 edition v. 7.2 upgrade package 1 processor tier p50 c. manufacturer: . price: 20069.43. Are Product A and Product B the same? No\n\nProduct A is title: netmanage upg rumba unix-hp edition level 902453-007-c ). manufacturer: netmanage. price: . Product B is title: 902453-007-b rumba unix-hp edition upgrade license 1 user volume level b 5-49 c. manufacturer: . price: 147.73. Are Product A and Product B the same? Yes\n\nProduct A is title: netmanage rumba unix-hp edition v. 7.3 upgrade license 902453-006 ). manufacturer: netmanage. price: . Product B is title: bhe3dus installanywhere enterprise edition v. 8 product upgrade license + 1 year. manufacturer: . price: 4746.16. Are Product A and Product B the same? No\n\nProduct A is title: netmanage rumba web as/400 edition 910584-001 ). manufacturer: netmanage. price: . Product B is title: 910519-004 rumba as/400 edition v. 7.4 upgrade license 1 processor tier p10 c. manufacturer: . price: 3223.26. Are Product A and Product B the same? No\n\nProduct A is title: netmanage rumba web as/400 edition 910584-001 ). manufacturer: netmanage. price: . Product B is title: 910584-001 rumba as/400 edition v. 7.2 upgrade package 1 processor tier p60 c. manufacturer: . price: 26759.24. Are Product A and Product B the same? No\n\nProduct A is title: netmanage rumba unix-hp edition v. 7.3 upgrade license 902453-006 ). manufacturer: netmanage. price: . Product B is title: bhe3dus installanywhere enterprise edition v. 8 product upgrade license + 1 year. manufacturer: . price: 4746.16. Are Product A and Product B the same? No\n\nProduct A is title: netmanage upg rumba unix-hp edition level 902453-007-c ). manufacturer: netmanage. price: . Product B is title: 902453-007-e rumba unix-hp edition upgrade license 1 user volume level e 1000-1999. manufacturer: . price: 106.48. Are Product A and Product B the same? Yes\n\n",
        "eidos interactive": "Product A is title: startopia. manufacturer: eidos interactive. price: 39.95. Product B is title: startopia. manufacturer: . price: 14.95. Are Product A and Product B the same? No\n\n",
        "atari": "Product A is title: home budget for dummies. manufacturer: atari. price: . Product B is title: home budget for dummies. manufacturer: . price: 10.95. Are Product A and Product B the same? No\n\nProduct A is title: tonka town ( jewel case ). manufacturer: atari. price: 9.99. Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: backyard skateboarding. manufacturer: atari. price: . Product B is title: backyard skateboarding. manufacturer: . price: 7.95. Are Product A and Product B the same? No\n\nProduct A is title: tonka town ( jewel case ). manufacturer: atari. price: 9.99. Product B is title: total 3d home deluxe ( jewel case ) ( jce-th6 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: encyclopedia britannica profiles : dinosaurs ( jewel case ). manufacturer: atari. price: 9.99. Product B is title: avanquest encyclopedia britannica 2006. manufacturer: . price: 29.0. Are Product A and Product B the same? No\n\n",
        "filemaker": "Product A is title: filemaker mobile 8 for palm os and pocket pc. manufacturer: filemaker. price: 69.0. Product B is title: resco explorer 2007 for pocketpc windows mobile software for pocket pc. manufacturer: . price: 29.95. Are Product A and Product B the same? No\n\nProduct A is title: filemaker mobile 8 for palm os and pocket pc. manufacturer: filemaker. price: 69.0. Product B is title: splashid windows mobile software for pocket pc. manufacturer: . price: 29.95. Are Product A and Product B the same? No\n\nProduct A is title: filemaker mobile 8 for palm os and pocket pc. manufacturer: filemaker. price: 69.0. Product B is title: filemaker mobile 8. manufacturer: . price: 65.99. Are Product A and Product B the same? No\n\nProduct A is title: filemaker mobile 8 for palm os and pocket pc. manufacturer: filemaker. price: 69.0. Product B is title: td824ll/a filemaker mobile v. 8 complete package 1 user cd win mac palm os. manufacturer: . price: 66.15. Are Product A and Product B the same? No\n\n",
        "oxford-university-press": "Product A is title: oxford spanish dictionary. manufacturer: oxford-university-press. price: 19.95. Product B is title: palmspring software 523 oxford american desk dictionary. manufacturer: palmspring software. price: 24.2. Are Product A and Product B the same? No\n\nProduct A is title: oxford spanish dictionary. manufacturer: oxford-university-press. price: 19.95. Product B is title: pop up oxford spanish / english dict. manufacturer: oxford-university-press. price: 13.95. Are Product A and Product B the same? Yes\n\n",
        "eagle games": "Product A is title: tournament poker 2005. manufacturer: eagle games. price: 20.99. Product B is title: tournament poker no limit texas holdem. manufacturer: . price: 19.99. Are Product A and Product B the same? No\n\n",
        "nero inc.": "Product A is title: nero 8 ultra edition. manufacturer: nero inc.. price: 99.95. Product B is title: nero inc nero 8 ultra edition. manufacturer: . price: 88.83. Are Product A and Product B the same? Yes\n\nProduct A is title: nero 7 ultra edition enhanced. manufacturer: nero inc.. price: 99.99. Product B is title: 70009 nero ultra edition enhanced v. 7 complete package 1 user cd win. manufacturer: . price: 87.76. Are Product A and Product B the same? No\n\nProduct A is title: nero 7 ultra edition enhanced. manufacturer: nero inc.. price: 99.99. Product B is title: nero nero 7 ultra edition enhanced cd/dvd burning software for windows authoring software. manufacturer: . price: 79.99. Are Product A and Product B the same? No\n\nProduct A is title: nero 8 ultra edition. manufacturer: nero inc.. price: 99.95. Product B is title: i copy dvds 2 ultra edition. manufacturer: . price: 29.9. Are Product A and Product B the same? No\n\nProduct A is title: nero 7 ultra edition enhanced. manufacturer: nero inc.. price: 99.99. Product B is title: ahead software 70115 nero 7 ultra edition enhanced mini pc & home entrmnt soln. manufacturer: . price: 83.97. Are Product A and Product B the same? Yes\n\n",
        "feral interactive": "Product A is title: lego star wars ii : the original trilogy for mac. manufacturer: feral interactive. price: 39.99. Product B is title: xbox : lego star wars. manufacturer: . price: 19.95. Are Product A and Product B the same? No\n\nProduct A is title: chessmaster 9000 ( mac ). manufacturer: feral interactive. price: 39.99. Product B is title: feral interactive limited chessmaster 9000. manufacturer: . price: 35.98. Are Product A and Product B the same? Yes\n\nProduct A is title: chessmaster 9000 ( mac ). manufacturer: feral interactive. price: 39.99. Product B is title: chessmaster 9000 : feral interactive. manufacturer: . price: 38.99. Are Product A and Product B the same? Yes\n\nProduct A is title: black & white platinum pack ( mac ). manufacturer: feral interactive. price: 29.99. Product B is title: feral interactive limited black & white platinum pack. manufacturer: . price: 27.37. Are Product A and Product B the same? Yes\n\nProduct A is title: black & white platinum pack ( mac ). manufacturer: feral interactive. price: 29.99. Product B is title: intervideo windvd 8 platinum ( pc ). manufacturer: . price: 59.99. Are Product A and Product B the same? No\n\nProduct A is title: lego star wars ii : the original trilogy for mac. manufacturer: feral interactive. price: 39.99. Product B is title: aspyr lego star wars ii. manufacturer: . price: 44.95. Are Product A and Product B the same? Yes\n\n",
        "nuance": "",
        "global marketing": "Product A is title: iview mediapro 2.5. manufacturer: global marketing. price: 199.99. Product B is title: iview mediapro 2.6 media management. manufacturer: . price: 155.99. Are Product A and Product B the same? Yes\n\n",
        "masque publishing": "Product A is title: sportsman 's double play. manufacturer: masque publishing. price: 19.95. Product B is title: masque publishing sportsmans double play with american billiards. manufacturer: . price: 17.55. Are Product A and Product B the same? No\n\nProduct A is title: solitaire antics deluxe. manufacturer: masque publishing. price: . Product B is title: masque publishing solitaire antics deluxe. manufacturer: . price: 8.46. Are Product A and Product B the same? Yes\n\nProduct A is title: best of slots 2 ( jewel case ). manufacturer: masque publishing. price: 9.99. Product B is title: cross stitch design studio ( jewel case ) ( 8006 ). manufacturer: m-and-r-technologies. price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: best of slots 2 ( jewel case ). manufacturer: masque publishing. price: 9.99. Product B is title: total 3d home deluxe ( jewel case ) ( jce-th6 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: sportsman 's double play. manufacturer: masque publishing. price: 19.95. Product B is title: masque publishing sportsmans double play. manufacturer: . price: 17.55. Are Product A and Product B the same? Yes\n\n",
        "lenovo": "Product A is title: the computrace lojack for laptops computer tracking and loss control solution by. manufacturer: lenovo. price: . Product B is title: absolute software lfl-m2-36 computrace lojack for laptops 3 year license mac 10.2 or higher. manufacturer: . price: 80.97. Are Product A and Product B the same? No\n\nProduct A is title: the computrace lojack for laptops computer tracking and loss control solution by. manufacturer: lenovo. price: . Product B is title: computrace lojack for laptops 3 yr subscription. manufacturer: . price: 89.99. Are Product A and Product B the same? No\n\nProduct A is title: the computrace lojack for laptops computer tracking and loss control solution by. manufacturer: lenovo. price: . Product B is title: absolute software lfl-n2-36 computrace lojack for laptops 3 year license ( win 2000 xp ). manufacturer: . price: 85.97. Are Product A and Product B the same? No\n\nProduct A is title: the computrace lojack for laptops computer tracking and loss control solution by. manufacturer: lenovo. price: . Product B is title: absolute software lfl-n2-48 computrace lojack for laptops 4 year license. manufacturer: . price: 105.68. Are Product A and Product B the same? No\n\nProduct A is title: the computrace lojack for laptops computer tracking and loss control solution by. manufacturer: lenovo. price: . Product B is title: computrace lojack for laptops 1 yr subscription. manufacturer: . price: 45.99. Are Product A and Product B the same? No\n\n",
        "cisco systems enterprise": "Product A is title: cisco ios enterprise services complete package cd28-esk9 = ). manufacturer: cisco systems enterprise. price: . Product B is title: cd72-ar1k9 = cisco ios enterprise/snasw ipsec 3des complete package cd. manufacturer: . price: 10576.76. Are Product A and Product B the same? No\n\nProduct A is title: cisco 2600xm ser ios ip/fw/ids cd26xm-ch = ). manufacturer: cisco systems enterprise. price: . Product B is title: cd26xm-ch = cisco ios ip/fw/ids complete package cd. manufacturer: . price: 360.82. Are Product A and Product B the same? No\n\nProduct A is title: cisco ios enterprise services complete package cd28-esk9 = ). manufacturer: cisco systems enterprise. price: . Product B is title: 22774 vsi-fax v. 6 complete package 1 server cd linux. manufacturer: . price: 1573.45. Are Product A and Product B the same? No\n\nProduct A is title: cisco syst . minor release : requires existing cww-6 .1 - win-mr ). manufacturer: cisco systems enterprise. price: . Product B is title: cisco ciscoworks v. 6.1 for windows upgrade cww-6 .1 - win-mr. manufacturer: . price: 225.0. Are Product A and Product B the same? No\n\nProduct A is title: cisco ios enterprise services complete package cd28-esk9 = ). manufacturer: cisco systems enterprise. price: . Product B is title: cd28-esk9 = cisco ios enterprise services complete package cd. manufacturer: . price: 1001.38. Are Product A and Product B the same? No\n\n",
        "abacus": "Product A is title: aircraft collector 's second edition. manufacturer: abacus. price: 19.95. Product B is title: neil diamond the neil diamond collection. manufacturer: . price: 22.13. Are Product A and Product B the same? No\n\nProduct A is title: train sim modeler design studio. manufacturer: abacus. price: 39.99. Product B is title: abacus train sim modeler. manufacturer: . price: 29.84. Are Product A and Product B the same? Yes\n\nProduct A is title: national parks : enhanced scenery add-on for microsoft flight simulator 2002 & 2000. manufacturer: abacus. price: 19.99. Product B is title: abacus national parks enhanced scenery. manufacturer: . price: 17.55. Are Product A and Product B the same? Yes\n\nProduct A is title: usa/world extreme landscapes bundle. manufacturer: abacus. price: 39.95. Product B is title: abacus software s546 usa extreme landscapes for fs2004 and fs2002 3cd set. manufacturer: abacus software. price: 28.99. Are Product A and Product B the same? No\n\nProduct A is title: usa/world extreme landscapes bundle. manufacturer: abacus. price: 39.95. Product B is title: abacus software s546 usa extreme landscapes for fs2004 and fs2002 3cd set. manufacturer: abacus software. price: 28.99. Are Product A and Product B the same? No\n\nProduct A is title: aircraft collector 's second edition. manufacturer: abacus. price: 19.95. Product B is title: abacus aircraft collectors second edition. manufacturer: . price: 17.55. Are Product A and Product B the same? Yes\n\nProduct A is title: usa/world extreme landscapes bundle. manufacturer: abacus. price: 39.95. Product B is title: abacus usa/world extreme landscapes bundle. manufacturer: . price: 38.43. Are Product A and Product B the same? Yes\n\nProduct A is title: usa/world extreme landscapes bundle. manufacturer: abacus. price: 39.95. Product B is title: abacus software s546 usa extreme landscapes for fs2004 and fs2002 3cd set. manufacturer: abacus software. price: 28.99. Are Product A and Product B the same? No\n\nProduct A is title: aircraft collector 's second edition. manufacturer: abacus. price: 19.95. Product B is title: family tree maker 16 collectors edition. manufacturer: . price: 99.99. Are Product A and Product B the same? No\n\nProduct A is title: ultra sudoku/ultra kakuro. manufacturer: abacus. price: 29.95. Product B is title: abacus ultra sudoku/ultra kakuro. manufacturer: . price: 17.55. Are Product A and Product B the same? Yes\n\n",
        "imaginova": "Product A is title: starry night galaxy explorer ( pc & mac ). manufacturer: imaginova. price: 9.99. Product B is title: orion/imaginova starry night galaxy explorer. manufacturer: . price: 8.8. Are Product A and Product B the same? Yes\n\nProduct A is title: starry night galaxy explorer ( pc & mac ). manufacturer: imaginova. price: 9.99. Product B is title: starry night galaxy explorer ( pc & mac ) ( snge-5-c1u ). manufacturer: . price: 9.59. Are Product A and Product B the same? Yes\n\n",
        "freeverse software": "Product A is title: burning monkey solitaire 2005 edition. manufacturer: freeverse software. price: 24.99. Product B is title: freeverse burning monkey solitaire 3d. manufacturer: . price: 24.99. Are Product A and Product B the same? Yes\n\nProduct A is title: burning monkey solitaire 2005 edition. manufacturer: freeverse software. price: 24.99. Product B is title: freeverse burning monkey mahjong 3d. manufacturer: . price: 24.99. Are Product A and Product B the same? No\n\nProduct A is title: 3d bridge deluxe ( mac ). manufacturer: freeverse software. price: 24.99. Product B is title: freeverse software 4012 3d bridge deluxe ( retail box ) for mac. manufacturer: freeverse software. price: 18.99. Are Product A and Product B the same? No\n\nProduct A is title: 3d bridge deluxe ( mac ). manufacturer: freeverse software. price: 24.99. Product B is title: freeverse 3d bridge deluxe. manufacturer: . price: 24.99. Are Product A and Product B the same? Yes\n\nProduct A is title: sound studio 3 : record edit add effects to audio ( mac ). manufacturer: freeverse software. price: 79.99. Product B is title: sound studio 3 for mac. manufacturer: . price: 79.99. Are Product A and Product B the same? Yes\n\nProduct A is title: burning monkey solitaire 4.0. manufacturer: freeverse software. price: 24.95. Product B is title: freeverse software 7001 burning monkey solitaire 4.0 mac 10.3.9 or later ub. manufacturer: . price: 22.97. Are Product A and Product B the same? No\n\nProduct A is title: burning monkey solitaire 4.0. manufacturer: freeverse software. price: 24.95. Product B is title: freeverse burning monkey solitaire 4.0. manufacturer: . price: 24.95. Are Product A and Product B the same? Yes\n\nProduct A is title: toysight gold for isight : includes 15 games ( mac ). manufacturer: freeverse software. price: 29.99. Product B is title: freeverse software 5014 toysight gold ( mac 10.2 or higher ). manufacturer: . price: 22.97. Are Product A and Product B the same? No\n\nProduct A is title: wingnuts 2 ( mac ). manufacturer: freeverse software. price: 29.99. Product B is title: wingnuts 2 : raina 's revenge. manufacturer: . price: 28.99. Are Product A and Product B the same? No\n\n",
        "kutoka": "Product A is title: mia 's math adventure : just in time. manufacturer: kutoka. price: 19.99. Product B is title: the land before time : kindergarten adventure. manufacturer: . price: 12.9. Are Product A and Product B the same? No\n\nProduct A is title: mia 's math adventure : just in time. manufacturer: kutoka. price: 19.99. Product B is title: kutoka interactive 61208 mias math adventure ( just in time ) ( win 95 98 me 2000 xp/mac 8.6-9.x ( classic ) x v10 .1 or higher ). manufacturer: . price: 18.97. Are Product A and Product B the same? Yes\n\nProduct A is title: didi & ditto : kindergarten win/mac. manufacturer: kutoka. price: 19.99. Product B is title: jumpstart kindergarten ( jc ). manufacturer: . price: 9.9. Are Product A and Product B the same? No\n\nProduct A is title: didi & ditto : kindergarten win/mac. manufacturer: kutoka. price: 19.99. Product B is title: the land before time : kindergarten adventure. manufacturer: . price: 12.9. Are Product A and Product B the same? No\n\n",
        "x-oom": "Product A is title: x-oom movie clone 3 gold. manufacturer: x-oom. price: . Product B is title: x-oom internet movies 2. manufacturer: . price: 22.9. Are Product A and Product B the same? No\n\nProduct A is title: x-oom internet movies 2. manufacturer: x-oom. price: . Product B is title: feral interactive the movies. manufacturer: . price: 47.99. Are Product A and Product B the same? No\n\nProduct A is title: x-oom mp3 radio recorder windows ). manufacturer: x-oom. price: . Product B is title: mp3 radio recorder for ipod. manufacturer: . price: 22.9. Are Product A and Product B the same? Yes\n\nProduct A is title: x-oom movie clone 3 gold. manufacturer: x-oom. price: . Product B is title: xoom movie clone 3 gold. manufacturer: . price: 22.9. Are Product A and Product B the same? Yes\n\n",
        "electronic arts": "Product A is title: battlefield 1942 : deluxe edition. manufacturer: electronic arts. price: 29.99. Product B is title: mastercook deluxe low carb edition. manufacturer: . price: 17.9. Are Product A and Product B the same? No\n\nProduct A is title: battlefield 1942 : deluxe edition. manufacturer: electronic arts. price: 29.99. Product B is title: allume aquazone seven seas deluxe edition. manufacturer: . price: 24.99. Are Product A and Product B the same? No\n\nProduct A is title: battlefield 1942 : deluxe edition. manufacturer: electronic arts. price: 29.99. Product B is title: instant immersion language lab deluxe edition. manufacturer: . price: 14.9. Are Product A and Product B the same? No\n\n",
        "the-learning-company": "Product A is title: reader rabbit preschool. manufacturer: the-learning-company. price: 29.99. Product B is title: blue 's clues preschool. manufacturer: . price: 7.95. Are Product A and Product B the same? No\n\nProduct A is title: reader rabbit preschool. manufacturer: the-learning-company. price: 29.99. Product B is title: reader rabbit learn to read with phonics ! preschool & kindergarten. manufacturer: . price: 12.95. Are Product A and Product B the same? No\n\nProduct A is title: reader rabbit preschool. manufacturer: the-learning-company. price: 29.99. Product B is title: reader rabbit 1st grade. manufacturer: . price: 17.9. Are Product A and Product B the same? No\n\nProduct A is title: reader rabbit preschool. manufacturer: the-learning-company. price: 29.99. Product B is title: reader rabbit reading learning system 2007. manufacturer: . price: 12.95. Are Product A and Product B the same? No\n\nProduct A is title: treasure mathstorm. manufacturer: the-learning-company. price: 29.99. Product B is title: treasure mathstorm. manufacturer: the-learning-company. price: 9.9. Are Product A and Product B the same? Yes\n\n",
        "now software": "Product A is title: now up-to-date & contact v5 .0 ( mac ). manufacturer: now software. price: . Product B is title: now software nuw45e1k now up-to-date & contact for windows. manufacturer: now software. price: 114.99. Are Product A and Product B the same? No\n\nProduct A is title: now up-to-date & contact v5 .0 ( mac ). manufacturer: now software. price: . Product B is title: now software nuw45e1k now up-to-date & contact complete product content management 1 user ( s ) complete product standard pc. manufacturer: . price: 107.97. Are Product A and Product B the same? No\n\nProduct A is title: now up-to-date & contact v5 .0 ( mac ). manufacturer: now software. price: . Product B is title: now software nuw45e1k now up-to-date & contact for windows. manufacturer: power-on-software. price: 114.99. Are Product A and Product B the same? No\n\nProduct A is title: now up-to-date & contact v5 .0 ( mac ). manufacturer: now software. price: . Product B is title: now software nuw45e1k now up-to-date & contact for windows. manufacturer: now software. price: 114.99. Are Product A and Product B the same? No\n\n",
        "destineer publishing": "",
        "topics-entertainment-software - ( consignment )": "Product A is title: learn2 training for microsoft excel. manufacturer: topics-entertainment-software - ( consignment ). price: 9.99. Product B is title: learn2 training for microsoft powerpoint ( win 95 98 me nt 2000 xp ). manufacturer: topics-entertainment-software - ( consignment ). price: 9.99. Are Product A and Product B the same? No\n\nProduct A is title: learn2 training for microsoft powerpoint. manufacturer: topics-entertainment-software - ( consignment ). price: 9.99. Product B is title: microsoft office and windows training professional. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: learn2 training for microsoft excel. manufacturer: topics-entertainment-software - ( consignment ). price: 9.99. Product B is title: microsoft ( r ) excel 2007. manufacturer: . price: 109.99. Are Product A and Product B the same? No\n\nProduct A is title: learn2 training for microsoft powerpoint. manufacturer: topics-entertainment-software - ( consignment ). price: 9.99. Product B is title: topics entertainment learn2 training for microsoft excel. manufacturer: topics-entertainment-software - ( consignment ). price: 6.56. Are Product A and Product B the same? No\n\nProduct A is title: learn2 training for microsoft excel. manufacturer: topics-entertainment-software - ( consignment ). price: 9.99. Product B is title: topics entertainment learn2 training for microsoft excel. manufacturer: topics-entertainment-software - ( consignment ). price: 6.56. Are Product A and Product B the same? Yes\n\nProduct A is title: learn2 training for microsoft powerpoint. manufacturer: topics-entertainment-software - ( consignment ). price: 9.99. Product B is title: microsoft ( r ) powerpoint ( r ) 2002 upgrade. manufacturer: . price: 108.87. Are Product A and Product B the same? No\n\n",
        "sonicwall": "Product A is title: sonicwall gms standard edition 10 node license. manufacturer: sonicwall. price: . Product B is title: 01-ssc-7309 sonicwall cfs standard edition for sonicwall pro 1260/2040/3060 / 4060 subscript. manufacturer: . price: 2124.92. Are Product A and Product B the same? No\n\nProduct A is title: sonicwall gms standard edition 10 node license. manufacturer: sonicwall. price: . Product B is title: 01-ssc-5670 sonicwall cfs standard edition for sonicwall pro 4100 subscription license 1. manufacturer: . price: 857.38. Are Product A and Product B the same? No\n\nProduct A is title: sonicwall gms standard edition 10 node license. manufacturer: sonicwall. price: . Product B is title: 01-ssc-3363 sonicwall global management system standard edition license 10 nodes win. manufacturer: . price: 1835.31. Are Product A and Product B the same? No\n\nProduct A is title: sonicwall internet upgrade ( 50-node ). manufacturer: sonicwall. price: . Product B is title: sonicwall gms 1000 upgrade. manufacturer: . price: 63074.12. Are Product A and Product B the same? No\n\nProduct A is title: sonicwall gms standard edition 10 node license. manufacturer: sonicwall. price: . Product B is title: 01-ssc-7309 sonicwall cfs standard edition for sonicwall pro 1260/2040/3060 / 4060 subscript. manufacturer: . price: 2124.92. Are Product A and Product B the same? No\n\n",
        "alsoft": "Product A is title: alsoft diskwarrior : mac univeral binary. manufacturer: alsoft. price: . Product B is title: alsoft disk warrior v/5 cd. manufacturer: . price: 107.95. Are Product A and Product B the same? No\n\n",
        "arcmedia": "Product A is title: easy drums ( jewel case ). manufacturer: arcmedia. price: 9.99. Product B is title: cross stitch design studio ( jewel case ) ( 8006 ). manufacturer: m-and-r-technologies. price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: easy drums ( jewel case ). manufacturer: arcmedia. price: 9.99. Product B is title: poetry fiction and drama ( win/mac ) ( jewel case ) ( 10172 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: easy drums ( jewel case ). manufacturer: arcmedia. price: 9.99. Product B is title: geometry ( win/mac ) ( jewel case ) ( 10158 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\n",
        "re : launch": "Product A is title: piano wizard premier software & stickers & midi cable. manufacturer: re : launch. price: 149.95. Product B is title: intuit quicken premier 2007 software for windows tax & finance software. manufacturer: . price: 73.95. Are Product A and Product B the same? No\n\nProduct A is title: piano wizard premier software & stickers & midi cable. manufacturer: re : launch. price: 149.95. Product B is title: re : launch piano wizard premier software & stickers & midi cable. manufacturer: . price: 113.86. Are Product A and Product B the same? Yes\n\nProduct A is title: bryce 6. manufacturer: re : launch. price: 109.95. Product B is title: re : launch bryce 6. manufacturer: . price: 94.94. Are Product A and Product B the same? Yes\n\n",
        "grisoft": "Product A is title: grisoft avg anti-virus & anti-spyware 2 year subscription. manufacturer: grisoft. price: 39.95. Product B is title: 01-ssc-6993 sonicwall client/server anti-virus suite subscription license 3 years 50. manufacturer: . price: 3992.98. Are Product A and Product B the same? No\n\nProduct A is title: grisoft avg anti-virus & anti-spyware 2 year subscription. manufacturer: grisoft. price: 39.95. Product B is title: 01-ssc-6997 sonicwall client/server anti-virus suite subscription license 3 years 10. manufacturer: . price: 64740.21. Are Product A and Product B the same? No\n\nProduct A is title: grisoft avg anti-virus & anti-spyware 2 year subscription. manufacturer: grisoft. price: 39.95. Product B is title: 01-ssc-6991 sonicwall client/server anti-virus suite subscription license 3 years 10. manufacturer: . price: 863.87. Are Product A and Product B the same? No\n\nProduct A is title: grisoft avg anti-virus & anti-spyware 2 year subscription. manufacturer: grisoft. price: 39.95. Product B is title: grisoft avg anti-virus & anti-spyware. manufacturer: . price: 32.29. Are Product A and Product B the same? Yes\n\n",
        "mariner software": "Product A is title: mariner calc speadsheet ( mac ). manufacturer: mariner software. price: 59.99. Product B is title: mariner calc spreadsheet software. manufacturer: . price: 52.99. Are Product A and Product B the same? Yes\n\nProduct A is title: montage ( mac ). manufacturer: mariner software. price: 149.95. Product B is title: adobe illustrator cs3 ( mac ). manufacturer: . price: 599.0. Are Product A and Product B the same? No\n\n",
        "warner-new-media": "Product A is title: hospital tycoon for pc. manufacturer: warner-new-media. price: 29.99. Product B is title: roller coaster tycoon 3 for pc. manufacturer: . price: 29.99. Are Product A and Product B the same? No\n\nProduct A is title: hospital tycoon for pc. manufacturer: warner-new-media. price: 29.99. Product B is title: warner home video games hospital tycoon. manufacturer: warner-new-media. price: 26.33. Are Product A and Product B the same? Yes\n\n",
        "extensis corporation": "",
        "school zone": "Product A is title: time money & fractions grades 1-2. manufacturer: school zone. price: 9.99. Product B is title: time money & fractions grades 1-2 workbook & cd-rom. manufacturer: . price: 12.99. Are Product A and Product B the same? No\n\nProduct A is title: school zone flash action addition/subtraction ( windows/macintosh ). manufacturer: school zone. price: 12.99. Product B is title: school zone interactive flash action phonics made easy. manufacturer: . price: 8.8. Are Product A and Product B the same? No\n\nProduct A is title: school zone flash action addition/subtraction ( windows/macintosh ). manufacturer: school zone. price: 12.99. Product B is title: school zone flash action alphabet/numbers 1-100 ( pc & mac ) ( 08410 ). manufacturer: . price: 9.59. Are Product A and Product B the same? No\n\nProduct A is title: vocabulary puzzles grade 1. manufacturer: school zone. price: 14.95. Product B is title: school zone interactive vocabulary puzzles 1 on track software. manufacturer: . price: 9.45. Are Product A and Product B the same? Yes\n\n",
        "sony-pictures-home-entertainment": "Product A is title: vegas 6 for pc. manufacturer: sony-pictures-home-entertainment. price: 599.99. Product B is title: dramatica pro 4 for pc. manufacturer: . price: 249.99. Are Product A and Product B the same? No\n\nProduct A is title: vegas 6 for pc. manufacturer: sony-pictures-home-entertainment. price: 599.99. Product B is title: vegas + dvd sony. manufacturer: sony-pictures-digital-entertainment. price: 899.99. Are Product A and Product B the same? No\n\n",
        "quark": "Product A is title: quarkxpress 7 windows & macintosh ). manufacturer: quark. price: 899.0. Product B is title: quark 119700 quarkxpress 6.5 mac. manufacturer: quark. price: 736.99. Are Product A and Product B the same? No\n\nProduct A is title: quarkxpress 7 windows & macintosh ). manufacturer: quark. price: 899.0. Product B is title: quark xpress 7 ( no returns ). manufacturer: . price: 714.99. Are Product A and Product B the same? No\n\nProduct A is title: quarkxpress 7 windows & macintosh ). manufacturer: quark. price: 899.0. Product B is title: quark quarkxpress 7 upgrade for mac & windows. manufacturer: . price: 244.99. Are Product A and Product B the same? No\n\nProduct A is title: quarkxpress 7 for mac/win .. manufacturer: quark. price: 1289.42. Product B is title: quark xpress 7 ( no returns ). manufacturer: . price: 714.99. Are Product A and Product B the same? No\n\nProduct A is title: quarkxpress 7 windows & macintosh ). manufacturer: quark. price: 899.0. Product B is title: quark xpress 7 upgrade ( no returns ). manufacturer: . price: 239.99. Are Product A and Product B the same? No\n\n",
        "symantec corporation .": "",
        "strategy first": "Product A is title: great invasions. manufacturer: strategy first. price: 39.99. Product B is title: strategy first inc. great invasions. manufacturer: . price: 34.07. Are Product A and Product B the same? Yes\n\nProduct A is title: gods : lands of infinity special edition. manufacturer: strategy first. price: 29.99. Product B is title: reader rabbit kindergarten special 2-cd edition. manufacturer: . price: 12.95. Are Product A and Product B the same? No\n\nProduct A is title: alliance future combat. manufacturer: strategy first. price: 19.99. Product B is title: strategy first inc. alliance future combat. manufacturer: . price: 16.37. Are Product A and Product B the same? Yes\n\n",
        "power-on-software": "Product A is title: power on now up to date & contact windows. manufacturer: power-on-software. price: 129.99. Product B is title: powerhouse migo for ipod for windows. manufacturer: . price: 71.99. Are Product A and Product B the same? No\n\nProduct A is title: power on now up to date & contact windows. manufacturer: power-on-software. price: 129.99. Product B is title: now software nuw45e1k now up-to-date & contact for windows. manufacturer: power-on-software. price: 114.99. Are Product A and Product B the same? Yes\n\n",
        "vivendi": "Product A is title: world of warcraft pc. manufacturer: vivendi. price: 35.7. Product B is title: world of warcraft burning crusade. manufacturer: . price: 39.99. Are Product A and Product B the same? No\n\nProduct A is title: world of warcraft pc. manufacturer: vivendi. price: 35.7. Product B is title: vivendi-universal games inc 72305 world of warcraft 60 day prepaid. manufacturer: . price: 27.99. Are Product A and Product B the same? No\n\n",
        "bling": "Product A is title: podmaxx '07. manufacturer: bling. price: 29.99. Product B is title: podmaxx '07 eb carlson. manufacturer: . price: 29.99. Are Product A and Product B the same? Yes\n\nProduct A is title: safekeeper plus. manufacturer: bling. price: 39.99. Product B is title: bling software limited safekeeper plus. manufacturer: . price: 39.99. Are Product A and Product B the same? No\n\nProduct A is title: safekeeper plus. manufacturer: bling. price: 39.99. Product B is title: bling software limited safekeeper plus. manufacturer: . price: 34.75. Are Product A and Product B the same? Yes\n\n",
        "sos-aggregation-company": "Product A is title: web graphics creator. manufacturer: sos-aggregation-company. price: 39.95. Product B is title: enteractive inc fps creator. manufacturer: . price: 44.58. Are Product A and Product B the same? No\n\nProduct A is title: listen for ipod magnetic time. manufacturer: sos-aggregation-company. price: 39.99. Product B is title: iaudioize magnetic time for pc and mac. manufacturer: . price: 39.99. Are Product A and Product B the same? No\n\nProduct A is title: web graphics creator. manufacturer: sos-aggregation-company. price: 39.95. Product B is title: graphics software. manufacturer: . price: 99.92. Are Product A and Product B the same? No\n\nProduct A is title: listen for ipod magnetic time. manufacturer: sos-aggregation-company. price: 39.99. Product B is title: sos aggregation company iaudioize magnetic time. manufacturer: . price: 31.53. Are Product A and Product B the same? No\n\nProduct A is title: calculation skills v1 add & sub basic core learning. manufacturer: sos-aggregation-company. price: . Product B is title: sos aggregation company brain builder core learning. manufacturer: . price: 15.91. Are Product A and Product B the same? No\n\nProduct A is title: listen for ipod magnetic time. manufacturer: sos-aggregation-company. price: 39.99. Product B is title: listen for ipod magnetic time ( win 2000 xp ). manufacturer: sos-aggregation-company. price: 39.99. Are Product A and Product B the same? Yes\n\nProduct A is title: calculation skills v1 add & sub basic core learning. manufacturer: sos-aggregation-company. price: . Product B is title: sos aggregation company calculation skills v1 add & sub basic core learning. manufacturer: sos-aggregation-company. price: 15.91. Are Product A and Product B the same? Yes\n\n",
        "avanquest publishing usa inc.": "",
        "iris": "Product A is title: readiris pro 11 corporate edition ( mac ). manufacturer: iris. price: 499.95. Product B is title: iris inc readiris pro 11 corporate edition. manufacturer: . price: 430.58. Are Product A and Product B the same? Yes\n\nProduct A is title: irispen express translator. manufacturer: iris. price: 149.99. Product B is title: iris inc irispen express translator. manufacturer: . price: 142.29. Are Product A and Product B the same? Yes\n\nProduct A is title: readiris pro 11 corporate edition ( mac ). manufacturer: iris. price: 499.95. Product B is title: smart-sound sonicfire pro 4 corporate edition mac os x audio software for video. manufacturer: . price: 448.95. Are Product A and Product B the same? No\n\nProduct A is title: iriscard pro business card scanner. manufacturer: iris. price: 199.99. Product B is title: iris card pro. manufacturer: navarre distribution services. price: 198.87. Are Product A and Product B the same? Yes\n\nProduct A is title: readiris pro 11 corporate edition ( mac ). manufacturer: iris. price: 499.95. Product B is title: readiris pro 11 corporate edition for pc. manufacturer: . price: 349.99. Are Product A and Product B the same? No\n\nProduct A is title: iriscard pro business card scanner. manufacturer: iris. price: 199.99. Product B is title: encore inc printshop business card maker. manufacturer: riverdeep-learning-company. price: 8.01. Are Product A and Product B the same? No\n\nProduct A is title: iriscard pro business card scanner. manufacturer: iris. price: 199.99. Product B is title: belight business card composer. manufacturer: . price: 39.99. Are Product A and Product B the same? No\n\nProduct A is title: readiris pro 11 corporate edition ( mac ). manufacturer: iris. price: 499.95. Product B is title: iris readiris pro 11 mac ocr. manufacturer: . price: 109.99. Are Product A and Product B the same? No\n\nProduct A is title: iriscard pro business card scanner. manufacturer: iris. price: 199.99. Product B is title: iriscard pro. manufacturer: . price: 174.99. Are Product A and Product B the same? Yes\n\nProduct A is title: irispen express translator. manufacturer: iris. price: 149.99. Product B is title: iris inc irispen express translator. manufacturer: . price: 142.29. Are Product A and Product B the same? Yes\n\n",
    },
    f"{DATASET_PATH}/data_imputation/Buy": "name: Transcend 8GB Compact Flash Card (133x) - TS8GCF133. description: Transcend 8GB CompactFlash Card (133x). Who is the manufacturer? TRANSCEND INFORMATION\n\nname: LG 42LG30 42' LCD TV. description: LG 42LG30 42' LCD HDTV - 12,000:1 Dynamic Contrast Ratio - Invisible Speaker. Who is the manufacturer? LG Electronics\n\nname: Speck Products SeeThru Case for Apple MacBook Air - MBA-PNK-SEE. description: Plastic - Pink. Who is the manufacturer? Speck Products\n\nname: Peerless Universal Tilt Wall Mount. description: Peerless Smart Mount ST660P Universal Tilt Wall Mount for 37' to 60' Screens (Black) up to 200lbs. Who is the manufacturer? Peerless\n\nname: Apple Time Capsule Network Hard Drive - MB277LL/A. description: 1TB - Type A USB. Who is the manufacturer? Apple\n\nname: Sirius SUPH1 Sirius Universal Home Kit. description: Sirius Satellite Radio Receiver. Who is the manufacturer? Sirius\n\nname: OmniMount TV Top Shelf Mount. description: OmniMount CCH1B Set-Top Center-Channel Shelf. Who is the manufacturer? OMNIMOUNT SYSTEMS, INC\n\nname: Monster Cable iFreePlay Cordless Headphone - AI SH HPHONE. description: Connectivity: Wireless - Stereo - Behind-the-neck. Who is the manufacturer? Monster\n\nname: Pure Digital Flip Mino Digital Camcorder - F360B. description: Flip Video Mino 60 min Black. Who is the manufacturer? Pure Digital Technol\n\nname: Elgato EyeTV Hybrid Analog/Digital TV Tuner Stick - 10020630. description: Elgato EyeTV Hybrid TV Tuner Stick for Analog and HDTV Reception - USB. manufacturer? ELGATO SYSTEMS\n",
    f"{DATASET_PATH}/data_imputation/Restaurant": "name: oceana. addr: 55 e. 54th st.. phone: 212/759-5941. type: seafood. What is the city? new york\n\nname: oceana. addr: 55 e. 54th st.. phone: 212-759-5941. type: seafood. What is the city? new york city\n",
    f"{DATASET_PATH}/error_detection/Hospital": {
        "EmergencyService": "Is there a x spelling error in EmergencyService: nan? No\n\nIs there a x spelling error in EmergencyService: yes? No\n\nIs there a x spelling error in EmergencyService: yes? No\n\nIs there a x spelling error in EmergencyService: yex? Yes\n",
        "City": "Is there a x spelling error in City: nan? No\n\nIs there a x spelling error in City: birmingham? No\n\nIs there a x spelling error in City: sheffield? No\n\nIs there a x spelling error in City: birminghxm? Yes\n",
        "Score": "Is there a x spelling error in Score: nan? No\n\nIs there a x spelling error in Score: 78%? No\n\nIs there a x spelling error in Score: 1xx%? Yes\n",
        "HospitalType": "Is there a x spelling error in HospitalType: nan? No\n\nIs there a x spelling error in HospitalType: acute care hospitals? No\n\nIs there a x spelling error in HospitalType: acute care hospitals? No\n\nIs there a x spelling error in HospitalType: acutexcarexhospitals? Yes\n",
        "CountyName": "Is there a x spelling error in CountyName: nan? No\n\nIs there a x spelling error in CountyName: calhoun? No\n\nIs there a x spelling error in CountyName: marion? No\n\nIs there a x spelling error in CountyName: xhambers? Yes\n",
        "MeasureName": "Is there a x spelling error in MeasureName: nan? No\n\nIs there a x spelling error in MeasureName: pneumonia patients given the most appropriate initial antibiotic(s)? No\n\nIs there a x spelling error in MeasureName: heart attack patients given aspirin at discharge? No\n\nIs there a x spelling error in MeasureName: allxheartxsurgeryxpatientsxwhosexbloodxsugarx(bloodxglucose)xisxkeptxunderxgoodxcontrolxinxthexdaysxrightxafterxsurgery? Yes\n",
        "HospitalName": "Is there a x spelling error in HospitalName: nan? No\n\nIs there a x spelling error in HospitalName: fayette medical center? No\n\nIs there a x spelling error in HospitalName: huntsville hospital? No\n\nIs there a x spelling error in HospitalName: medixal xenter enterprise? Yes\n",
        "Condition": "Is there a x spelling error in Condition: nan? No\n\nIs there a x spelling error in Condition: heart attack? No\n\nIs there a x spelling error in Condition: surgical infection prevention? No\n\nIs there a x spelling error in Condition: hearx axxack? Yes\n",
        "Stateavg": "Is there a x spelling error in Stateavg: nan? No\n\nIs there a x spelling error in Stateavg: al_pn-6? No\n\nIs there a x spelling error in Stateavg: al_scip-inf-3? No\n\nIs there a x spelling error in Stateavg: al_scipxinfx2? Yes\n\nIs there a x spelling error in Stateavg: al_ami-7a? No\n\nIs there a x spelling error in Stateavg: alxami-7a? Yes\n",
        "MeasureCode": "Is there a x spelling error in MeasureCode: nan? No\n\nIs there a x spelling error in MeasureCode: ami-7a? No\n\nIs there a x spelling error in MeasureCode: scip-inf-2? No\n\nIs there a x spelling error in MeasureCode: xax-1? Yes\n",
        "ZipCode": "Is there a x spelling error in ZipCode: nan? No\n\nIs there a x spelling error in ZipCode: 35640? No\n\nIs there a x spelling error in ZipCode: 36302? No\n\nIs there a x spelling error in ZipCode: x6x05? Yes\n",
        "Sample": "Is there a x spelling error in Sample: nan? No\n\nIs there a x spelling error in Sample: 13 patients? No\n\nIs there a x spelling error in Sample: 97 patients? No\n\nIs there a x spelling error in Sample: 4 xatients? Yes\n",
        "HospitalOwner": "Is there a x spelling error in HospitalOwner: nan? No\n\nIs there a x spelling error in HospitalOwner: proprietary? No\n\nIs there a x spelling error in HospitalOwner: government - hospital district or authority? No\n\nIs there a x spelling error in HospitalOwner: goverxmext - hospital district or authority? Yes\n",
        "Address1": "Is there a x spelling error in Address1: nan? No\n\nIs there a x spelling error in Address1: 126 hospital ave? No\n\nIs there a x spelling error in Address1: 2000 pepperell parkway? No\n\nIs there a x spelling error in Address1: 400xnxedwardsxstreet? Yes\n",
        "Address2": "Is there a x spelling error in Address2: nan? No\n\nIs there a x spelling error in Address2: ami-7a? No\n\nIs there a x spelling error in Address2: scip-inf-2? No\n\nIs there a x spelling error in Address2: xax-1? Yes\n",
        "Address3": "Is there a x spelling error in Address3: nan? No\n\nIs there a x spelling error in Address3: ami-7a? No\n\nIs there a x spelling error in Address3: scip-inf-2? No\n\nIs there a x spelling error in Address3: xax-1? Yes\n",
        "ProviderNumber": "Is there a x spelling error in ProviderNumber: nan? No\n\nIs there a x spelling error in ProviderNumber: 10005? No\n\nIs there a x spelling error in ProviderNumber: 10055? No\n\nIs there a x spelling error in ProviderNumber: 1xx32? Yes\n",
        "State": "Is there a x spelling error in State: nan? No\n\nIs there a x spelling error in State: al? No\n\nIs there a x spelling error in State: al? No\n\nIs there a x spelling error in State: ax? Yes\n",
        "PhoneNumber": "Is there a x spelling error in PhoneNumber: nan? No\n\nIs there a x spelling error in PhoneNumber: 3343762205? No\n\nIs there a x spelling error in PhoneNumber: 2562651000? No\n\nIs there a x spelling error in PhoneNumber: 2568453x50? Yes\n",
    },
}
